/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 */
#ifndef __MEGAIDE_DEBUG_H__
#define __MEGAIDE_DEBUG_H__

#ifdef MIDE_DEBUG
# define debug(level, fmt, args...)	do {if (level <= MIDE_DEBUG) \
    printk_Rlsi("%s: " fmt "\n", __FUNCTION__ , ##args);} while(0)
# define debug_called(level) \
    do {if (level <= MIDE_DEBUG) printk_Rlsi("%s: called\n", __FUNCTION__);} while(0)
#else
#    define debug(level, fmt, args...)
#    define debug_called(level)
#endif

#define xdebug(fmt, args...)	printk_Rlsi("%s: " fmt "\n", __FUNCTION__ , ##args)

/*
*-------------------------------------------------------------------
* Macros Definition
*-------------------------------------------------------------------
*/

#define AR_DEBUG_EVERYTHING		((u_long)0xFFFFFFFF)
#define AR_DEBUG_MODULE			((u_long)0x10000000)

#define AR_DEBUG_DRVR_ENTRY		((u_long)0x00000001)
#define AR_DEBUG_INITDEV		((u_long)0x00000002)
#define AR_DEBUG_FINDDEV		((u_long)0x00000004)
#define AR_DEBUG_SCAN_OP_ROM		((u_long)0x00000008)

#define AR_DEBUG_WARNING		((u_long)0x00000010)
#define AR_DEBUG_ERRORS		((u_long)0x00000020)
#define AR_DEVEL_DEBUG			((u_long)0x000000F0)

#define AR_DEBUG_SCSI_CMD		((u_long)0x00000100)
#define AR_DEBUG_SCSI_CMD1	((u_long)0x00000200)
#define AR_DEBUG_SCSI_CMD2	((u_long)0x00000400)
#define AR_DEBUG_INTERUPT		((u_long)0x00000800)

#define AR_DEBUG_RELEASE		((u_long)0x00001000)
#define AR_DEBUG_IOCTL			((u_long)0x00002000)
#define AR_DEBUG_ATA			((u_long)0x00004000)
#define AR_DEBUG_SGLIST			((u_long)0x00008000)

#define AR_DEBUG_CONFIG		((u_long)0x00010000)
#define AR_UPDATE_IRCD			((u_long)0x00020000)
#define AR_BP						((u_long)0x00040000)
#define AR_DEBUG_PASSTHRU	((u_long)0x00080000)

#define AR_DEBUG_VANI_CODE		((u_long)0x00100000)
#define AR_DEBUG_ERROR_RECOVERY		((u_long)0x00200000)
#define AR_DEBUG_DYN_DMA			((u_long)0x00400000)
#define AR_DEBUG_CHAR_IOCTL		((u_long)0x00800000)
#define AR_DEBUG_SATA				((u_long)0x01000000)
#define AR_DEBUG_NEW_IRCD		((u_long)0x02000000)
#define AR_DEBUG_48BIT				((u_long)0x04000000)
#define AR_DEBUG_PM					((u_long)0x08000000)

/* global parmeter used in mide_debug.h*/
/* u_long debug_level=AR_DEBUG_SCSI_CMD | AR_DEBUG_SCSI_CMD1; */
extern u_long debug_level;

#define INTERNAL_REBUILD

#define DEBUG_ON 1

#ifdef DEBUG_ON
#define BREAKPOINT		{asm("int $3");}
#define AR_DEBUG(level, msg_string) {\
    if(debug_level & level) {\
	printk_Rlsi msg_string;	\
    } \
}
#else
    #define BREAKPOINT
    #define AR_DEBUG(level,msg_string) /* DO NOTHING */
#endif /*AR_DEBUG*/

/*
*-------------------------------------------------------------------
* Structure Definition
*-------------------------------------------------------------------
*/


/*
*-------------------------------------------------------------------
* Function Protypes
*-------------------------------------------------------------------
*/

#endif /*__MEGAIDE_DEBUG_H__*/
/*vi set ts=8:*/
