/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/inc/mide_devext.h,v emoore Exp$
 */

#ifndef __MEGAIDE_DEVEXT_H__
#define __MEGAIDE_DEVEXT_H__
/* megaide_devext.h */

#define		PROCBUFSIZE			4096

#ifdef BP_SUPPORT

/*
Initialize Backplane Structure
*/

typedef
void
(*BP_EVENT_CALLBACK_ROUTINE)(unsigned long status);

typedef
ULONG
(*BP_ENTRY_POINT)(unsigned int cmd, unsigned long arg);

typedef struct _INIT_BACKPLANE
{
    BP_EVENT_CALLBACK_ROUTINE callback;
    BP_ENTRY_POINT bp_control;
    USHORT	MajorVersion;
    USHORT	MinorVersion;
    ULONG	Capability;
}
INIT_BACKPLANE, *PINIT_BACKPLANE;


typedef struct  _BACKPLANE_DATA {
    UCHAR	bpRegistered;
    UCHAR	NotifyController;
    BP_ENTRY_POINT pIdeBackplaneControl;
    BP_EVENT_CALLBACK_ROUTINE pCallBackFunc;
    INIT_BACKPLANE init_bp;
    BP_STATUS_INFO bpControlStatus;
    UCHAR	StartEvent_Index;
    UCHAR	EndEvent_Index;
    ADD_EVENT	EventLogBuffer[256];
} BACKPLANE_DATA, *PBACKPLANE_DATA;

#endif

typedef enum _IRCD_TYPE
{
    NO_IRCD = 0,
    XIDE_IRCD = 1,
    _IDE_IRCD = 2
}IRCD_TYPE;


typedef enum _CONTROLLER_TYPE
{
    CMD_648 = 0,
    CMD_649 = 1,
    CMD_680 = 2,
    RCC_SERVER_WORKS_CSB5 = 3,
    SIL_3112 = 4,
    RCC_SERVER_WORKS_CSB6 = 5
}CONTROLLER_TYPE;


/*
SGL definition for DMA operations.
*/

typedef struct _PHYSICAL_SGL_ENTRY {
    PVOID	Address;
    ULONG	Length:31;
    ULONG	EndOfListFlag:1;
} PHYSICAL_SGL_ENTRY, *PPHYSICAL_SGL_ENTRY;

/*
SGL definition for PIO operations.
*/
typedef struct _LOGICAL_SGL_ENTRY {
    PVOID	Address;
    ULONG	Length:31;
    ULONG	EndOfListFlag:1;
} LOGICAL_SGL_ENTRY, *PLOGICAL_SGL_ENTRY;

/*
SGL definition for PIO operations.
*/
typedef struct _PRINTABLE_SGL_ENTRY {
   PVOID	Address;
   ULONG	Length;
} PRINTABLE_SGL_ENTRY, *PPRINTABLE_SGL_ENTRY;

/*
Scatter/Gather list definition.
*/
typedef union _SGL_ENTRY {
    PHYSICAL_SGL_ENTRY Physical;
    LOGICAL_SGL_ENTRY Logical;
    PRINTABLE_SGL_ENTRY Printable;
} SGL_ENTRY, *PSGL_ENTRY;

typedef struct _COMMAND_SPLIT_STATUS {
    union {
	struct {
	    UCHAR Drive0:1;
	    UCHAR Drive1:1;
	    UCHAR Drive2:1;
	    UCHAR Drive3:1;
	    UCHAR Reserved:4;
	}Bits;
	UCHAR Drives;
    }u;
}COMMAND_SPLIT_STATUS, *PCOMMAND_SPLIT_STATUS;


typedef struct _FAST_IDE_SCB_EXTENSION {
    COMMAND_SPLIT_STATUS CmdBuildStatus;
    ULONG	NumberOfPendingCmd;
    UCHAR	Reserved[2];
    UCHAR	CmdStatus;
    UCHAR	buffer_is_mapped;
    VOID 	*mide_buffer_dmamap;
    VOID 	*mide_buffer_dmat;
    UINT	scb_queue_index;
    SGL_ENTRY	SglEntry[M_IDE_MAX_SGLIST];
/* scsi cmd stuff copied here */
    UINT	target;
    UINT	lun;
    UINT	channel;
    UCHAR	scsi_cdb[16];
    UCHAR	cdb_len;
    VOID 	*request_buffer;
    VOID	*ioctl_buffer;
    UCHAR	 IoctlFlag;
    USHORT	sglem;
    VOID 	*pstScsiCmd;
/* PHYSICAL_DRIVE_COMMAND PhysicalDrvCmd[MAX_DRIVES_PER_CONTROLLER];*/
    UCHAR	PhysicalDrvCmd[1];
}
__attribute__ ((packed))
FAST_IDE_SCB_EXTENSION, *PFAST_IDE_SCB_EXTENSION;

/* scsi_malloc() must be on 512 boundry's,, no larger than 4K
we dont check for the upper 4K boundry ( since we should never have that )
#define SCB_EXTENSION_SCSI_MALLOC_SIZE   ( ( sizeof( FAST_IDE_SCB_EXTENSION ) / 512 + 1 ) * 512 )
#define SCB_EXTENSION_SCSI_MALLOC_SIZE ( ( get_scb_extension_size() / 512 + 1 ) * 512 )*/

/*Logical drive descriptor*/
typedef struct _LOGICAL_DRIVE_INFO{
    MP_RAID_LEVEL RaidLevel;
    UCHAR	LogicalDriveNumber;
    MP_LOGICAL_DRIVE_STATUS Status;
    ULONG	Sectors; /* Total number of available sectors.*/
    ULONG	StripeSize; /* In 512-byte units.*/
    UCHAR	StripesPerRow;
    UCHAR	PhysicalDriveCount;
    UCHAR	PhysicalDriveTid[MAX_DRIVES_PER_CONTROLLER];
    UCHAR	SlotPhysicalDriveTid[MAX_DRIVES_PER_CONTROLLER];
    ULONG	StartBlock; /* start block on phy drive below */
    BOOLEAN	IsBkOperationStarted;
    BOOLEAN	IsAbortReceived;
    ULONG	EachArmLength;
    ULONG	RebuildStartBlock;
    ULONG	CheckConStartBlock;
    ULONG	CompletedBlocks; /* number of completed blocks within the log drive*/
    ULONG	ulInconsistencyCount;
    ULONG	ulErrorCountDuringBackgroundTask;
    UCHAR	ucLastBackgroundTaskOperation;
    MP_ACTIVITY_STATUS ucLastBackgroundTaskStatus;
/* GATE_KEEPER	GateKeeper; */
    void 	*GateKeeper;
/* FAST_IDE_SCB_EXTENSION	InternalScbExtension; */
    PFAST_IDE_SCB_EXTENSION InternalScbExtension;
    UCHAR	ucControllerId;
    UCHAR	TagDisable;
    struct _ARRAY_DRIVE_INFO *pArrayInfo;
    struct _LOGICAL_DRIVE_INFO *NextLogDrive;
    struct _LOGICAL_DRIVE_INFO *PrevLogDrive;
}
__attribute__ ((packed))
LOGICAL_DRIVE_INFO, *PLOGICAL_DRIVE_INFO;


/*Array drive descriptor*/
/* +++ edm 1 March 2002 - */
typedef struct _ARRAY_DRIVE_INFO{
    UCHAR	ucArrayDriveNumber;
    MP_RAID_LEVEL RaidLevel;
    MP_LOGICAL_DRIVE_STATUS Status;
    UCHAR	ucLogicalDriveCount;
    PLOGICAL_DRIVE_INFO HeadLogDrive;
    PLOGICAL_DRIVE_INFO TailLogDrive;
}
__attribute__ ((packed))
ARRAY_DRIVE_INFO, *PARRAY_DRIVE_INFO;
/* --- edm 1 March 2002 - */

/* Physical drive descriptor.*/
typedef struct _PHYSICAL_DRIVE_INFO {
    UCHAR	PhysicalDriveId; /*0,1,2,3 for four physcial Drive*/
    UCHAR	LogicalDriveId;	 /* these are needed for old IRCD*/
    UCHAR	LogicalIndex;	 /* these are needed for old IRCD*/
    ULONG	BiosExposedSectors;
    UCHAR	ucMirrorDriveId;
    PARRAY_DRIVE_INFO pArrayInfo; /* points to the logical data area of this drive*/
    UCHAR	ucConfigurationWasSwitched; /* we switched the Config in when reading IRCD*/
    UCHAR	error_log_needs_reading;
    USHORT	ErrorFound; 	/* count of total errors*/
/* Dell is wanting us to program down the drives if we get 3 sequentail errors on a physical drive
this is the counter for this */
    USHORT	SequentialErrorCount;
/* PERROR_LOG_SECTOR	ErrorLogSector;*/
    void 	*ErrorLogSector;
    SGL_ENTRY	DiskErrorLogSgl;
    ULONG	ErrorLogSectorIndex;
    USHORT	ErrorReported;
    ULONG	IrcdSectorIndex;
    ULONG	EnquirySectorIndex;
    ULONG	DeviceFlags;
    BOOLEAN	bEnableRwCache;
    BOOLEAN	bEnableReadAheadCache;
    BOOLEAN	ucIsPowerConnected;
    MP_TRANSFER_MODE TransferMode;
    ULONG	StartBlock;
    ULONG	NumberOfBlocks;
    ULONG	NumberOfBlocks48bit; /*manoj 0122*/
    MP_PHYSICAL_DRIVE_STATUS Status;
    BOOLEAN	bIsSpareDrive;
    IDENTIFY_DATA IdentifyData;
    ULONG	MaximumBlockXfer;
    UCHAR	RetryCount;
    ULONG	LastBlockRead;
    ULONG	BadErrorSector;	/* location of last error found on drive*/
    UCHAR	errorByte;      /*type of error found on drive.*/
    BOOLEAN	IsFrozenQueue;
    ULONG	CurrentQueueLength;
/* PPHYSICAL_DRIVE_COMMAND HoldQueue;  */
    void 	*HoldQueue;  /*Head Pointer */
/* PPHYSICAL_DRIVE_COMMAND TailQueue;	*/
    void 	*TailQueue; /*Tail pointer*/
    UCHAR	ucControllerId; /*Nagendra 10/10/01*/
    BOOLEAN	bErrorLogUpdatePending;
}
__attribute__ ((packed))
PHYSICAL_DRIVE_INFO, *PPHYSICAL_DRIVE_INFO;

/*
* Channel descriptor.
*/
typedef struct _CHANNEL{
    UCHAR	LastDriveFed;
    UCHAR	SwitchDrive;
    PSGL_ENTRY	pSglEntry;
    PSGL_ENTRY	SglBaseVirtualAddress;
    ULONG	SglBasePhysicalAddress;
    ULONG	MaxSglEntries;
    PIDE_REGISTERS_1 BaseIoAddress1;
    PIDE_REGISTERS_2 BaseIoAddress2;
    PBM_REGISTERS BaseBmAddress;
   PCONTROL_FUNTION_REG_SIL680 CntrlFunc; /* manoj 1213 Sil Image 680 Base address 5 Register */
/* resource handles */
    void 	*base1_resource;
    void 	*base2_resource;
    void 	*bmaddr_resource;
/* sglist tags */
    void 	*mide_sg_dmamap;
    void 	*mide_sg_dmat;
/*	PPHYSICAL_DRIVE_COMMAND ActiveQueue; */
    void 	*ActiveQueue;
    UCHAR	ExpectingInterrupt;
    UCHAR	ucActiveCmd;
    UCHAR	ulActiveTargetId;
    UCHAR	SglElemCnt;

    SGL_ENTRY DiskConfigurationSgl;
    PUCHAR	DiskConfiguration;
    SGL_ENTRY DiskEnquirySgl;
    PUCHAR	DiskEnquiry;
    PASS_THRU_DATA passthru_data;
    UCHAR	ChannelBusy;
    UCHAR  ucCurrentPMMode;
    ULONG	ulActiveCmdTimeOut;
    BOOLEAN	AltStatus1Byte;

}
__attribute__ ((packed))
CHANNEL, *PCHANNEL;

/*
* Device extension.
*/

typedef struct _HW_DEVICE_EXTENSION
{
    UCHAR	ucControllerId;
    ULONG	ulVendorId; /* PCI PNP IDs.*/
    ULONG	ulDeviceId;
    USHORT	usSubsysVendorId;
    USHORT	usSubsysDeviceId;
    ULONG	ulUniqueId;
    ULONG	ulProdId;
    USHORT	usSlot;
    UCHAR	ucPCIBus;
    UCHAR	ucPCIDev;
    UCHAR	ucPCIFun;
    ULONG	ulIntLine; /* Base register locations.*/
/* resource handles */
    void 	*isr_resource;		/* interrupt resource */
    void 	*isr_handle;            /* interrupt handle */
    CONTROLLER_SPEED ControllerSpeed; /* Controller speed (UDMA/33, UDMA/66, etc.)*/
    ULONG	MaximumTransferLength;
    ULONG	ulLogicalDriveCount;
    ULONG	ulLastChannelFed;
    ULONG	ulArrayDriveCount;
    ARRAY_DRIVE_INFO ArrayDrive[MAX_DRIVES_PER_CONTROLLER];
    LOGICAL_DRIVE_INFO LogicalDrive[MAX_LOGICAL_DRIVES];
    UCHAR	BootLogDrvID;
    UCHAR	BootLogDrvMapping[MAX_LOGICAL_DRIVES];
    PHYSICAL_DRIVE_INFO PhysicalDrive[MAX_DRIVES_PER_CONTROLLER];
    CHANNEL	Channel[MAX_CHANNELS_PER_CONTROLLER];
/* PGATE_KEEPER	ActiveGateKeeper;*/
    void *base5_resource;
    ULONG BaseMemAddress5;
    UCHAR  *virtBaseMemAddress5;
    void 	*ActiveGateKeeper;
    BOOLEAN	IsUsed;
    PUCHAR	GateKeeperBuffer;
    ULONG	GateKeeperPhysicalBuffer;
    void 	*GateKeeperBufferDmat;
    void 	*GateKeeperBufferDmaMap;
    ULONG	GateKeeperBufferLength;
    UCHAR	ConfigurationVirtualBuffer[IDE_SECTOR_SIZE];
    UCHAR	EnquiryVirtualBuffer[ IDE_SECTOR_SIZE ];
    ULONG	NumberOfPendingCmd;
    ULONG	NumberOfPendingScb;
    UCHAR	BusyTarget;
#ifdef	BP_SUPPORT
    BP_BEEP	bBuzzerEnable;
    BACKPLANE_DATA BackPlane;
#endif
    COMMAND_SPLIT_STATUS drive_inserted;
    COMMAND_SPLIT_STATUS drive_removed;
    /* GATE_KEEPER	GateKeeper;*/
    void 	*GateKeeper;
/* FAST_IDE_SCB_EXTENSION	GateScbExtension;*/
    PFAST_IDE_SCB_EXTENSION GateScbExtension;
    MP_IDE_VERSION stFirmwareVersion;
    MP_IDE_VERSION stIRCDVersion;
    IRCD_TYPE	ucIRCDType;
    CONTROLLER_TYPE ucCardType;
    ULONG	uiFlag;
    UCHAR	ucErrorRecovery;
    UCHAR	aucChannelList[MAX_CHANNELS_PER_CONTROLLER];
    char	achProcBuf[PROCBUFSIZE];
    int		iProcIdx;
/* data buffer tags */
    void 	*mide_buffer_dmat;
    void 	*parent_sglist_dmat;
    void 	*parent_datatransfer_dmat;
    PFAST_IDE_SCB_EXTENSION scb_queue;
    BOOLEAN	scb_queue_in_use[M_IDE_MAX_COMMANDS];
    int		last_free_scb;
    BOOLEAN	resource_shortage;
    int		FreezeCnt;
    UCHAR PMHandle;
    UCHAR PMState;
#ifdef  ENABLE_ERROR_INJECTOR
    ULONG	ErrorAfterCounter;
    PDErrorInjectorInfo pdErrorInjectorInfo[ MAX_DRIVES_PER_CONTROLLER];
    ULONG	ErrorCounter[10];
	struct {
	    ULONG	sector;
	    BOOLEAN 	isWrite;
	    ULONG	pattern[MAX_STATE];
	}ErrorPattern[MAX_PATTERN];
#endif /* ENABLE_ERROR_INJECTOR*/
}
__attribute__ ((packed))
HW_DEVICE_EXTENSION, *PHW_DEVICE_EXTENSION;

#endif /* __MEGAIDE_DEVEXT_H__*/
