/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/inc/mide_macros.h,v emoore Exp$
 */

#ifndef __MACROS_H__
#define __MACROS_H__

/*
*-------------------------------------------------------------------
* Macros Definition
*-------------------------------------------------------------------
*/


/*
* Macros for port communication
* This is a Linux dependent code
*/

#define STM_PIO(mode)			((UCHAR) ((1 << 3) | (mode)))
#define STM_MULTIWORD_DMA(mode)		((UCHAR) ((1 << 5) | (mode)))
#define STM_UDMA(mode)			((UCHAR) ((1 << 6) | (mode)))


#define ABS_DIFF(pos1, pos2) \
    (ULONG)((pos1 > pos2) ? (pos1-pos2) : (pos2 - pos1))

/*
* Beautification macros
*/


#define BIG_ENDIAN_ULONG(Ulong) \
    (((PUCHAR)&Ulong)[0] << 24) |\
    (((PUCHAR)&Ulong)[1] << 16) |\
    (((PUCHAR)&Ulong)[2] << 8) |\
    ((PUCHAR)&Ulong)[3]

#define BIG_ENDIAN_USHORT(Ushort) \
    (((PUCHAR)&Ushort)[0] << 8) |\
    (((PUCHAR)&Ushort)[1] >> 8)

/*manoj 1213 Macros to Disable and Enable interrupt for Sil 680 and CMD 649*/
#define GET_DEVICE_ID(DeviceExtension) \
    DeviceExtension->ulDeviceId

/*manoj 1213 Check for 48 bit addressable support*/
#define IS_48_BIT_ENABLED(targetId)\
    ((DeviceExtension->PhysicalDrive[targetId].IdentifyData.CmdSupported2 &(1<<10))!=0)

/*edm only let drives over 28bit use 48bit addressing*/
#define IS_48_BIT_ADDESSABLE(targetId)\
    ( IS_48_BIT_ENABLED(targetId) && \
    ( DeviceExtension->PhysicalDrive[targetId].NumberOfBlocks > 0xFFFFFFF ) )

#define USES_DMA(TargetId) \
    ((DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_USES_EITHER_DMA) != 0)

#define SELECT_DEVICE(resource, BaseIoAddress, TargetId) \
    ScsiPortWritePortUchar(\
	resource, \
	(unsigned)&BaseIoAddress->DriveSelect, \
	(UCHAR)((((TargetId) & 1) << 4) | IDE_CHS_MODE));\
	ScsiPortStallExecution(60)

#define SELECT_CHS_DEVICE(resource, BaseIoAddress, TargetId, HeadNumber) \
    ScsiPortWritePortUchar(\
	resource, \
	(unsigned)&BaseIoAddress->DriveSelect, \
	(UCHAR)((((TargetId) & 1) << 4) | IDE_CHS_MODE | (HeadNumber & 15)));\
	ScsiPortStallExecution(60)

#define SELECT_LBA_DEVICE(resource, BaseIoAddress, TargetId, Lba) \
    ScsiPortWritePortUchar(\
	resource, \
	(unsigned)&BaseIoAddress->DriveSelect, \
	(UCHAR)((((TargetId) & 1) << 4) | IDE_LBA_MODE | (Lba & 0x0f000000) >> 24));\
	ScsiPortStallExecution(60)

/*manoj 0122 for 48bit addressing support*/
#define SELECT_LBA_DEVICE_48BIT(resource, BaseIoAddress, TargetId) \
    ScsiPortWritePortUchar(\
	resource, \
	(unsigned)&BaseIoAddress->DriveSelect, \
	(UCHAR)((((TargetId) & 1) << 4) | IDE_LBA_MODE ));\
	ScsiPortStallExecution(60)

#define GET_STATUS(resource, BaseIoAddress1, Status) \
    Status = ScsiPortReadPortUchar( resource, (unsigned)&BaseIoAddress1->Command);

#define GET_BASE_STATUS(resource, BaseIoAddress1, Status) \
    ScsiPortStallExecution(1);\
    Status = ScsiPortReadPortUchar(resource, (unsigned)&BaseIoAddress1->Command);


#define WRITE_COMMAND(resource, BaseIoAddress1, Command) \
    ScsiPortWritePortUchar( resource, (unsigned)&BaseIoAddress1->Command, Command);

#define AR_READ_BUFFER(resource,BaseIoAddress, Buffer, Count) \
    ScsiPortReadPortBufferUshort(resource,(unsigned)(&BaseIoAddress->Data), \
	Buffer, \
	Count);

#define AR_WRITE_BUFFER(resource,BaseIoAddress, Buffer, Count) \
    ScsiPortWritePortBufferUshort(resource,(unsigned)(&BaseIoAddress->Data), \
	Buffer, \
	Count);

#define AR_READ_BUFFER2(resource,BaseIoAddress, Buffer, Count) \
    ScsiPortReadPortBufferUlong(resource, (unsigned)(&BaseIoAddress->Data), \
	Buffer, \
	Count);

#define AR_WRITE_BUFFER2(resource,BaseIoAddress, Buffer, Count) \
    ScsiPortWritePortBufferUlong( resource, (unsigned)(&BaseIoAddress->Data), \
	Buffer, \
	Count);

/*
WAIT_IN_BUSY waits for up to 1 s.
*/

#define WAIT_ON_BUSY(resource, BaseIoAddress, Status) {\
    ULONG i; \
    for (i = 0; i < 200000; i++) { \
	GET_STATUS(resource, BaseIoAddress, Status); \
	if (Status & (IDE_STATUS_BUSY/*|IDE_STATUS_DRQ*/)) { \
	    ScsiPortStallExecution(5); \
	    continue; \
	} else { \
	    break; \
	} \
    } \
}

/*#define WAIT_ON_ALTERNATE_STATUS_BUSY(resource, BaseIoAddress2, Status) \ Vani: 0911*/
#define WAIT_ON_ALTERNATE_STATUS_BUSY(resource, baseIoAddress2, Status, AltStatus1Byte) {\
    ULONG i; \
    for (i = 0; i < 200000; i++) { \
	Status = GET_ALTERNATE_STATUS_REGISTER_VALUE(resource, baseIoAddress2, AltStatus1Byte); \
	if (Status & (IDE_STATUS_BUSY/*|IDE_STATUS_DRQ*/)) { \
	    ScsiPortStallExecution(5); \
	    continue; \
	} else { \
	    break; \
	} \
    } \
}

#define SET_ENQUIRY_FLAG(DeviceExtension) {\
    ULONG i; \
    UCHAR Ch1=0, Ch2=0; \
    if( (DeviceExtension->PhysicalDrive[0].Status == PDS_Online) || (DeviceExtension->PhysicalDrive[1].Status == PDS_Online) ) \
	Ch1=1; \
    if( (DeviceExtension->PhysicalDrive[0].Status == PDS_Online) || (DeviceExtension->PhysicalDrive[1].Status == PDS_Online) ) \
	Ch2=1; \
    if(Ch1 && !Ch2) \
	DeviceExtension->bChangedEnqValid = FIRST_CHANNEL; \
    else if(!Ch1 && Ch2) \
	DeviceExtension->bChangedEnqValid = SECOND_CHANNEL; \
    else if(Ch1 && Ch2) \
	DeviceExtension->bChangedEnqValid = BOTH_CHANNELS; \
}

#define WAIT_ON_ALTERNATE_DRIVE_READY(resource, BaseIoAddress2, Status, AltStatus1Byte) {\
    ULONG i; \
    for (i = 0; i < 20000; i++) { \
	Status = GET_ALTERNATE_STATUS_REGISTER_VALUE(resource, baseIoAddress2, AltStatus1Byte); \
	if ( Status != 0x50 ) { \
	    ScsiPortStallExecution(50); \
	    continue; \
    	} else { \
	    break; \
    	} \
    } \
}

#define WAIT_ON_BASE_BUSY(resource, BaseIoAddress, Status) {\
    ULONG i; \
    for (i = 0; i < 200000; i++) { \
	GET_STATUS(resource, BaseIoAddress, Status); \
	if (Status & IDE_STATUS_BUSY) { \
	    ScsiPortStallExecution(5); \
	    continue; \
	} else { \
	    break; \
	} \
    } \
}

#define WAIT_FOR_DRQ(resource, BaseIoAddress, Status) {\
    ULONG i; \
    for (i = 0; i < 200000; i++) { \
	GET_STATUS(resource, BaseIoAddress, Status); \
	if (Status & IDE_STATUS_BUSY) { \
	    ScsiPortStallExecution(100); \
	} else if (Status & IDE_STATUS_DRQ) { \
	    break; \
	} else { \
	    ScsiPortStallExecution(5); \
	} \
    } \
}

#define WAIT_FOR_ALTERNATE_DRQ(resource, baseIoAddress2, Status, AltStatus1Byte) {\
    ULONG i; \
    for (i = 0; i < 200000; i++) { \
	Status = GET_ALTERNATE_STATUS_REGISTER_VALUE(resource, baseIoAddress2, AltStatus1Byte); \
	if (Status & IDE_STATUS_BUSY) { \
	    ScsiPortStallExecution(100); \
	} else if (Status & IDE_STATUS_DRQ) { \
	    break; \
	} else { \
	    ScsiPortStallExecution(5); \
	} \
    } \
}


#define WAIT_SHORT_FOR_DRQ(resource, BaseIoAddress, Status) {\
    ULONG i; \
    for (i = 0; i < 40; i++) { \
	GET_STATUS(resource, BaseIoAddress, Status); \
	if (Status & IDE_STATUS_BUSY) { \
	    ScsiPortStallExecution(100); \
	} else if (Status & IDE_STATUS_DRQ) { \
	    break; \
	} else { \
	    ScsiPortStallExecution(5); \
	} \
    } \
}

#define ATAPI_SOFT_RESET(resource, BaseIoAddress, DeviceNumber) {\
    UCHAR statusByte; \
    ScsiPortWritePortUchar(resource,(unsigned)&BaseIoAddress->DriveSelect, (UCHAR)(((DeviceNumber & 0x1) << 4) | 0xA0)); \
    ScsiPortStallExecution(500);\
    ScsiPortWritePortUchar(resource,(unsigned)&BaseIoAddress->Command, IDE_COMMAND_ATAPI_RESET); \
    ScsiPortStallExecution(50000);    \
    ScsiPortWritePortUchar(resource,(unsigned)&BaseIoAddress->DriveSelect,(UCHAR)((DeviceNumber << 4) | 0xA0)); \
    WAIT_ON_BUSY(resource,BaseIoAddress, statusByte); \
    ScsiPortStallExecution(500);\
}

#define IDE_DC_RESET_CONTROLLER 0x04
#define IDE_DC_REENABLE_CONTROLLER 0x00
/* Vani 1007 This IDE_HARD_RESET code will be needed to be changed to include check for AltStatus1Byte,
As of now the below macro is not used anywhere */
#define IDE_HARD_RESET(resource,BaseIoAddress1, BaseIoAddress2, ucTargetId, result) {\
    UCHAR statusByte;\
    ULONG i;\
    ScsiPortWritePortUchar( resource,(unsigned)&BaseIoAddress2->AlternateStatus,IDE_DC_RESET_CONTROLLER );\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    ScsiPortWritePortUchar( resource,(unsigned)&BaseIoAddress2->AlternateStatus,IDE_DC_REENABLE_CONTROLLER);\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    ScsiPortStallExecution(1000);\
    SELECT_DEVICE(resource,BaseIoAddress1, ucTargetId);\
    for (i = 0; i < 1000; i++) {\
	statusByte = ScsiPortReadPortUchar( resource,(unsigned)&BaseIoAddress1->Command);\
	if (statusByte != IDE_STATUS_IDLE && statusByte != 0x0) {\
	    ScsiPortStallExecution(1000);\
	} else {\
	    break;\
	}\
    }\
    if (i == 1000) {\
	result = FALSE;\
    }\
    result = TRUE;\
}

#define IS_RDP(OperationCode)\
    ((OperationCode == SCSIOP_ERASE)||\
    (OperationCode == SCSIOP_LOAD_UNLOAD)||\
    (OperationCode == SCSIOP_LOCATE)||\
    (OperationCode == SCSIOP_REWIND) ||\
    (OperationCode == SCSIOP_SPACE)||\
    (OperationCode == SCSIOP_SEEK)||\
    (OperationCode == SCSIOP_WRITE_FILEMARKS))

#define SET_BM_COMMAND_REGISTER(resource,Base, Field, Value) { \
    BM_COMMAND_REGISTER buffer;\
    buffer.AsUchar = ScsiPortReadPortUchar(resource,(unsigned)&(Base->Command.AsUchar));\
    buffer.Field = Value;\
    ScsiPortWritePortUchar(resource,(unsigned)&(Base->Command.AsUchar), buffer.AsUchar); \
}

#define SET_BM_STATUS_REGISTER(resource,Base, Field, Value) {\
    BM_STATUS_REGISTER buffer;\
    buffer.AsUchar = ScsiPortReadPortUchar( resource,(unsigned)&(Base->Status.AsUchar));\
    buffer.Field = Value;\
    ScsiPortWritePortUchar(resource,(unsigned)&(Base->Status.AsUchar), buffer.AsUchar);\
}

#define CLEAR_BM_INT(resource,Base, StatusByte) \
    StatusByte = ScsiPortReadPortUchar(resource,(unsigned)&(Base->Status.AsUchar));\
    StatusByte |= 0x06; \
    ScsiPortWritePortUchar( resource,(unsigned)&(Base->Status.AsUchar), StatusByte)

#define DRIVE_PRESENT(DeviceExtension,TargetId)   \
    ((TargetId < MAX_DRIVES_PER_CONTROLLER) && (DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_DEVICE_PRESENT))

#define ENABLE_IRCD_UPDATE_ON_DRIVE(TargetId) \
    ( DeviceExtension->PhysicalDrive[TargetId].DeviceFlags |= DFLAGS_UPDATE_IRCD )

#define DISABLE_IRCD_UPDATE_ON_DRIVE(TargetId) \
    ( DeviceExtension->PhysicalDrive[TargetId].DeviceFlags &= ~ DFLAGS_UPDATE_IRCD )

#define IRCD_UPDATE_ON_DRIVE(TargetId) \
    ( DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_UPDATE_IRCD )

#define IS_IDE_DRIVE(DeviceExtension, TargetId) \
    (DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_DEVICE_PRESENT)

#define IS_ATAPI_DRIVE(TargetId) \
    ((DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_DEVICE_PRESENT) && \
    (DeviceExtension->PhysicalDrive[TargetId].DeviceFlags & DFLAGS_ATAPI_DEVICE))

#define FEED_ALL_CHANNELS(DeviceExtension) {\
    StartChannelIo(DeviceExtension, 0); \
    StartChannelIo(DeviceExtension, 1); \
}

#define CLEAR_CHANNEL( DeviceExtension , ulChannelId)    \
    DeviceExtension->Channel[ulChannelId].ActiveQueue = NULL;   \
    DeviceExtension->Channel[ulChannelId].ExpectingInterrupt = 0;

/*
* Manoj 0604 FIXME DANGER, I dont know why  here again
* we are calling StartChannelIO
*/

#if 1
#define CLEAR_AND_POST_COMMAND(DeviceExtension, ulChannel) \
    MarkChannelFree(DeviceExtension, ulChannel);
#else
#define CLEAR_AND_POST_COMMAND(DeviceExtension, ulChannel) {\
    MarkChannelFree(DeviceExtension, ulChannel); \
    StartChannelIo(DeviceExtension, ulChannel); \
}
#endif

#define ACTIVE_COMMAND_PRESENT(DeviceExtension , ucChannelId) \
    ( DeviceExtension->Channel[ucChannelId].ActiveQueue)

/* VANI: 1009 Read alternate status register from offset 0 or offset 2 based on how much resource we were able
to allocate in bus_alloc_resource, if we allocated 4 bytes read alternate status from offset 2, else if we allocated 1 byte resource, read
altrenate status from offset 1*/
#define GET_ALTERNATE_STATUS_REGISTER_VALUE(resource,baseIoAddress2, AltStatus1Byte)  \
    (AltStatus1Byte == FALSE) ? ScsiPortReadPortUchar(resource, (unsigned)&(baseIoAddress2->AlternateStatus) ): \
    ScsiPortReadPortUchar(resource, (unsigned)&(baseIoAddress2->ucReserved1) );

/*Read the Drive Head Register*/
#define GET_DRIVE_HEAD_REGISTER_VALUE(resource,baseIoAddress1)   \
    ScsiPortReadPortUchar( resource,(unsigned)&(baseIoAddress1->DriveSelect))
/*Read the error Register*/
#define GET_ERROR_REGISTER_VALUE(resource,baseIoAddress1)    \
    ScsiPortReadPortUchar( resource,(unsigned)&(baseIoAddress1->SectorCount))

/*IDE Error Sense Macro*/
#define  SET_IDE_ERROR_SENSE(S,D,E)       ((0x00000000) | ((S) << 16) | ((D) << 8) | (E))

#define IS_CHANNEL_BUSY(DeviceExtension, ulChannelId)   (DeviceExtension->Channel[ucChannelId].pstActiveCommand)

#define DRIVE_HAS_COMMANDS(DeviceExtension, ucTargetId) DeviceExtension->PhysicalDrive[ucTargetId].stScsiCmdQueue.pHead

#define DEBUG_INQUIRY_LEVEL		(0x1 << 31)

#define CHANNEL_IS_BUSY( DeviceExtension , ulChannelId)   \
    ( DeviceExtension->Channel[ulChannelId].ActiveQueue)

#define IS_DRIVE_READY(DeviceExtension, TargetId) \
    ((TargetId < MAX_DRIVES_PER_CONTROLLER) && (DeviceExtension->PhysicalDrive[TargetId].Status < PDS_Failed))

#define IS_DRIVE_ONLINE(DeviceExtension, TargetId) \
    ((TargetId < MAX_DRIVES_PER_CONTROLLER) && (DeviceExtension->PhysicalDrive[TargetId].Status == PDS_Online))

#define IS_DRIVE_REBUILDING(DeviceExtension, TargetId) \
    ((TargetId < MAX_DRIVES_PER_CONTROLLER) && (DeviceExtension->PhysicalDrive[TargetId].Status == PDS_Rebuilding))

#define IS_DRIVE_FAILED(TargetId) \
    ((TargetId == INVALID_DRIVE_ID) || (DeviceExtension->PhysicalDrive[TargetId].Status == PDS_Failed))


/*Absolute sure that there is a logical drive present for this id
#define GET_LD_INFO(ld)	(((ULONG)ld < DeviceExtension->ulLogicalDriveCount) ? &DeviceExtension->LogicalDrive[ld] : NULL)
*/
#define GET_LD_INFO(ld)	(((ULONG)ld < DeviceExtension->ulLogicalDriveCount) ? &DeviceExtension->LogicalDrive[ DeviceExtension->BootLogDrvMapping[ld] ] : NULL)

#define SET_IDENTIFY_STAMP(Ld, PdIndex) ((((USHORT)Ld<<8) & 0xff00)|(PdIndex & 0xff))

#define UPDATE_IDENTIFY_STAMP(Buffer, IdentityStamp) \
    Buffer[0x1f0] = (UCHAR)(IdentityStamp & 0xff); \
    Buffer[0x1f1] = (UCHAR)(IdentityStamp >> 8)


#define IS_OTHER_ERROR(ucExtendedErrorValue)         ((!(ucExtendedErrorValue & (ICRC_ERR|ECC_ERR))) && (ucExtendedErrorValue & ABRT_ERR ))

/* +++ edm March 27, 2002
Dell is wanting us to program down the drives if we get 3 sequentail errors on a physical drive
here is the macros for handling this
*/
#define CLEAR_SEQ_ERROR_FLAG( TargetId ) \
    DeviceExtension->PhysicalDrive[TargetId].SequentialErrorCount = 0;

#define INC_SEQ_ERROR_FLAG( TargetId ) \
    DeviceExtension->PhysicalDrive[TargetId].SequentialErrorCount ++; \
    if ( DeviceExtension->PhysicalDrive[TargetId].SequentialErrorCount >= 3 ) {\
	SetTransferMode ( DeviceExtension, TargetId , ( DeviceExtension->PhysicalDrive[TargetId].TransferMode - 1 )  ) ; \
	CLEAR_SEQ_ERROR_FLAG( TargetId ); \
    }
/* --- edm March 27, 2002*/


#define COPY_BACK() \
    if (offset > DeviceExtension->iProcIdx) { \
	*eof = TRUE; \
	DeviceExtension->iProcIdx = 0; \
	DeviceExtension->achProcBuf[0] = 0; \
	return 0; \
    } \
    if ((iCount + offset) > DeviceExtension->iProcIdx) { \
	iCount = DeviceExtension->iProcIdx - offset; \
	*eof = TRUE; \
    } \
    memcpy(page, &DeviceExtension->achProcBuf[offset], iCount); \
    DeviceExtension->iProcIdx = 0; \
    DeviceExtension->achProcBuf[0] = 0;

#define	POLL_FOR_INTERRUPT(resource) {\
    ULONG i; \
    ScsiPortStallExecution(1000); \
    for (i = 0; i < 5000; i++) /* for a whole second delay*/ {\
	ucStatus = ScsiPortReadPortUchar(resource,(unsigned)&(DeviceExtension->Channel[0].BaseBmAddress->Reserved)); \
	if (ucStatus & ucInterruptValue) {\
	    break; \
	} else {\
	    ScsiPortStallExecution(1000); \
	    ScsiPortStallExecution(1000); \
	    continue; \
	}\
    }\
}

#define	POLL_FOR_DMA_INTERRUPT(resource,baseBm) {\
    ULONG i; \
    ScsiPortStallExecution(1000); \
    for (i = 0; i < 5000; i++) /* for a whole second delay*/ {\
	ucStatus = ScsiPortReadPortUchar(resource,(unsigned)&(baseBm->Status.AsUchar)); \
	if (((BM_STATUS_REGISTER) ucStatus) .f.InterruptStatus) {\
	/*WAIT_ON_ALTERNATE_STATUS_BUSY(BaseIoAddress2, ucStatus); */ \
	    break; \
	} else {\
	    ScsiPortStallExecution(1000); \
	    ScsiPortStallExecution(1000); \
	    continue; \
	}\
    }\
}

#define	POLL_FOR_INTERRUPT_68X() {\
    ULONG i; \
    ScsiPortStallExecution(1000); \
    for (i = 0; i < 5000; i++) /* for a whole second delay*/ {\
	if (!(ulTargetId >> 1)) /*secondary*/ {\
	    pci_read_config_byte_Rlsi(DeviceExtension, (IDE0_TFT_CONF_STAT_REG680 + 1), &ucStatus); \
	    if (ucStatus & PRIMARY_CHAN_INTERRUPT_680) {\
		break; \
	    } else {\
		ScsiPortStallExecution(1000); \
		ScsiPortStallExecution(1000); \
		continue; \
	    }\
	} else/*Primary channel*/ {\
	    pci_read_config_byte_Rlsi(DeviceExtension, (IDE1_TFT_CONF_STAT_REG680 + 1), &ucStatus); \
	    if (ucStatus & SECONDARY_CHAN_INTERRUPT_680) {\
		break; \
	    } else {\
		ScsiPortStallExecution(1000); \
		ScsiPortStallExecution(1000); \
		continue; \
	    }\
	}\
    }\
}

#define	POLL_FOR_INTERRUPT_CSB5(resource,baseBm) {\
    ULONG i; \
    ScsiPortStallExecution(1000); \
    for (i = 0; i < 5000; i++) /* for a whole second delay*/ {\
	if (!(ulTargetId >> 1)) /*secondary*/ {\
	    ucStatus = ScsiPortReadPortUchar(resource, (unsigned)(baseBm+0x0A)); \
	    if (ucStatus & CHANNEL_INTERRUPT_SERVERWORKS) {\
		break; \
	    } else {\
		ScsiPortStallExecution(1000); \
		ScsiPortStallExecution(1000); \
		continue; \
	    }\
	} else/*Primary channel*/ {\
	    ucStatus = ScsiPortReadPortUchar(resource,(unsigned)(baseBm+ 0x02)); \
	    if (ucStatus & CHANNEL_INTERRUPT_SERVERWORKS) {\
		break; \
	    } else {\
		ScsiPortStallExecution(1000); \
		ScsiPortStallExecution(1000); \
		continue; \
	    }\
	}\
    }\
}

#endif /*__MACROS_H__*/
/*vi set ts=8:*/
