/*
 * Copyright (c) 2002 Eric Moore
 * Copyright (c) 2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/mide_char.c,v emoore Exp$
 */


#define	MIN(a, b) (((a) < (b)) ? (a) :(b))

#define	CHAR_IOCTL_TIMEOUT  20	/* 20Sec */
#define	CHAR_IOCTL_FINISHED 0xFF

static int
mide_open(dev_t dev, int oflags, int devtype, d_thread_t * td)
{
	int err = 0;
	return (err);
}

static int
mide_close(dev_t dev, int fflag, int devtype, d_thread_t * td)
{
	int err = 0;
	return (err);
}

static int
mide_ioctl(dev_t dev, u_long cmd, caddr_t arg, int flag, d_thread_t * td)
{
	pchar_megaide_ioctl p_ioctl = (pchar_megaide_ioctl) arg;
	int error = 0;

	if (cmd == MEGAIDE_DEV_IOCTL) {
		if (memcmp_Rlsi(p_ioctl->acSignature, "MEGANIT", 7) != 0) {
			AR_DEBUG(AR_DEBUG_CHAR_IOCTL, (" wrong signature\n"));
			return (EINVAL);
		}
		if ((p_ioctl->iAdapNumber > giNumberOfCtrlrs) ||
		    (!p_ioctl->iAdapNumber)) {
			AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
			    (" Wrong adapter number\n"));
			return (EINVAL);
		}
		error = HandleCharIOCTLCall(cmd, arg,
		    p_ioctl->inlen, p_ioctl->outlen, p_ioctl->iAdapNumber);

	}
	return (error);
}

static void
megaidedev_ioctl_over(void *pstScsiCmd)
{
	AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
	    (" megaide: megaidedev_ioctl_over\n"));
	wakeup(pstScsiCmd);
}				/* megaidedev_ioctl_over() */

int
HandleCharIOCTLCall(unsigned int cmd, caddr_t arg, u_long inlen,
    u_long outlen, int adp_no)
{
	int i;
	int iRetVal = EFAULT;
	struct ccb_hdr *ccb_h;
	struct ccb_scsiio *pstScsiCmd;
	pchar_ioctl_data pIdata;
	pchar_megaide_ioctl p_ioctl = (pchar_megaide_ioctl) arg;
	PHW_DEVICE_EXTENSION DeviceExtension = NULL;
	int data_size = 0;
	PFAST_IDE_SCB_EXTENSION ScbExtension;
	UCHAR IoctlRet;
	int s;

	s = splcam();

	data_size = (inlen > outlen) ? inlen : outlen;

	DeviceExtension = gastCardInfo[adp_no - 1].DeviceExtension;

	if (DeviceExtension == NULL) {
		printk_Rlsi("megaide: Ooops Device Exension is Null, \
		adp_no = %d\n",
		    adp_no);
		splx(s);
		return (EINVAL);
	}
	pstScsiCmd = (struct ccb_scsiio *) malloc(sizeof(union ccb),
	    M_MIDE, M_WAITOK | M_ZERO);
	if (pstScsiCmd == NULL) {
		AR_DEBUG(AR_DEBUG_CHAR_IOCTL, ("megaide: \
			SCSI command Memory allocation Failed!!\n"));
		splx(s);
		return (ENOMEM);
	}
	bzero(pstScsiCmd, sizeof(union ccb));

	ScbExtension = scsi_malloc(DeviceExtension,
	    (void *)pstScsiCmd, TRUE);
	if (ScbExtension == NULL) {
		printk_Rlsi("megaide: unable to \
			allocate ScbExtension in mide_char.c\n");
		free(pstScsiCmd, M_MIDE);
		splx(s);
		return (ENOMEM);
	}
	pIdata = (pchar_ioctl_data) malloc(
	    (data_size + sizeof(ioctl_header)),
	    M_MIDE, M_WAITOK | M_ZERO);
	if (pIdata == NULL) {
		AR_DEBUG(AR_DEBUG_CHAR_IOCTL, ("megaide: \
			IOCTL Memory allocation Failed!!\n"));
		free(pstScsiCmd, M_MIDE);
		splx(s);
		return (ENOMEM);
	}
	bzero(pIdata, data_size);


	if (copyin(p_ioctl->wdata, pIdata,
		(inlen + sizeof(ioctl_header))) != 0) {
		printk_Rlsi("megaide: Copy From user Failed!!\n");
		splx(s);
		return (EFAULT);
	}
	ccb_h = (struct ccb_hdr *) pstScsiCmd;
	ccb_h->spriv_ptr1 = megaidedev_ioctl_over;

	ScbExtension->request_buffer = (VOID *) pIdata;
	ScbExtension->pstScsiCmd = (void *)pstScsiCmd;
	pstScsiCmd->data_ptr = (VOID *) pIdata;
	pstScsiCmd->dxfer_len = (data_size + sizeof(ioctl_header));

	IoctlRet = HandleIOCTLCall(DeviceExtension, ScbExtension);

	switch (IoctlRet) {
	case IOCTL_OK:
		{
			AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
			    ("megaide:IOCTL_OK\n"));
			megaide_complete_command(DeviceExtension,
			    ScbExtension, (MIDE_STATUS_OK << 16));
			iRetVal = 0;
			break;
		}

	case IOCTL_ERROR:
		{
			AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
			    ("megaide:IOCTL error\n"));
			megaide_complete_command(DeviceExtension,
			    ScbExtension,
			    (MIDE_STATUS_ERROR << 16));
			iRetVal = 0;
			break;
		}

	case IOCTL_QUEUE:
		{
			AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
			    ("megaide:IOCTL_QUEUE\n"));

			for (i = 0; i < (CHAR_IOCTL_TIMEOUT * hz); i++) {
				tsleep(pstScsiCmd, PRIBIO | PCATCH,
				    "midewcmd", 5 * hz);
				if (ccb_h->status == CAM_REQ_CMP) {
					break;
				}
			}

			if (ccb_h->status != CAM_REQ_CMP) {
				AR_DEBUG(AR_DEBUG_CHAR_IOCTL, ("mide: \
					Cmd timed out, Need To abort\n"));
				abort_passthu_command(DeviceExtension,
				    pstScsiCmd);
			}
			iRetVal = 0;

			break;
		}

	default:
		break;

	}			/* switch */

	if (!iRetVal) {
		iRetVal = copyout(pIdata, p_ioctl->wdata,
		    (outlen + sizeof(ioctl_header)));
	}
	free(pstScsiCmd, M_MIDE);
	free(pIdata, M_MIDE);
	splx(s);

	return (iRetVal);

}

UCHAR
abort_passthu_command(PHW_DEVICE_EXTENSION DeviceExtension,
    void *pstScsiCmd)
{
	PFAST_IDE_SCB_EXTENSION ScbExtension = NULL;
	struct ccb_hdr *ccb_h;

	ccb_h = (struct ccb_hdr *) pstScsiCmd;

	AR_DEBUG(AR_DEBUG_CHAR_IOCTL,
	    (" megaide: abort_passthu_command\n"));

	ScbExtension = (PFAST_IDE_SCB_EXTENSION) ccb_h->spriv_ptr0;
	megaide_abort_ioctl(DeviceExtension, ScbExtension);
	return (0);
}				/* abort_passthu_command() */

/*
 *-------------------------------------------------------------------
 * Description
 *  	This funtion returns the integer equvalent value of the array
 * Arguments
 * 	pcArray-> Array holds the integer number
 *
 * Return Value
 * 	Signed integer equivalent value
 *-------------------------------------------------------------------
 */

long
array_to_int(const char *pcArray)
{
	int iCurrent;		/* current char */
	long ulTotal;		/* current total */
	int iSign;		/* if '-', then negative, otherwise positive */

	/* skip whitespace */
	while (isspace((int)(unsigned char)*pcArray)) {
		++pcArray;
	}


	iCurrent = (int)(unsigned char)*pcArray++;
	iSign = iCurrent;
	if (iCurrent == '-' || iCurrent == '+') {
		/* skip sign */
		iCurrent = (int)(unsigned char)*pcArray++;
	}
	ulTotal = 0;

	while (isdigit(iCurrent)) {
		/* accumulate digit */
		ulTotal = 10 * ulTotal + (iCurrent - '0');
		/* get next char */
		iCurrent = (int)(unsigned char)*pcArray++;
	}

	if (iSign == '-') {
		return (-ulTotal);
	} else {
		return (ulTotal);
	}

}				/* array_to_int() */
