/*
 * Copyright (c) 2001, 2002 Eric Moore
 * Copyright (c) 2001, 2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/mide_kernel_lib.c,v emoore Exp$
 */


/* megaide_kernel_lib.c */


#if defined __alpha__
#define	vtobus(va) 	alpha_XXX_dmamap((vm_offset_t)va)
#else
#define	vtobus(va)		vtophys(va)
#endif

/* library calls */
void
ScsiPortStallExecution(unsigned long usecs)
{
	DELAY(usecs);
}

unsigned char
ScsiPortReadPortUchar(void *resource, unsigned short port)
{
	return bus_space_read_1(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start);
}

unsigned short
ScsiPortReadPortUshort(void *resource, unsigned short port)
{
	return bus_space_read_2(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start);
}

void
ScsiPortWritePortUchar(void *resource,
    unsigned short port, unsigned char value)
{
	bus_space_write_1(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start,
	    value);
}

void
ScsiPortWritePortUlong(void *resource,
    unsigned short port, unsigned int value)
{
	bus_space_write_4(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start,
	    value);
}

unsigned long
ScsiPortReadPortUlong(void *resource,
    unsigned short port)
{
	return bus_space_read_4(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start);
}
void
ScsiPortReadPortBufferUshort(void *resource,
    unsigned short port, void *addr, unsigned long count)
{
	return bus_space_read_multi_2(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start,
	    addr, count);
}

void
ScsiPortWritePortBufferUshort(void *resource,
    unsigned short port, const void *addr, unsigned long count)
{
	bus_space_write_multi_2(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    port - ((struct resource *) resource)->r_start,
	    addr, count);
}

unsigned long
ReadMemIoRegUlong(void *resource,
    unsigned long offset)
{
	return bus_space_read_4(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    offset);
}
void
WriteMemIoRegUlong(void *resource,
     unsigned long offset, unsigned long value)
{
	bus_space_write_4(
	    rman_get_bustag((struct resource *) resource),
	    rman_get_bushandle((struct resource *) resource),
	    offset, value);
}

unsigned char
inb_Rlsi(unsigned short port)
{
	return inb(port);
}

void
outb_Rlsi(unsigned short port, unsigned char value)
{
	outb(port, value);
}

int
memcmp_Rlsi(const void *cs, const void *ct, size_t count)
{
	return memcmp(cs, ct, count);
}

void *
memcpy_Rlsi(void *dest, const void *src, size_t count)
{
	return memcpy(dest, src, count);
}

void *
memmove_Rlsi(void *dest, const void *src, size_t count)
{
	return memcpy(dest, src, count);
}

void *
memset_Rlsi(void *s, int c, size_t count)
{
	return memset(s, c, count);
}

char *
strncpy_Rlsi(char *dest, const char *src, size_t count)
{
	return strncpy(dest, src, count);
}

char *
strcpy_Rlsi(char *dest, const char *src)
{
	return strcpy(dest, src);
}

int
strcmp_Rlsi(const char *s1, const char *s2)
{
	return strcmp(s1, s2);
}

int
strncmp_Rlsi(const char *cs, const char *ct, size_t count)
{
	return strncmp(cs, ct, count);
}


int
pci_read_config_word_Rlsi(
    PHW_DEVICE_EXTENSION DeviceExtension,
    int where, unsigned short *val)
{
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	*val = pci_read_config(gastCardInfo[ucControllerId].pstDev,
	    where, 2);
	return 0;
}

int
pci_read_config_byte_Rlsi(
    PHW_DEVICE_EXTENSION DeviceExtension,
    int where, unsigned char *val)
{
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	*val = pci_read_config(gastCardInfo[ucControllerId].pstDev,
	    where, 1);
	return 0;
}

int
pci_write_config_word_Rlsi(
    PHW_DEVICE_EXTENSION DeviceExtension,
    int where, unsigned short val)
{
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	pci_write_config(gastCardInfo[ucControllerId].pstDev,
	    where, val, 2);
	return 0;
}

int
pci_write_config_byte_Rlsi(
    PHW_DEVICE_EXTENSION DeviceExtension,
    int where, unsigned char val)
{
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	pci_write_config(gastCardInfo[ucControllerId].pstDev,
	    where, val, 1);
	return 0;
}

void
spin_lock_Rlsi(PHW_DEVICE_EXTENSION DeviceExtension)
{
/*
 *	FIXME
 *	FIXED: VANI 0827
 */
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
#ifdef SMPNG
	mtx_lock(&gastCardInfo[ucControllerId].midemtx);
#endif
}

void
spin_unlock_Rlsi(PHW_DEVICE_EXTENSION DeviceExtension)
{
/*
 *	FIXME
 *	FIXED: VANI 0827
 */

	UCHAR ucControllerId;
	ucControllerId = DeviceExtension->ucControllerId;
#ifdef SMPNG
	mtx_unlock(&gastCardInfo[ucControllerId].midemtx);
#endif
}

void
wake_up_Rlsi(PHW_DEVICE_EXTENSION DeviceExtension)
{
/*
 *	FIXME
 *	FIXED: VANI 0827
 */
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	wakeup(&gastCardInfo[ucControllerId].internal_wait);
}

void
sleep_on_timeout_Rlsi(PHW_DEVICE_EXTENSION DeviceExtension)
{
/*
 *	FIXME
 *	FIXED: VANI 0827
 */
	UCHAR ucControllerId;

	ucControllerId = DeviceExtension->ucControllerId;
	tsleep(&gastCardInfo[ucControllerId].internal_wait,
	    PRIBIO | PCATCH, "midewcmd", 5 * hz);
}

int
printk_Rlsi(const char *fmt,...)
{
	device_t pstDev;
	va_list ap;
	char buf[256];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	/*
	 * FIXME - we are assuming we are picking up one controller here
	 */
	pstDev = gastCardInfo[0].pstDev;
	if (!pstDev) {
		printf(buf);
		return 0;
	}
	return device_printf(pstDev, buf);
}

u_long
CURRENT_TIME_RLSI(void)
{
	struct timeval now;
	microtime(&now);
	return (now.tv_sec);	/* returns seconds */
}
/* FIXME: Vani 1118: if yellowstone support is needed then following code needs to be worked on
void *
request_region_Rlsi( USHORT port, ULONG count )
{
	device_t dev;
        int rid;

}
void
release_region_Rlsi( USHORT port, ULONG count )
{
	device_t dev;
        int rid;

	release_region ( port , count );

 }
*/
