/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/mide_pci.c,v emoore Exp$
 */



#if __FreeBSD_version >= 500005
#include <sys/taskqueue.h>
#endif
#

#include <machine/stdarg.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/ctype.h>


#include <sys/bus.h>
#include <sys/bus_private.h>
#include <sys/conf.h>
#include <sys/devicestat.h>
#include <sys/disk.h>
#include <sys/stat.h>
#include <sys/malloc.h>

#include <sys/file.h>

#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>

#include <machine/bus_memio.h>
#include <machine/bus_pio.h>
#include <machine/bus.h>

#include <vm/vm.h>
#include <vm/pmap.h>
#include <machine/md_var.h>
#include <machine/pc/bios.h>

#include <machine/resource.h>
#include <machine/clock.h>
#include <sys/rman.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#include <cam/cam.h>
#include <cam/cam_ccb.h>
#include <cam/cam_periph.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt.h>
#include <cam/cam_xpt_sim.h>
#include <cam/cam_debug.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>

#include <contrib/dev/mide/inc/mide_datatype.h>
#include <contrib/dev/mide/inc/mide_debug.h>
#include <contrib/dev/mide/inc/mide_adapter.h>
#include <contrib/dev/mide/inc/mide_ata.h>
#include <contrib/dev/mide/util/mp_hd_ioctls.h>
#include    <contrib/dev/mide/util/bp_ioctl.h>

#include <contrib/dev/mide/inc/mide_devext.h>
#include <contrib/dev/mide/inc/mide_pci.h>
#include <contrib/dev/mide/inc/mide_macros.h>
#include <contrib/dev/mide/inc/mide_scsi.h>
#include <contrib/dev/mide/inc/mide_compat.h>
#include <contrib/dev/mide/inc/mide_char.h>


#include <contrib/dev/mide/inc/mide_kernel_lib.h>


#ifndef	KMALLOC_MAXSIZE
/* Maximum allowed request size */
#define	KMALLOC_MAXSIZE		131072UL
#endif



/* Global Functions */
static MALLOC_DEFINE(M_MIDE, "MIDE driver", "MIDE RAID driver");

CARD_INFO gastCardInfo[MAX_CONTROLLERS] = {};
MP_RAID_OPTIONS_EX driverOptions;
#ifdef BP_SUPPORT
BP_OPTIONS bpOptions;
#endif
static int giNumberOfCtrlrs = 0;
static int mide_pci_identify_called = FALSE;
u_long debug_level = 0;


CARD_SPECIFIC_FUNCTION gaCardSpecificFunctions[NUM_SUPPORTED_CONTROLLERS] =
{
	{Found64xInterrupt, Disable64xInterrupts, Enable64xInterrupts,
		Is64x80PinCable, ProgramDriveFor64xChip,
		IsPostCommandToDriveSuccess64x},
	{Found64xInterrupt, Disable64xInterrupts, Enable64xInterrupts,
		Is64x80PinCable, ProgramDriveFor64xChip,
		IsPostCommandToDriveSuccess64x},
	{Found68xInterrupt, Disable68xInterrupts, Enable68xInterrupts,
		Is68x80PinCable, ProgramDriveFor68xChip,
		IsPostCommandToDriveSuccess68x},
	{FoundCSB5Interrupt, DisableCSB5Interrupts, EnableCSB5Interrupts,
		IsCSB580PinCable, ProgramDriveForCSB5Chip,
		IsPostCommandToDriveSuccessCSB5},
	{Found68xInterrupt, DisableSata3112Interrupts, EnableSata3112Interrupts,
		Is68x80PinCable, ProgramDriveFor68xChip,
		IsPostCommandToDriveSuccess68x},
	{FoundCSB5Interrupt, DisableCSB5Interrupts, EnableCSB5Interrupts,
		IsCSB580PinCable, ProgramDriveForCSB5Chip,
		IsPostCommandToDriveSuccessCSB5}
};

#ifdef MIDE_COMPILE_CLOSED_SOURCE
#include    </sys/contrib/dev/mide/lib/megaide_lib.c>
#endif

#include    <contrib/dev/mide/mide_kernel_lib.c>
#include    <contrib/dev/mide/mide_scsi.c>
#include    <contrib/dev/mide/mide_char.c>


static void mide_pci_identify(driver_t * driver, device_t parent);
static int mide_pci_probe(device_t dev);
static int mide_pci_attach(device_t dev);
static int mide_pci_detach(device_t dev);
static int mide_pci_shutdown(device_t dev);
static int mide_pci_suspend(device_t dev);
static int mide_pci_resume(device_t dev);


static device_method_t mide_methods[] = {
	/* Device interface */
	DEVMETHOD(device_identify, mide_pci_identify),
	DEVMETHOD(device_probe, mide_pci_probe),
	DEVMETHOD(device_attach, mide_pci_attach),
	DEVMETHOD(device_detach, mide_pci_detach),
	DEVMETHOD(device_shutdown, mide_pci_shutdown),
	DEVMETHOD(device_suspend, mide_pci_suspend),
	DEVMETHOD(device_resume, mide_pci_resume),
	DEVMETHOD(bus_print_child, bus_generic_print_child),
	DEVMETHOD(bus_driver_added, bus_generic_driver_added),
	{0, 0}
};

static driver_t mide_pci_driver = {
	"mide",
	mide_methods,
	sizeof(struct _HW_DEVICE_EXTENSION)
};

devclass_t mide_devclass;
DRIVER_MODULE(mide, pci, mide_pci_driver, mide_devclass, 0, 0);

static struct {
	int vendor;
	int device;
	int SubVendor;
	int SubDevice;
	int flag;
#define	PROBE_SIGNATURE	(1<<0)
} mide_device_ids[] = {

	{
		PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_648, 0, 0, 0
	},
	{
		PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_649, 0, 0, 0
	},
	{
		PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_680, 0, 0, 0
	},
	{
		PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_3112, 0, 0, 0
	},
	{
		PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SERVERWORKS_CSB6IDE, 0, 0, 0
	},
	{
		0, 0, 0
	}
};



/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
mide_pci_identify(driver_t * driver, device_t parent)
{
	BUS_ADD_CHILD(parent, 0, "mide", -1);
	/* printf("mide_pci_identify\n"); */
	RemoveBuildInIDESupport(driver, parent);
	mide_pci_identify_called = TRUE;
}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
RemoveBuildInIDESupport(driver_t * driver, device_t parent)
{
	device_t *children;
	int nchildren, i;
	short sPciVendor;
	int unregister_ide = 0;

	if (device_get_children(parent, &children, &nchildren))
		return;

	for (i = 0; i < nchildren; i++) {
		sPciVendor = pci_get_vendor(children[i]);

		if ((sPciVendor == PCI_VENDOR_ID_CMD) ||
		    (sPciVendor == PCI_VENDOR_ID_SERVERWORKS) ) {

			 AR_DEBUG(AR_DEBUG_FINDDEV, ("mide : VendorID %x  Support \n" , sPciVendor));

			unregister_ide = CheckSubSysID(children[i]);
			if (unregister_ide) {

				  AR_DEBUG(AR_DEBUG_FINDDEV, ("mide : UnRegistered Builtin IDE Support \n" ));
				ide_unregister(driver, children[i]);
			}
		}
	}
}

int
CheckSubSysID(device_t dev)
{
	short sPciDev;
	short sPciSubVendor;
	short sPciSubDevice;
	u_char class_dev;
	int unregister = 0;

	sPciDev = pci_get_device(dev);
	sPciSubVendor = pci_read_config(dev, PCIR_SUBVEND_0, 2);
	sPciSubDevice = pci_read_config(dev, PCIR_SUBDEV_0, 2);
	class_dev = pci_read_config(dev, PCIR_CLASS, 1);

	AR_DEBUG(AR_DEBUG_FINDDEV, ("SubDev %x, SubVen %x \n", sPciSubDevice, sPciSubVendor));

	switch (sPciDev) {
	case PCI_DEVICE_ID_CMD_648:
	case PCI_DEVICE_ID_CMD_649:
		unregister = 1;
		break;
	case PCI_DEVICE_ID_SERVERWORKS_CSB6IDE:
		if (sPciSubVendor == PCI_VENDOR_ID_LSI_LOGIC) {
			unregister = 1;
		}
		break;
	case PCI_DEVICE_ID_CMD_3112:
		if ((sPciSubVendor == PCI_SUB_VENDOR_ID_CMD_3112 ) &&
			(sPciSubDevice == PCI_SUB_DEVICE_ID_CMD_3112)) {
			unregister = 1;
		}
		break;
	case PCI_DEVICE_ID_CMD_680:
		/*
		 * FIXME SUBSYS ID CHECK NEEDS TO BE DONE HERE FIXED: VANI
		 * 0827
		 */
		if (sPciSubVendor == PCI_VENDOR_ID_DELL) {
			if ((sPciSubDevice == PCI_DEVICE_ID_YELLOWSTONE) ||
			    (sPciSubDevice == PCI_DEVICE_ID_CMD_680) ||
			    (sPciSubDevice == PCI_DEVICE_ID_CMD_3112) ||
			    (sPciSubDevice == PCI_DEVICE_ID_DELL_CERC_ATA_AS)) {
				unregister = 1;
			}
		} else if (sPciSubVendor == PCI_VENDOR_ID_CMD) {
			if (sPciSubDevice == PCI_DEVICE_ID_CMD_680) {
				unregister = 1;
			}
		} else if (sPciSubVendor == PCI_VENDOR_ID_LSI_LOGIC) {
			if (sPciSubDevice == PCI_DEVICE_ID_LSI_524) {
				unregister = 1;
			}
		}
		break;
	default:
		break;
	}

	return unregister;
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
ide_unregister(driver_t * driver, device_t pDEV)
{

	if (!pDEV->nameunit) {
		device_set_devclass(pDEV, driver->name);
		return;
	}
	if (strcmp(device_get_name(pDEV), "atapci") == 0) {
		bus_generic_detach(pDEV);
		pDEV->unit = -1;
		pDEV->devclass = NULL;
		pDEV->nameunit = NULL;
#ifdef DEVICE_SYSCTLS
		device_unregister_oids(pDEV);
#endif
		pDEV->state = DS_NOTPRESENT;
		device_set_driver(pDEV, NULL);
		device_set_devclass(pDEV, driver->name);
		return;
	}
}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_probe(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);
	int i;
	short sPciSubVendor;
	short sPciSubDevice;
	int attach = 0;

	if (!dev || !(DeviceExtension))
		return ENXIO;

	/* we need to handle the case of loading driver at boot time */
	if (mide_pci_identify_called == FALSE) {
		RemoveBuildInIDESupport(device_get_driver(dev), device_get_parent(dev));
	}
	sPciSubVendor = pci_get_vendor(dev);
	sPciSubDevice = pci_get_device(dev);

	/*
	 * printf("mide0: mide_pci_probe: VendorID=%x,DeviceID=%x\n",
	 * pci_get_vendor(dev),pci_get_device(dev));
	 */
	for (i = 0; mide_device_ids[i].vendor != 0; i++) {
		if ((pci_get_vendor(dev) == mide_device_ids[i].vendor) &&
		    (pci_get_device(dev) == mide_device_ids[i].device)) {

			/*
			 * FIXME SUBSYS ID CHECK NEEDS TO BE DONE HERE FIXED:
			 * VANI 0827 */
   			AR_DEBUG(AR_DEBUG_FINDDEV, ("mide: probe: Device ID %x, Vendor ID %x\n",
				pci_get_device(dev), pci_get_vendor(dev)));

			attach = CheckSubSysID(dev);
			if (!attach) {
				continue;
			} else {
				device_set_desc(dev, "LSILogic MegaRAID IDE");
				/*
				 *	allow room to be
				 * overridden by returning 0
				 */
				return (0);
			}
		}
	}

	return (ENXIO);
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_attach(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);
	int error;

	if (!dev || !(DeviceExtension))
		return (ENXIO);

	error = ENXIO;

	/* PLEASE don't use printk's till after this point */
	gastCardInfo[giNumberOfCtrlrs].pstDev = dev;

	if (bootverbose)
		printk_Rlsi(MEGAIDE_VERSION);
#ifdef SMPNG
	mtx_init(gastCardInfo[giNumberOfCtrlrs].midemtx, device_get_nameunit(dev), MTX_DEF | MTX_TOPHALF);
#endif

	if (megaide_find_card(DeviceExtension, dev) != ENXIO)
		error = 0;

	if (!error) {
		/*
		 * FIXME: change Later, for making only one instance of dev
		 * node for all adapters Vani 0826
		 * Vani: Commented as it causes conflict, will create device node using shell script
		 */

		    gastCardInfo[giNumberOfCtrlrs].mide_dev = make_dev(
		    &mide_cdevsw,
		    device_get_unit(gastCardInfo[giNumberOfCtrlrs].pstDev),
		    UID_ROOT, GID_OPERATOR,
		    S_IRUSR|S_IWUSR, "mide");
                /*
		 * if (gastCardInfo[giNumberOfCtrlrs].mide_dev)
		 * printk_Rlsi("Created mided for priv_node \n");
		 */
		mide_periodic((void *)DeviceExtension);
		DeviceExtension->FreezeCnt = 0;
		giNumberOfCtrlrs++;
	}
	return (error);

}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

int
megaide_find_card(
    PHW_DEVICE_EXTENSION DeviceExtension,
    device_t dev)
{
	int ControllerNumber;
	u_char i, j;
	int error;

	/* debug_called(1); */

	/* make sure this is RAID STORAGE CLASS */
	if (pci_get_subclass(dev) != PCIS_STORAGE_RAID)
		return (ENXIO);

	memset_Rlsi(DeviceExtension, 0, sizeof(HW_DEVICE_EXTENSION));
	ControllerNumber = giNumberOfCtrlrs;

	for (j = 0; j < MAX_LOGICAL_DRIVES; j++) {
		DeviceExtension->LogicalDrive[j].ucControllerId = ControllerNumber;
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
		DeviceExtension->PhysicalDrive[i].ucControllerId = ControllerNumber;
	}

	/* Adjust SGL's buffer pointer for the channels */
	AssignSglPtrsForChannels(DeviceExtension);


	if (AssignPtrsForDevExtension(DeviceExtension) == FALSE) {
		CleanUpAdapter(DeviceExtension);
		return (ENXIO);
	}
	/* initializing device extension */
	gastCardInfo[ControllerNumber].DeviceExtension = DeviceExtension;
	DeviceExtension->ucControllerId = ControllerNumber;
	DeviceExtension->ulVendorId = pci_get_vendor(dev);
	DeviceExtension->ulDeviceId = pci_get_device(dev);
	DeviceExtension->usSubsysVendorId = pci_get_vendor(dev);
	DeviceExtension->usSubsysDeviceId = pci_get_device(dev);
	DeviceExtension->ucPCIFun = pci_get_function(dev);
	DeviceExtension->usSlot = pci_get_slot(dev);
	DeviceExtension->ucPCIBus = pci_get_bus(dev);
	DeviceExtension->ucPCIDev = (DeviceExtension->ucPCIFun >> 3);
	bus_get_resource(dev, SYS_RES_IRQ, 0, &DeviceExtension->ulIntLine, NULL);

	driverOptions.ulRebuildRate = DEFAULT_RATE;
	driverOptions.ulIdentical = TRUE;
	driverOptions.ulAutoRebuild = FALSE;
	driverOptions.ulCheckArray = 0;
	driverOptions.ulStopRebuild = 0;
	driverOptions.ulErrorThreshold = 0;
	driverOptions.ulFailDrive = 0;

#ifdef BP_SUPPORT
	/*backplane discovery*/
	idebp_discovery(DeviceExtension);
#endif


	switch (DeviceExtension->ulDeviceId) {
	case PCI_DEVICE_ID_CMD_648:
		DeviceExtension->ControllerSpeed = Udma66;
		DeviceExtension->ucCardType = CMD_648;
		break;

	case PCI_DEVICE_ID_CMD_649:
		DeviceExtension->ControllerSpeed = Udma100;
		DeviceExtension->ucCardType = CMD_649;
		break;

	case PCI_DEVICE_ID_CMD_680:
		DeviceExtension->ControllerSpeed = Udma133;
		DeviceExtension->ucCardType = CMD_680;
		break;

	case PCI_DEVICE_ID_CMD_3112:
		DeviceExtension->ControllerSpeed = Udma133;
		DeviceExtension->ucCardType = SIL_3112;
		break;

	case PCI_DEVICE_ID_SERVERWORKS_CSB5IDE:
		DeviceExtension->ControllerSpeed = Udma100;
		DeviceExtension->ucCardType = RCC_SERVER_WORKS_CSB5;
		break;

	case PCI_DEVICE_ID_SERVERWORKS_CSB6IDE:
		DeviceExtension->ControllerSpeed = Udma100;
		DeviceExtension->ucCardType = RCC_SERVER_WORKS_CSB6;
		break;

	default:
		printk_Rlsi("Unknown device!!!! sPciDev %ld\n", DeviceExtension->ulDeviceId);
		break;
	}
/* Vani: Initialise PowerManagement */
	mide_pm_init(DeviceExtension);

	DeviceExtension->Channel[0].BaseIoAddress1 =
	    (PIDE_REGISTERS_1)(bus_get_resource_start(dev, SYS_RES_IOPORT, 0x10) & 0xfffffffe);
	DeviceExtension->Channel[0].BaseIoAddress2 =
	    (PIDE_REGISTERS_2)(bus_get_resource_start(dev, SYS_RES_IOPORT, 0x14) & 0xfffffffe);
	DeviceExtension->Channel[1].BaseIoAddress1 =
	    (PIDE_REGISTERS_1)(bus_get_resource_start(dev, SYS_RES_IOPORT, 0x18) & 0xfffffffe);
	DeviceExtension->Channel[1].BaseIoAddress2 =
	    (PIDE_REGISTERS_2)(bus_get_resource_start(dev, SYS_RES_IOPORT, 0x1C) & 0xfffffffe);
	DeviceExtension->Channel[0].BaseBmAddress =
	    (PBM_REGISTERS)(bus_get_resource_start(dev, SYS_RES_IOPORT, 0x20) & 0xfffffffe);
	DeviceExtension->Channel[1].BaseBmAddress =
	    (PBM_REGISTERS)((bus_get_resource_start(dev, SYS_RES_IOPORT, 0x20) & 0xfffffffe) + 8);

	    /* Vani 1121: Need to comment out the following as we need to allocate memory resource for it */
/*
	if ((DeviceExtension->ulDeviceId == PCI_DEVICE_ID_CMD_680) ||
	    (DeviceExtension->ulDeviceId == PCI_DEVICE_ID_CMD_3112)) {
		DeviceExtension->Channel[0].CntrlFunc =
		    (PCONTROL_FUNTION_REG_SIL680) (bus_get_resource_start(dev, SYS_RES_MEMORY, 0x24) & 0xfffffffe);
		DeviceExtension->Channel[1].CntrlFunc = DeviceExtension->Channel[0].CntrlFunc;
	}
*/
	DeviceExtension->last_free_scb = 0;
	if (AR_FALSE == get_hyperdisk_option_rom_info(DeviceExtension)) {
		printk_Rlsi("get_hyperdisk option_rom_infofailed\n");
		CleanUpAdapter(DeviceExtension);
		return (ENXIO);
	}
	if (ClaimOurIOPorts(DeviceExtension) == FALSE) {
		xdebug("ClaimOurIOPorts: failed");
		CleanUpAdapter(DeviceExtension);
		ReleaseOurIOPorts(DeviceExtension);
		return (ENXIO);
	}
	DisableInterrupts(DeviceExtension);

	if (AR_ERROR == HyperDiskScanForDevices(DeviceExtension)) {
		printk_Rlsi("Scanning Devices failed\n");
		CleanUpAdapter(DeviceExtension);
		ReleaseOurIOPorts(DeviceExtension);
		return (ENXIO);
	}
	if (megaide_cam_attach(DeviceExtension)) {
		printk_Rlsi("megaide_cam_attach failed\n");
		CleanUpAdapter(DeviceExtension);
		ReleaseOurIOPorts(DeviceExtension);
		return (ENXIO);
	}
	if ((error = bus_setup_intr(dev,
		(struct resource *)DeviceExtension->isr_resource, INTR_TYPE_CAM,
		megaide_intr, DeviceExtension, &DeviceExtension->isr_handle))) {
		printk_Rlsi("bus_setup_intr: failed");
		CleanUpAdapter(DeviceExtension);
		ReleaseOurIOPorts(DeviceExtension);
		return (ENXIO);
	}
	return (0);
}

/*
 * Author : Vanishree
 * Date :11/25/02
 * Purpose: To allocate the memory resource for SATA and 680 controllers.
 */
BOOLEAN
alloc_mem_resource(PHW_DEVICE_EXTENSION pDE)
{
	int rid;
	device_t dev;
	ULONG Count = 0;
	dev = gastCardInfo[pDE->ucControllerId].pstDev;

	rid = 0x24;
	if (GET_DEVICE_ID(pDE) == PCI_DEVICE_ID_CMD_3112)
	{
		Count = SIL_3112_BASE_REG_5_DEPTH;
	}
	if(GET_DEVICE_ID(pDE) == PCI_DEVICE_ID_CMD_680)
	{
		Count = SIL_680_BASE_REG_5_DEPTH;
	}
	pDE->base5_resource =  (void *)bus_alloc_resource(dev, SYS_RES_MEMORY, &rid, 0, ~0, Count, RF_ACTIVE);
	if (!pDE->base5_resource)
		return FALSE;
	pDE->BaseMemAddress5 = rman_get_start((struct resource *)pDE->base5_resource);
    	pDE->virtBaseMemAddress5 = (u_char *)rman_get_virtual((struct resource *)pDE->base5_resource);

	return TRUE;

}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

BOOLEAN
ClaimOurIOPorts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	int ControllerNumber;
	device_t dev;
	int rid;
	int iobase;

	ControllerNumber = DeviceExtension->ucControllerId;
	dev = gastCardInfo[ControllerNumber].pstDev;

	rid = 0x10;		/* primary channel - 1st range */
	DeviceExtension->Channel[0].base1_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 8, RF_ACTIVE);
	if (!DeviceExtension->Channel[0].base1_resource)
		return FALSE;

	rid = 0x14;		/* primary channel - 2nd range */
	DeviceExtension->Channel[0].base2_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 4, RF_ACTIVE);
	iobase = bus_get_resource_start(dev, SYS_RES_IOPORT, 0x14);
	if (!DeviceExtension->Channel[0].base2_resource) {

		DeviceExtension->Channel[0].base2_resource =
		    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, iobase, iobase, 1, RF_ACTIVE);
		if (!DeviceExtension->Channel[0].base2_resource)
			return FALSE;
		else {
			DeviceExtension->Channel[0].AltStatus1Byte = TRUE;

			/*
			 * FIXME: Vani 1008: There is some problem in OS,
			 * tough it allocates 4 bytes for 2nd range, it
			 * does not read alternate status register from
			 * offset 2
			 */

			DeviceExtension->Channel[1].AltStatus1Byte = TRUE;
			AR_DEBUG(AR_DEBUG_WARNING, ("AltStatus1Byte Set to 1 chan0\n"));
		}
	}
	rid = 0x18;		/* secondary channel - 1st range */
	DeviceExtension->Channel[1].base1_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 8, RF_ACTIVE);
	if (!DeviceExtension->Channel[1].base1_resource)
		return FALSE;

	rid = 0x1C;		/* secondary channel - 2nd range */
	DeviceExtension->Channel[1].base2_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 4, RF_ACTIVE);
	iobase = bus_get_resource_start(dev, SYS_RES_IOPORT, 0x1C);
	if (!DeviceExtension->Channel[1].base2_resource) {

		DeviceExtension->Channel[1].base2_resource =
		    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, iobase, iobase, 1, RF_ACTIVE);
		if (!DeviceExtension->Channel[1].base2_resource)
			return FALSE;
		else {
			DeviceExtension->Channel[1].AltStatus1Byte = TRUE;
			AR_DEBUG(AR_DEBUG_WARNING, ("AltStatus1Byte Set to 1 chan1\n"));
		}
	}
	rid = 0x20;		/* secondary channel - bmaddr */
	DeviceExtension->Channel[0].bmaddr_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 16, RF_ACTIVE);
	if (!DeviceExtension->Channel[0].bmaddr_resource)
		return FALSE;

	DeviceExtension->Channel[1].bmaddr_resource =
	    DeviceExtension->Channel[0].bmaddr_resource;

 	if ((GET_DEVICE_ID(DeviceExtension) == PCI_DEVICE_ID_CMD_3112) || (GET_DEVICE_ID(DeviceExtension) == PCI_DEVICE_ID_CMD_680) ) {
	   if(FALSE ==  alloc_mem_resource(DeviceExtension))
	   	return FALSE;
	 }


	rid = 0;		/* irq */
	DeviceExtension->isr_resource =
	    (void *)bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0,
	    ~0, 1, RF_SHAREABLE | RF_ACTIVE);
	if (!DeviceExtension->isr_resource)
		return FALSE;


	return (TRUE);
}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

BOOLEAN
ReleaseOurIOPorts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	int ControllerNumber;
	device_t dev;

	ControllerNumber = DeviceExtension->ucControllerId;
	dev = gastCardInfo[ControllerNumber].pstDev;

#ifdef BP_SUPPORT
	idebp_release(DeviceExtension);
#endif
	if (DeviceExtension->Channel[0].base1_resource)
		bus_release_resource(dev, SYS_RES_IOPORT, 0x10,
		    (struct resource *)DeviceExtension->Channel[0].base1_resource);

	if (DeviceExtension->Channel[0].base2_resource)
		bus_release_resource(dev, SYS_RES_IOPORT, 0x14,
		    (struct resource *)DeviceExtension->Channel[0].base2_resource);

	if (DeviceExtension->Channel[1].base1_resource)
		bus_release_resource(dev, SYS_RES_IOPORT, 0x18,
		    (struct resource *)DeviceExtension->Channel[1].base1_resource);

	if (DeviceExtension->Channel[1].base2_resource)
		bus_release_resource(dev, SYS_RES_IOPORT, 0x1C,
		    (struct resource *)DeviceExtension->Channel[1].base2_resource);

	if (DeviceExtension->Channel[0].bmaddr_resource)
		bus_release_resource(dev, SYS_RES_IOPORT, 0x20,
		    (struct resource *)DeviceExtension->Channel[0].bmaddr_resource);

	if (DeviceExtension->base5_resource)
		bus_release_resource(dev, SYS_RES_MEMORY, 0x24,
		    (struct resource *)DeviceExtension->base5_resource);

	if (DeviceExtension->isr_resource)
		bus_release_resource(dev, SYS_RES_IRQ, 0,
		    (struct resource *)DeviceExtension->isr_resource);

	return (0);
}

/*
 * -------------------------------------------------------------------
 * Description This function dissables the controller interupt
 *
 * Arguments - Input Device extension structure pointer
 *
 * Return Value
 *
 * Returns AR_TRUE for success
 * -------------------------------------------------------------------
 */

BOOLEAN
DisableInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR CardType = DeviceExtension->ucCardType;

	return (
	    gaCardSpecificFunctions[CardType].pfnDisableInterrupts(DeviceExtension));
}				/* Disable Interrupts */

BOOLEAN
Disable64xInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	PBM_REGISTERS pstBMRegister = NULL;
	UCHAR opcimcr;

	pstBMRegister = DeviceExtension->Channel[0].BaseBmAddress;

	/*
	 * Enable the Interrupt notification so that further interrupts are
	 * got. This is done because, there is no interrupt handler at the
	 * time before the actual registration of the Int. handler..
	 */
	opcimcr = ScsiPortReadPortUchar(
	    DeviceExtension->Channel[0].bmaddr_resource,
	    (unsigned)((PUCHAR)pstBMRegister + 1));
	opcimcr |= 0x30;
	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[0].bmaddr_resource,
	    (unsigned)((PUCHAR)pstBMRegister + 1), opcimcr);

	return AR_TRUE;

}				/* Disable64xInterrupts */

BOOLEAN
Disable68xInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR opcimcr;

	pci_read_config_byte_Rlsi(DeviceExtension,
	    SYS_CONF_STAT_CMD_REG680 + 2, &opcimcr);
	opcimcr |= 0xC0;	/* 1100 0000 */
	pci_write_config_byte_Rlsi(DeviceExtension,
	    SYS_CONF_STAT_CMD_REG680 + 2, opcimcr);

	return AR_TRUE;

}				/* Disable68xInterrupts */

BOOLEAN
DisableCSB5Interrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	outb_Rlsi(0xC00, 0x02);
	outb_Rlsi(0xC01, 0x00);
	outb_Rlsi(0xC00, 0x04);
	outb_Rlsi(0xC01, 0x00);

	return AR_TRUE;
}				/* DisableCSB5Interrupts */

BOOLEAN
DisableSata3112Interrupts(PHW_DEVICE_EXTENSION pDE)
{
	UCHAR opcimcr;

	pci_read_config_byte_Rlsi(pDE, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr |= 0xC0; /*1100 0000*/
	pci_write_config_byte_Rlsi(pDE, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	if(pDE->PMState == NORMAL)
		sataDisablePMInterrupts(pDE);

	sataDisableHPInterrupt(pDE);
	return AR_TRUE;

}/*DisablesataInterrupts*/

/**
 ** Silicon Image sata Related Functions
 **/


/*
 * -------------------------------------------------------------------
 * Description This function Enables the controller interupt
 *
 * Arguments - Input Device extension structure pointer
 *
 * Return Value
 *
 * Returns AR_TRUE for success
 * -------------------------------------------------------------------
 */
BOOLEAN
EnableInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PBM_REGISTERS pstBMRegister = NULL;
	ULONG ulDrive, ulChannel;
	UCHAR statusByte;
	UCHAR ucTargetId;
	PCHANNEL pChannel;

	/* clear drive interrupts */
	for (ulChannel = 0;
	    ulChannel < MAX_CHANNELS_PER_CONTROLLER;
	    ulChannel++) {
		pChannel = &DeviceExtension->Channel[ulChannel];
		baseIoAddress1 = (PIDE_REGISTERS_1)pChannel->BaseIoAddress1;
		baseIoAddress2 = (PIDE_REGISTERS_2)pChannel->BaseIoAddress2;

		for (ulDrive = 0; ulDrive < MAX_DRIVES_PER_CHANNEL; ulDrive++) {
			ucTargetId = (ulChannel << 1) + ulDrive;
			if (!IS_IDE_DRIVE(DeviceExtension, ucTargetId))
				continue;

			SELECT_DEVICE(pChannel->base1_resource, baseIoAddress1, ulDrive);
			GET_BASE_STATUS(pChannel->base1_resource, baseIoAddress1, statusByte);
		}
	}

	/* clear drive interrupt */
	for (ulChannel = 0;
	    ulChannel < MAX_CHANNELS_PER_CONTROLLER;
	    ulChannel++) {
		pChannel = &DeviceExtension->Channel[ulChannel];
		pstBMRegister = pChannel->BaseBmAddress;
		CLEAR_BM_INT(pChannel->bmaddr_resource, pstBMRegister, statusByte);
	}

	pstBMRegister = DeviceExtension->Channel[0].BaseBmAddress;

	/*
	 * Enable the Interrupt notification so that further interrupts are
	 * got. This is done because, there is no interrupt handler at the
	 * time before the actual registration of the Int. handler..
	 */

	return gaCardSpecificFunctions[DeviceExtension->ucCardType].pfnEnableInterrupts(DeviceExtension);

}				/* Enable Interrupts() */



BOOLEAN
Enable64xInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	PBM_REGISTERS pstBMRegister = NULL;
	UCHAR opcimcr;
	PCHANNEL pChannel = &DeviceExtension->Channel[0];

	pstBMRegister = pChannel->BaseBmAddress;

	opcimcr = ScsiPortReadPortUchar(
	    pChannel->bmaddr_resource,
	    (unsigned)((PUCHAR)pstBMRegister + 1));
	opcimcr &= 0xCF;
	opcimcr |= 0x01;
	ScsiPortWritePortUchar(
	    pChannel->bmaddr_resource,
	    (unsigned)((PUCHAR)pstBMRegister + 1), opcimcr);

	return AR_TRUE;
}				/* Enable64xInterrupts() */

BOOLEAN
Enable68xInterrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR opcimcr;

	pci_read_config_byte_Rlsi(DeviceExtension,
	    SYS_CONF_STAT_CMD_REG680 + 2, &opcimcr);
	opcimcr &= 0x3F;	/* 0011 111 */
	pci_write_config_byte_Rlsi(DeviceExtension,
	    SYS_CONF_STAT_CMD_REG680 + 2, opcimcr);

	return AR_TRUE;
}				/* Enable68xInterrupts() */

BOOLEAN
EnableCSB5Interrupts(PHW_DEVICE_EXTENSION DeviceExtension)
{
	outb_Rlsi(0xC00, 0x02);
	outb_Rlsi(0xC01, DeviceExtension->ulIntLine);
	outb_Rlsi(0xC00, 0x04);
	outb_Rlsi(0xC01, DeviceExtension->ulIntLine);

	return AR_TRUE;
}				/* EnableCSB5Interrupts() */

BOOLEAN
EnableSata3112Interrupts(PHW_DEVICE_EXTENSION pDE)
{
	UCHAR opcimcr;

	pci_read_config_byte_Rlsi(pDE, SYS_CONF_STAT_CMD_REG680+2, &opcimcr);
	opcimcr &= 0x3F; /*0011 111*/
	pci_write_config_byte_Rlsi(pDE, SYS_CONF_STAT_CMD_REG680+2, opcimcr);

	sataEnablePMInterrupts(pDE);
	if(pDE->PMState == NORMAL)
		sataEnableHPInterrupt(pDE);
	return AR_TRUE;
}/*EnablesataInterrupts() */


/*
 * -------------------------------------------------------------------
 *
 * Description : This function will check for the presence of the device with
 * the given target ID
 *
 * Input Parameters: Ponter to the device extention structure and target ID
 *
 * Output Parameters: AR_TRUE		If present AR_FALSE	It not
 * present
 * -------------------------------------------------------------------
 */

BOOLEAN
IsDrivePresent(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR ucTargetId)
{
	UCHAR channel = (ucTargetId >> 1);
	PIDE_REGISTERS_1 pstBaseIoAddress1 =
	DeviceExtension->Channel[channel].BaseIoAddress1;
	UCHAR ucStatus;
	UCHAR ucDrvSelVal, ucReadDrvNum;
	PCHANNEL pChannel = &DeviceExtension->Channel[channel];

	AR_DEBUG(AR_DEBUG_FINDDEV,
	    ("is_drive _present: channel=%d \n", channel));

	SELECT_DEVICE(pChannel->base1_resource, pstBaseIoAddress1, ucTargetId);

	GET_STATUS(pChannel->base1_resource, pstBaseIoAddress1, ucStatus);

	if (0xff == ucStatus) {
		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("No device Status returned as 0xff\n"));
		return AR_FALSE;
	}
	SELECT_DEVICE(pChannel->base1_resource,
	    pstBaseIoAddress1, ucTargetId);

	ScsiPortWritePortUchar(
	    pChannel->base1_resource,
	    (unsigned)&pstBaseIoAddress1->SectorCount, 0xff);

	ScsiPortStallExecution(60);

	ucReadDrvNum = ScsiPortReadPortUchar(pChannel->base1_resource,
	    (unsigned)&pstBaseIoAddress1->DriveSelect);

	ucDrvSelVal = IDE_CHS_MODE | ((ucTargetId & 0x01) << 4);

	if (ucReadDrvNum == ucDrvSelVal) {
		AR_DEBUG(AR_DEBUG_FINDDEV, ("Drive responding at %d\n",
			ucTargetId));
		return AR_TRUE;
	} else {

		AR_DEBUG(AR_DEBUG_FINDDEV, ("No drive present at %d\n",
			ucTargetId));
		return AR_FALSE;
	}
}				/* IsDrivePresent() */



/*
 * -------------------------------------------------------------------
 * Description Issue IDENTIFY command to a device.
 *
 * Arguments - Input DeviceExtension - HBA miniport driver's adapter data
 * storage DeviceNumber - Indicates which device. Command - Either the
 * standard(EC) or the ATAPI packet(A1) IDENTIFY.
 *
 * Return Value
 *
 * Returns AR_TRUE for success Returns AR_TRUE for failure
 * -------------------------------------------------------------------
 */

BOOLEAN
IssueIdentify(
    PHW_DEVICE_EXTENSION DeviceExtension,
    UCHAR DeviceNumber,
    UCHAR Channel,
    UCHAR Command)
{
	PIDE_REGISTERS_1 baseIoAddress1 =
	DeviceExtension->Channel[Channel].BaseIoAddress1;
	PCHANNEL pChannel = &DeviceExtension->Channel[Channel];
	ULONG i;
	UCHAR statusByte;
	UCHAR targetId;
	PUSHORT puIdentifyData;

	AR_DEBUG(AR_DEBUG_FINDDEV, ("Issue Identify:\n"));

	/* Select device 0 or 1 */

	SELECT_DEVICE(pChannel->base1_resource, baseIoAddress1, DeviceNumber);

	/*
	 * Check that the status register makes sense. The call came here
	 * since there is a drive. so let us not worry about whether there is
	 * any drive at this place
	 */
	GET_BASE_STATUS(pChannel->base1_resource, baseIoAddress1, statusByte);

	/*
	 * Load CylinderHigh and CylinderLow with number bytes to transfer.
	 */

	ScsiPortWritePortUchar(pChannel->base1_resource, (unsigned)&baseIoAddress1->CylinderHigh, (0x200 >> 8));
	ScsiPortWritePortUchar(pChannel->base1_resource, (unsigned)&baseIoAddress1->CylinderLow, (0x200 & 0xFF));

	/* Send IDENTIFY command. */
	ScsiPortWritePortUchar(pChannel->base1_resource, (unsigned)&baseIoAddress1->Command, Command);

	WAIT_ON_BUSY(pChannel->base1_resource, baseIoAddress1, statusByte);

	if ((!(statusByte & IDE_STATUS_BUSY)) &&
	    (!(statusByte & IDE_STATUS_DRQ))) {
		AR_DEBUG(AR_DEBUG_FINDDEV, ("Identify Failed\n"));
		/* this is an error... so let us not try any more. */
		return AR_FALSE;
	}
	WAIT_ON_BUSY(pChannel->base1_resource, baseIoAddress1, statusByte);

	/* Wait for DRQ. */

	WAIT_FOR_DRQ(pChannel->base1_resource, baseIoAddress1, statusByte);


	/* Read status to acknowledge any interrupts generated. */

	GET_BASE_STATUS(pChannel->base1_resource, baseIoAddress1, statusByte);

	/*
	 * Check for error on really stupid master devices that assert random
	 * patterns of bits in the status register at the slave address.
	 */

	if ((Command == IDE_COMMAND_IDENTIFY) &&
	    (statusByte & IDE_STATUS_ERROR)) {
		AR_DEBUG(AR_DEBUG_FINDDEV, ("Identify Failed : 3\n"));
		return AR_FALSE;
	}
	AR_DEBUG(AR_DEBUG_FINDDEV, ("Stat b4 read words %x\n", statusByte));


	/*
	 * Suck out 256 words. After waiting for one model that asserts busy
	 * after receiving the Packet Identify command.
	 */

	WAIT_ON_BUSY(pChannel->base1_resource, baseIoAddress1, statusByte);

	if (!(statusByte & IDE_STATUS_DRQ)) {
		AR_DEBUG(AR_DEBUG_FINDDEV, ("Failed to access\n"));
		return AR_FALSE;
	}
	targetId = (Channel << 1) + DeviceNumber;

	memset_Rlsi(
	    &DeviceExtension->PhysicalDrive[targetId].IdentifyData,
	    0, sizeof(IDENTIFY_DATA));

	AR_READ_BUFFER(pChannel->base1_resource, baseIoAddress1,
	    (PUSHORT)&DeviceExtension->PhysicalDrive[targetId].IdentifyData, 256);
	/*
	 * Check out a few capabilities / limitations of the device.
	 */

	if (DeviceExtension->PhysicalDrive[targetId].IdentifyData.MediaStatusNotification & 1) {
		/* Determine if this drive supports the MSN functions. */

		AR_DEBUG(AR_DEBUG_WARNING,
		    ("Issue Identify:Marking drive %d as removable. MSN = %d!!!\n",
			Channel * 2 + DeviceNumber,
			DeviceExtension->PhysicalDrive[targetId].IdentifyData.MediaStatusNotification));
		/*
		 * Manoj FIXME Some time I am getting wrong Identify
		 * information from the disc and it says that the device is
		 * removable drive
		 */
		DeviceExtension->PhysicalDrive[(Channel * 2) + DeviceNumber].DeviceFlags |=
		    DFLAGS_REMOVABLE_DRIVE;
	}
	if (DeviceExtension->PhysicalDrive[targetId].IdentifyData.MaximumBlockTransfer) {
		/*
		 * Determine max. block transfer for this device.
		 */

		DeviceExtension->PhysicalDrive[targetId].MaximumBlockXfer =
		    (UCHAR)(DeviceExtension->PhysicalDrive[targetId].IdentifyData.\
		    MaximumBlockTransfer & 0xFF);
	}
	puIdentifyData = (PUSHORT)&(DeviceExtension->PhysicalDrive[targetId].IdentifyData);

	if (DeviceExtension->PhysicalDrive[targetId].IdentifyData.GeneralConfiguration & 0x20 &&
	    Command != IDE_COMMAND_IDENTIFY) {
		/*
		 * This device interrupts with the assertion of DRQ after
		 * receiving Atapi Packet Command
		 */

		DeviceExtension->PhysicalDrive[targetId].DeviceFlags |= DFLAGS_INT_DRQ;

		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("Device interrupts on assertion of DRQ.\n"));

	}
#ifdef AR_DEBUG
	else {

		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("Device does not interrupt on assertion of DRQ.\n"));
	}
#endif
	if (((DeviceExtension->PhysicalDrive[targetId].IdentifyData.GeneralConfiguration & 0xF00)
		== 0x100) && Command != IDE_COMMAND_IDENTIFY) {
		/* This is a tape. */

		DeviceExtension->PhysicalDrive[targetId].DeviceFlags |= DFLAGS_TAPE_DEVICE;

		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("Issue Identify: Device is a tape drive.\n"));

	} else {
		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("Issue Identify:Device is not a tape drive.\n"));
	}

	/*
	 * Work around for some IDE and one model Atapi that will present
	 * more than 256 bytes for the Identify data.
	 */

	WAIT_ON_BUSY(pChannel->base1_resource, baseIoAddress1, statusByte);

	for (i = 0; i < 0x10000; i++) {
		GET_STATUS(pChannel->base1_resource, baseIoAddress1, statusByte);

		if (statusByte & IDE_STATUS_DRQ) {
			/* Suck out any remaining bytes and throw away. */

		} else {
			break;
		}
	}

	AR_DEBUG(AR_DEBUG_FINDDEV,
	    ("Status after read words(%x) \n", statusByte));

	return AR_TRUE;

}				/* IssueIdentify() */

/*
 * -------------------------------------------------------------------
 * Description Initialize some features on drive(which are required at the
 * initialization time. Right now we are enabling Cache implementation
 * features...(this is to resolve the bug of bad performance on some IBM
 * Drives)
 *
 * Arguments - Input DeviceExtension - HBA miniport driver's adapter data
 * storage
 *
 * Return Value
 *
 * Returns AR_TRUE for success Returns AR_TRUE for failure
 * -------------------------------------------------------------------
 */
BOOLEAN
InitDriveFeatures(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR ucDrvInd;

	for (ucDrvInd = 0; ucDrvInd < MAX_DRIVES_PER_CONTROLLER; ucDrvInd++) {
		InitSingleDriveFeatures(DeviceExtension, ucDrvInd);
	}
	return AR_TRUE;

}				/* Init DriveFeatures */


BOOLEAN
InitSingleDriveFeatures(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR ucDrvInd)
{
	UCHAR channel = (ucDrvInd >> 1);
	PIDE_REGISTERS_1 baseIoAddress;
	UCHAR statusByte;

	if (!IS_IDE_DRIVE(DeviceExtension, ucDrvInd))
		return AR_FALSE;

	baseIoAddress = DeviceExtension->Channel[channel].BaseIoAddress1;

	SELECT_DEVICE(DeviceExtension->Channel[channel].base1_resource, baseIoAddress, ucDrvInd);

	if (DeviceExtension->PhysicalDrive[ucDrvInd].bEnableRwCache) {
		AR_DEBUG(AR_DEBUG_INITDEV, ("Write Cache feature Enabled \n"));
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)((PUCHAR)baseIoAddress + 1),
		    FEATURE_ENABLE_WRITE_CACHE);
	} else {

		AR_DEBUG(AR_DEBUG_INITDEV, ("Write Cache feature Disabled\n"));
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)((PUCHAR)baseIoAddress + 1),
		    FEATURE_DISABLE_WRITE_CACHE);
	}

	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[channel].base1_resource,
	    (unsigned)&(baseIoAddress->Command),
	    IDE_COMMAND_SET_FEATURES);
	WAIT_ON_BASE_BUSY(
	    DeviceExtension->Channel[channel].base1_resource,
	    baseIoAddress, statusByte);

	if (DeviceExtension->PhysicalDrive[ucDrvInd].bEnableReadAheadCache) {
		AR_DEBUG(AR_DEBUG_INITDEV, ("Read Ahead feature Enabled \n"));
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)((PUCHAR)baseIoAddress + 1),
		    FEATURE_ENABLE_READ_CACHE);
	} else {
		AR_DEBUG(AR_DEBUG_INITDEV, ("Read Ahead feature Disabled \n"));
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)((PUCHAR) baseIoAddress + 1),
		    FEATURE_DISABLE_READ_CACHE);
	}

	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[channel].base1_resource,
	    (unsigned)&(baseIoAddress->Command),
	    IDE_COMMAND_SET_FEATURES);
	WAIT_ON_BASE_BUSY(
	    DeviceExtension->Channel[channel].base1_resource,
	    baseIoAddress, statusByte);

	set_multi_block_xfers(DeviceExtension, ucDrvInd);

	return (AR_TRUE);
}				/* Init DriveFeatures */


BOOLEAN
set_multi_block_xfers(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR ucTargetId)
{
	UCHAR channel = (ucTargetId >> 1);
	PIDE_REGISTERS_1 baseIoAddress;
	UCHAR statusByte = 0, errorByte;

	baseIoAddress = DeviceExtension->Channel[channel].BaseIoAddress1;

	AR_DEBUG(AR_DEBUG_INITDEV,
	    ("set_multi_block_xfers TID=%d \n", ucTargetId));

	/* If supported, setup Multi-block transfers. */
	if (DeviceExtension->PhysicalDrive[ucTargetId].MaximumBlockXfer != 0) {
		/* Select the device. */

		SELECT_DEVICE(
		    DeviceExtension->Channel[channel].base1_resource,
		    baseIoAddress, ucTargetId);

		/* Setup sector count to reflect the # of blocks. */
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)&baseIoAddress->SectorCount,
		    DeviceExtension->PhysicalDrive[ucTargetId].MaximumBlockXfer);

		/* Issue the command. */
		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[channel].base1_resource,
		    (unsigned)&baseIoAddress->Command,
		    IDE_COMMAND_SET_MULTIPLE);

		/* Wait for busy to drop. */
		WAIT_ON_BASE_BUSY(
		    DeviceExtension->Channel[channel].base1_resource,
		    baseIoAddress, statusByte);

		/*
		 * Check for errors. Reset the value to 0(disable MultiBlock)
		 * if the command was aborted.
		 */

		if (statusByte & IDE_STATUS_ERROR) {

			/* Read the error register. */
			errorByte = ScsiPortReadPortUchar(
			    DeviceExtension->Channel[channel].base1_resource,
			    (unsigned)((PUCHAR)baseIoAddress + 1));

			AR_DEBUG(AR_DEBUG_INITDEV,
			    ("Error setting mult mode.Stat %x, err %x\n",
				statusByte, errorByte));

			/* Adjust the devExt. value, if necessary. */
			DeviceExtension->PhysicalDrive[ucTargetId].MaximumBlockXfer = 0;

		} else {
			AR_DEBUG(AR_DEBUG_INITDEV,
			    ("Using Multi on TID %d. Blocks, int - %ld\n",
				ucTargetId,
				DeviceExtension->PhysicalDrive[ucTargetId].MaximumBlockXfer));
		}
	}
	return AR_TRUE;

}				/* set_multi_block_xfers() */




void
SetDriveParameters(PHW_DEVICE_EXTENSION DeviceExtension)
{
	ULONG i;
	UCHAR statusByte = 0;
	UCHAR DeviceNumber, Channel;
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PIDENTIFY_DATA identifyData = NULL;


	for (DeviceNumber = 0;
	    DeviceNumber < MAX_DRIVES_PER_CONTROLLER;
	    DeviceNumber++) {

		if (!IS_IDE_DRIVE(DeviceExtension, DeviceNumber))
			continue;

		Channel = DeviceNumber >> 1;
		baseIoAddress1 = (PIDE_REGISTERS_1)DeviceExtension->Channel[Channel].BaseIoAddress1;
		baseIoAddress2 = (PIDE_REGISTERS_2)DeviceExtension->Channel[Channel].BaseIoAddress2;
		identifyData = &DeviceExtension->PhysicalDrive[DeviceNumber].IdentifyData;


		/*
		 * Set up registers for SET PARAMETER command.
		 */

		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[Channel].base1_resource,
		    (unsigned)&baseIoAddress1->DriveSelect,
		    (UCHAR) (((DeviceNumber << 4) | 0xA0) | (identifyData->NumberOfHeads - 1)));

		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[Channel].base1_resource,
		    (unsigned)&baseIoAddress1->SectorCount,
		    (UCHAR)identifyData->SectorsPerTrack);

		/*
		 * Send SET PARAMETER command.
		 */

		ScsiPortWritePortUchar(
		    DeviceExtension->Channel[Channel].base1_resource,
		    (unsigned)&baseIoAddress1->Command,
		    IDE_COMMAND_SET_DRIVE_PARAMETERS);


		ScsiPortStallExecution(1000);

		/*
		 * Wait for up to 30 milliseconds for ERROR or command
		 * complete.
		 */

		for (i = 0; i < 30 * 1000; i++) {
			UCHAR errorByte;

			statusByte = GET_ALTERNATE_STATUS_REGISTER_VALUE(
			    DeviceExtension->Channel[Channel].base2_resource,
			    baseIoAddress2, DeviceExtension->Channel[Channel].AltStatus1Byte);
			if (statusByte & IDE_STATUS_ERROR) {
				errorByte = ScsiPortReadPortUchar(
				    DeviceExtension->Channel[Channel].base1_resource,
				    (unsigned)((PUCHAR)baseIoAddress1 + 1));

				AR_DEBUG(AR_DEBUG_INITDEV, ("SetDriveParameters: Error bit set. Status %x, error %x\n",
					errorByte,
					statusByte));

				break;

			} else if ((statusByte & ~IDE_STATUS_INDEX) == IDE_STATUS_IDLE) {

				break;

			} else {

				ScsiPortStallExecution(100);
			}
		}


		AR_DEBUG(AR_DEBUG_INITDEV,
		    ("Status after setting Drive Parameters : %X\n", statusByte));

	}


}				/* end SetDriveParameters() */



void
SetDriveParams(
    PHW_DEVICE_EXTENSION DeviceExtension,
    UCHAR DeviceNumber)
{
	ULONG i;
	UCHAR statusByte = 0;
	UCHAR Channel;
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PIDENTIFY_DATA identifyData = NULL;


	if (!IS_IDE_DRIVE(DeviceExtension, DeviceNumber))
		return;

	Channel = DeviceNumber >> 1;
	baseIoAddress1 = (PIDE_REGISTERS_1)DeviceExtension->Channel[Channel].BaseIoAddress1;
	baseIoAddress2 = (PIDE_REGISTERS_2)DeviceExtension->Channel[Channel].BaseIoAddress2;
	identifyData = &DeviceExtension->PhysicalDrive[DeviceNumber].IdentifyData;


	/*
	 * Set up registers for SET PARAMETER command.
	 */

	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[Channel].base1_resource,
	    (unsigned)&baseIoAddress1->DriveSelect,
	    (UCHAR)(((DeviceNumber << 4) | 0xA0) | (identifyData->NumberOfHeads - 1)));

	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[Channel].base1_resource,
	    (unsigned)&baseIoAddress1->SectorCount,
	    (UCHAR)identifyData->SectorsPerTrack);

	/*
	 * Send SET PARAMETER command.
	 */

	ScsiPortWritePortUchar(
	    DeviceExtension->Channel[Channel].base1_resource,
	    (unsigned)&baseIoAddress1->Command,
	    IDE_COMMAND_SET_DRIVE_PARAMETERS);


	ScsiPortStallExecution(1000);

	/*
	 * Wait for up to 30 milliseconds for ERROR or command complete.
	 */

	for (i = 0; i < 30 * 1000; i++) {
		UCHAR errorByte;

		statusByte = GET_ALTERNATE_STATUS_REGISTER_VALUE(
		    DeviceExtension->Channel[Channel].base2_resource,
		    baseIoAddress2, DeviceExtension->Channel[Channel].AltStatus1Byte);
		if (statusByte & IDE_STATUS_ERROR) {
			errorByte = ScsiPortReadPortUchar(
			    DeviceExtension->Channel[Channel].base1_resource,
			    (unsigned)((PUCHAR)baseIoAddress1 + 1));

			AR_DEBUG(AR_DEBUG_INITDEV,
			    ("SetDriveParameters: Error bit set. Status %x, error %x\n",
				errorByte,
				statusByte));

			break;

		} else if ((statusByte & ~IDE_STATUS_INDEX) == IDE_STATUS_IDLE) {
			break;
		} else {
			ScsiPortStallExecution(100);
		}
	}

	AR_DEBUG(AR_DEBUG_INITDEV, ("Status after setting Drive Parameters : %X\n", statusByte));

}				/* end SetDriveParams() */


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
int
get_hyperdisk_option_rom_info(
    PHW_DEVICE_EXTENSION DeviceExtension)
{
	ULONG ulRomAddr;
	PUCHAR pcVirtualRomAddr;


	int iRetVal = AR_FALSE;

	AR_DEBUG(AR_DEBUG_SCAN_OP_ROM,
	    ("get_hyperdisk_option_rom_info:\n"));
	/*
	 * If we already found the card, return there is only one shadow,
	 * regardles on the number of cards installed in the system
	 */


	for (ulRomAddr = 0xC0000, iRetVal = 0;
	    (ulRomAddr < 0xF0000) && (iRetVal == 0);
	    ulRomAddr += (PAGE_SIZE * 8)) {

		pcVirtualRomAddr =
		/* (PUCHAR) ioremap(ulRomAddr & PAGE_MASK, (PAGE_SIZE*8)); */
		    (PUCHAR)BIOS_PADDRTOVADDR(ulRomAddr);

		if (!pcVirtualRomAddr)
			break;

		if (AR_TRUE ==
		    scan_option_rom(pcVirtualRomAddr,
			(PAGE_SIZE * 8), DeviceExtension)) {
#ifdef DEBUG_ON
			AR_DEBUG(AR_DEBUG_SCAN_OP_ROM,
			    ("CMD64X: ROM enabled at 0x%08lx\n", ulRomAddr));
#endif
			iRetVal = AR_TRUE;
		}
		/* iounmap(pcVirtualRomAddr); */

	}

#ifdef DEBUG_ON
	if (AR_FALSE == iRetVal)
		AR_DEBUG(AR_DEBUG_SCAN_OP_ROM,
		    ("CMD64X: WARNING: ROM not enabled\n"));
#endif

	return (iRetVal);
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

int
scan_option_rom(unsigned char *pcRom,
    unsigned long ulRomSize,
    PHW_DEVICE_EXTENSION DeviceExtension)
{
	char acTempStr[80];
	u_long i, j, k;
	u_long *pulTempPtr;
	u_short *pusROMSig;
	u_short bFoundVersion = 0;
	u_long scan_depth;
	char acTempMinor[10];

	AR_DEBUG(AR_DEBUG_SCAN_OP_ROM, ("scan_option_rom:\n"));
	/* look every 2k */

	for (i = 0; i < ulRomSize; i += 0x800) {

		pusROMSig = (u_short *) & pcRom[i];

		if (*pusROMSig != 0xAA55)
			continue;

		/* look for our signature at offset 6 bytes */
		memmove_Rlsi(acTempStr, &pcRom[i + 6], 8);
		acTempStr[8] = (char)0;

		if ((!strcmp_Rlsi(acTempStr, "HyperDis")) ||
		    (!strcmp_Rlsi(acTempStr, "MegaRAID"))) {

			/* get firmware revision */
			scan_depth = ((pcRom[i + 2] * 512) > ulRomSize) ? ulRomSize : (pcRom[i + 2] * 512);

			/* for (j=14; j < 0x2000; j++) { */
			for (j = 14; j < scan_depth; j++) {

				pulTempPtr = (u_long *)&pcRom[i + j];

				/*
				 * looking for 'Hype' looking for 'Mega'
				 * looking for 'IDE '(this is for Compaq
				 * Bios) looking for ''CERC(this is for Dell
				 * Bios) looking for $XIDE$
				 */
				if ((*pulTempPtr == 0x65707948) ||
					(*pulTempPtr == 0x6167654D) ||
				    (*pulTempPtr == 0x20454449) ||
				    (*pulTempPtr == 0x43524543) ||
				    (*pulTempPtr == 0x44495824)) {

					memmove_Rlsi(acTempStr, &pcRom[i + j], 14);
					acTempStr[14] = (char)0;
					acTempMinor[10] = (char)0;

					/* heres $XIDE$ */
					if (!strncmp_Rlsi(acTempStr, "$XIDE$", 6)) {
						j += 6;
						DeviceExtension->stFirmwareVersion.usMajorVer = pcRom[i + j] - '0';
						j++;
						DeviceExtension->stFirmwareVersion.usMinorVer = pcRom[i + j] - '0';
						j++;
						DeviceExtension->stFirmwareVersion.ulBuild = array_to_int(&pcRom[i + j]);
						memmove_Rlsi(acTempMinor, &pcRom[i + j], 8);
						acTempMinor[9] = (char)0;
						printk_Rlsi("<LSILogic MegaRAID IDE> BIOS %d.%d.%s\n",
						    DeviceExtension->stFirmwareVersion.usMajorVer,
						    DeviceExtension->stFirmwareVersion.usMinorVer,
						/*
						 * DeviceExtension->stFirmware
						 * Version.ulBuild);
						 */
						    acTempMinor);

						/* yes, we want this */
						return (AR_TRUE);
					}
					/*
					 * the strcmp string should be 14
					 * bytes long. the 'j' counter should
					 * increment pointer to the location
					 * of 'V' begining at Version look at
					 * offset 26 for firmware version(12
					 * bytes long),
					 */
					else if (!strcmp_Rlsi(acTempStr, "MegaRAID IDE B")) {
						j += 26;
						bFoundVersion = 1;
					}
					/*
					 * heres old hypderdisk string look
					 * at offset 23 for firmware
					 * version(12 bytes long)
					 */
					else if (!strcmp_Rlsi(acTempStr, "HyperDisk BIOS")) {
						j += 23;
						bFoundVersion = 1;
					}
					/*
					 * heres CPQ bios string look at
					 * offset 22 for firmware version(12
					 * bytes long)
					 */
					else if (!strcmp_Rlsi(acTempStr, "IDE RAID BIOS ")) {
						j += 22;
						bFoundVersion = 1;
					}
					/*
					 * heres DELL bios string look at
					 * offset 22 for firmware version(12
					 * bytes long)
					 */
					else if (!strcmp_Rlsi(acTempStr, "CERC ATA/2S BI")) {
						j += 25;
						bFoundVersion = 1;
					}
					if (bFoundVersion) {

						/* get usMajorVer */
						k = 0;
						/* look for '.' */
						while (pcRom[i + j + k] != '.')
							k++;
						DeviceExtension->stFirmwareVersion.usMajorVer =
						    array_to_int(&pcRom[i + j]);


						/* get usMinorVer */
						j += (k + 1);
						k = 0;
						/* look for '.' */
						while (pcRom[i + j + k] != '.')
							k++;
						DeviceExtension->stFirmwareVersion.usMinorVer =
						    array_to_int(&pcRom[i + j]);

						/* get ulBuild */
						j += (k + 1);
						DeviceExtension->stFirmwareVersion.ulBuild =
						    array_to_int(&pcRom[i + j]);


						printk_Rlsi("<LSILogic MegaRAID IDE> BIOS %2.2d.%2.2d.%8.8ld\n",
						    DeviceExtension->stFirmwareVersion.usMajorVer,
						    DeviceExtension->stFirmwareVersion.usMinorVer,
						    DeviceExtension->stFirmwareVersion.ulBuild);

						/* yes, we want this */
						return (AR_TRUE);
					}	/* if (bFoundVersion) */
				}	/* if ((*pulTempPtr == 0x65707948) */
			}	/* for (j=14; */

			return (AR_TRUE);	/* yes, we want this */
		}		/* if ((!strcmp(acTempStr */
	}			/* for (i = 0; i */

	return (AR_FALSE);
}				/* scan_option_rom() */

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
initialize_device_extension(PHW_DEVICE_EXTENSION DeviceExtension)
{
	u_char i, j;

	for (j = 0; j < MAX_LOGICAL_DRIVES; j++) {
		DeviceExtension->LogicalDrive[j].Status = LDS_NotPresent;
		for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
			DeviceExtension->LogicalDrive[j].PhysicalDriveTid[i] = INVALID_DRIVE_ID;
			DeviceExtension->LogicalDrive[j].SlotPhysicalDriveTid[i] = INVALID_DRIVE_ID;
		}
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
		if (DeviceExtension->ucIRCDType == _IDE_IRCD)
			DeviceExtension->PhysicalDrive[i].Status = PDS_Undefined;
		else
			DeviceExtension->PhysicalDrive[i].Status = PDS_OffLine;

		DeviceExtension->PhysicalDrive[i].ucMirrorDriveId = INVALID_DRIVE_ID;
		DeviceExtension->PhysicalDrive[i].LogicalDriveId = INVALID_DRIVE_ID;
		DeviceExtension->PhysicalDrive[i].pArrayInfo = NULL;
		ENABLE_IRCD_UPDATE_ON_DRIVE(i);
	}

	for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
		DeviceExtension->ArrayDrive[i].RaidLevel = None;
		DeviceExtension->ArrayDrive[i].Status = LDS_NotPresent;
		DeviceExtension->ArrayDrive[i].ucArrayDriveNumber = 0;
		DeviceExtension->ArrayDrive[i].ucLogicalDriveCount = 0;
		DeviceExtension->ArrayDrive[i].HeadLogDrive = NULL;
		DeviceExtension->ArrayDrive[i].TailLogDrive = NULL;
	}

}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
LONG
HyperDiskScanForDevices(PHW_DEVICE_EXTENSION DeviceExtension)
{
	LONG lRetVal = AR_SUCCESSS;

	/*
	 * ISR is not ready yet... so let controller not raise interrupts
	 */

	if (!InitIdeRaidController(DeviceExtension)) {
		lRetVal = AR_ERROR;
		AR_DEBUG(AR_DEBUG_WARNING,
		    ("Initializing IDE Raid Controllers failed\n"));
	}
	AR_DEBUG(AR_DEBUG_MODULE,
	    ("Now enable interrupts\n"));

	/* Enable interrutps back */
	AtapiHwInitialize(DeviceExtension);

	AR_DEBUG(AR_DEBUG_MODULE,
	    ("Exiting HyperDiskScanForDevices\n"));
	return lRetVal;

	return AR_SUCCESSS;
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
BOOLEAN
InitIdeRaidController(PHW_DEVICE_EXTENSION DeviceExtension)
{
	BOOLEAN foundDevices;
	UCHAR channel;
	UCHAR targetId;
	UCHAR LogDrvInd;
	UCHAR ulPhyDrvInd;
	char raid_type[20];
	char raid_status[20];

	foundDevices = AR_FALSE;

	for (channel = 0;
	    channel < MAX_CHANNELS_PER_CONTROLLER;
	    channel++) {
		/* Search for devices on this controller. */
		if (FindDevices(DeviceExtension, channel)) {

			AR_DEBUG(AR_DEBUG_INITDEV,
			    ("Found device in Channel%d\n", channel));
			/* Remember that some devices were found. */
			foundDevices = AR_TRUE;

			if ((DeviceExtension->PhysicalDrive[channel << 1].DeviceFlags &
				DFLAGS_DEVICE_PRESENT) &&
			    (DeviceExtension->PhysicalDrive[(channel << 1) + 1].DeviceFlags &
				DFLAGS_DEVICE_PRESENT)) {

				/*
				 * If 2 drives are present, set SwitchDrive
				 * to 1(default is 0) . SwitchDrive is used
				 * to toggle bewteen the two drives'work
				 * queues, to keep the drives fed in a fair
				 * fashion. See Start ChannelIo() .
				 */

				DeviceExtension->Channel[channel].SwitchDrive = 1;
			}
		}		/* if (find_device */
	}			/* for () */

	if (foundDevices) {
		/* Program I/O mode for each drive. */
		for (targetId = 0; targetId < MAX_DRIVES_PER_CONTROLLER;
		    targetId++) {

			if (IS_IDE_DRIVE(DeviceExtension, targetId)) {

				BOOLEAN success;

				success = GetIoMode(DeviceExtension, targetId);

				if (!success) {
					/* Disable device. */
					DeviceExtension->PhysicalDrive[targetId].DeviceFlags &=
					    ~DFLAGS_DEVICE_PRESENT;
				}
				/*
				 * manoj 08022002 programming drives after
				 * Inquiry data collection
				 */
				/*
				 * SetMaxTransferMode(DeviceExtension,
				 * targetId);
				 */

				DeviceExtension->PhysicalDrive[targetId].ucIsPowerConnected = 1;

			}
		}
	}
	/* initialize boot log drv mapping */
	for (ulPhyDrvInd = 0; ulPhyDrvInd < MAX_LOGICAL_DRIVES; ulPhyDrvInd++)
		DeviceExtension->BootLogDrvMapping[ulPhyDrvInd] = ulPhyDrvInd;

	initialize_device_extension(DeviceExtension);

	if (GetConfigInfoAndErrorLogSectorInfo(DeviceExtension)) {
		ConfigureRaidDrives(DeviceExtension,
		    DeviceExtension->ConfigurationVirtualBuffer);
		if (DeviceExtension->ucIRCDType == _IDE_IRCD) {
			EnquireProperties(DeviceExtension, TRUE);
		}
	}
	/* manoj 08022002 programming drives after Inquiry data collection */
	for (ulPhyDrvInd = 0; ulPhyDrvInd < MAX_DRIVES_PER_CONTROLLER; ulPhyDrvInd++) {
		if (DeviceExtension->PhysicalDrive[ulPhyDrvInd].ucIsPowerConnected) {
			SetMaxTransferMode(DeviceExtension, ulPhyDrvInd);
		}
	}

	AR_DEBUG(AR_DEBUG_VANI_CODE,
	    ("completed translation to from New Ircd\n"));

	/*
	 * if (DeviceExtension->ulLogicalDriveCount == 0) {
	 * CreateMultipleLogRAID1Drives(DeviceExtension);
	 * CreateRAID0SingleDrives(DeviceExtension);
	 * CreateRAID0Drives(DeviceExtension);
	 * CreateRAID1Drives(DeviceExtension); }
	 */

	AR_DEBUG(AR_DEBUG_VANI_CODE,
	    ("logicaldrivecnt %ld\n", DeviceExtension->ulLogicalDriveCount));

	if (DeviceExtension->ulLogicalDriveCount) {

		for (LogDrvInd = 0;
		    LogDrvInd < DeviceExtension->ulLogicalDriveCount;
		    LogDrvInd++) {

			switch (DeviceExtension->LogicalDrive[LogDrvInd].RaidLevel) {
			case None:
				sprintf(raid_type, "%s", "NON-RAID");
				break;
			case Raid0:
				sprintf(raid_type, "%s", "RAID0");
				break;
			case Raid1:
				sprintf(raid_type, "%s", "RAID1");
				break;
			case Raid10:
				sprintf(raid_type, "%s", "RAID10");
				break;
			case SpareDrivePool:
				sprintf(raid_type, "%s", "SPAREPOOL");
				break;
			default:
				sprintf(raid_type, "%s", "UNKNOWN");
				break;

			}

			switch (DeviceExtension->LogicalDrive[LogDrvInd].Status) {

			case LDS_Online:
				sprintf(raid_status, "%s", "ONLINE");
				break;
			case LDS_Critical:
				sprintf(raid_status, "%s", "CRITICAL");
				break;
			case LDS_Degraded:
				sprintf(raid_status, "%s", "DEGRADED");
				break;
			case LDS_Deleted:
				sprintf(raid_status, "%s", "DELETED");
				break;
			case LDS_OffLine:
			default:
				sprintf(raid_status, "%s", "OFFLINE");
				break;
			}

			printk_Rlsi("LD %d %s status = %s \
				sectors = %ld capacity = %ld MB drives = %ld\n",
			    LogDrvInd,
			    raid_type,
			    raid_status,
			    (ULONG) DeviceExtension->LogicalDrive[LogDrvInd].Sectors - 1,
			    (ULONG) ((DeviceExtension->LogicalDrive[LogDrvInd].Sectors / 2) / 1024),
			    (ULONG) DeviceExtension->LogicalDrive[LogDrvInd].PhysicalDriveCount);

		}


		InitDriveFeatures(DeviceExtension);

	} else {
		return AR_FALSE;
	}

	return AR_TRUE;

}

/*
 * -------------------------------------------------------------------
 * Description ????
 * Arguments
 * Return Value
 * -------------------------------------------------------------------
 */

BOOLEAN
FindDevices(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR Channel)
{
	PIDE_REGISTERS_1 baseIoAddress1 =
	DeviceExtension->Channel[Channel].BaseIoAddress1;
	BOOLEAN deviceResponded = AR_FALSE;
	UCHAR deviceNumber;
	UCHAR i;
	UCHAR targetId;
	PPHYSICAL_DRIVE_INFO pPhysInfo;
	BOOLEAN IssueIdent ;


	AR_DEBUG(AR_DEBUG_FINDDEV, ("find_ devices Called\n"));

	/*
	 * Clear expecting interrupt flags.
	 */

	DeviceExtension->Channel[Channel].ExpectingInterrupt = 0;

	/*
	 * Search for devices.
	 */

	for (deviceNumber = 0; deviceNumber < 2; deviceNumber++) {
		 IssueIdent = TRUE;
		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("Searching Channel %d deviceNumber %d\n",
			Channel, deviceNumber));

		targetId = (Channel << 1) + deviceNumber;

 		/* Vani 1120 SATA : No use in making this check after checking for IsDrivePresent */
                /* serial ATA has only one drive per channel */
		if (DeviceExtension->ucCardType == SIL_3112)
		 {
		 	if(deviceNumber > 0)
			{
				IssueIdent = FALSE;
				break;
			}
                        /*
			 * perform the check for Serial ATA Sstatus register to determine if drive is present
			 * Only if Drive is present issue identify
			 */
			IssueIdent = Sata_Drv_Ready(DeviceExtension,  Channel);


 		}
		else if (!IsDrivePresent(DeviceExtension, targetId))
		{
			IssueIdent = FALSE;
			continue;
		}

		/*
		 * Issue IDE Identify. If an Atapi device is actually
		 * present, the signature will be asserted, and the drive
		 * will be recognized as such.
		 */
		 if(TRUE == IssueIdent )
		 {
			if (IssueIdentify(DeviceExtension, deviceNumber, Channel,
				IDE_COMMAND_IDENTIFY)) {
				/* IDE drive found */
				AR_DEBUG(AR_DEBUG_FINDDEV,
				("Find Devices: TID %d is IDE\n", targetId));

				DeviceExtension->PhysicalDrive[targetId].DeviceFlags |=
				DFLAGS_DEVICE_PRESENT;

				deviceResponded = AR_TRUE;

				/* Indicate IDE - not ATAPI device. */

				DeviceExtension->PhysicalDrive[targetId].DeviceFlags &=
				~DFLAGS_ATAPI_DEVICE;
			} else {
				/* Apparently no device */
				continue;
			}
		}
	}

	for (i = 0; i < MAX_CHANNELS_PER_CONTROLLER; i++) {
		targetId = (Channel << 1) + i;
		pPhysInfo = &DeviceExtension->PhysicalDrive[targetId];

		if (pPhysInfo->DeviceFlags & DFLAGS_DEVICE_PRESENT) {
			/* Initialize LastDriveFed. */
			DeviceExtension->Channel[Channel].LastDriveFed = i;

			if (!(pPhysInfo->DeviceFlags &
				DFLAGS_ATAPI_DEVICE) && deviceResponded) {
				/*
				 * This hideous hack is to deal with ESDI
				 * devices that return garbage geometry in
				 * the IDENTIFY data. This is ONLY for the
				 * crashdump environment as these are ESDI
				 * devices.
				 */

				if (pPhysInfo->IdentifyData.SectorsPerTrack == 0x35 &&
				    pPhysInfo->IdentifyData.NumberOfHeads == 0x07) {

					/*
					 * AR_DEBUG(AR_DEBUG_FINDDEV,
					 * (KERN_DEBUG "megaide: Find
					 * Devices: Found nasty Compaq
					 * ESDI!\n"));
					 */

					/*
					 * Change these values to something
					 * reasonable.
					 */

					pPhysInfo->IdentifyData.SectorsPerTrack = 0x34;
					pPhysInfo->IdentifyData.NumberOfHeads = 0x0E;
				}
				if (pPhysInfo->IdentifyData.SectorsPerTrack == 0x35 &&
				    pPhysInfo->IdentifyData.NumberOfHeads == 0x0F) {

					/*
					 * AR_DEBUG(AR_DEBUG_FINDDEV,
					 * (KERN_DEBUG "megaide: Find
					 * Devices: Found nasty Compaq
					 * ESDI!\n"));
					 */

					/*
					 * Change these values to something
					 * reasonable.
					 */
					pPhysInfo->IdentifyData.SectorsPerTrack = 0x34;
					pPhysInfo->IdentifyData.NumberOfHeads = 0x0F;
				}
				if (pPhysInfo->IdentifyData.SectorsPerTrack == 0x36 &&
				    pPhysInfo->IdentifyData.NumberOfHeads == 0x07) {

					/* AR_DEBUG(AR_DEBUG_FINDDEV, */
					/*
					 * (KERN_DEBUG "megaide: Find
					 * Devices: Found nasty UltraStor
					 * ESDI!\n"));
					 */

					/*
					 * Change these values to something
					 * reasonable.
					 */
					pPhysInfo->IdentifyData.SectorsPerTrack = 0x3F;
					pPhysInfo->IdentifyData.NumberOfHeads = 0x10;
				}
			}
		}
	}

	for (i = 0; i < 2; i++) {

		targetId = (Channel << 1) + i;
		pPhysInfo = &DeviceExtension->PhysicalDrive[targetId];

		if (pPhysInfo->DeviceFlags & DFLAGS_DEVICE_PRESENT) {
			/* select first available drive */
			SELECT_DEVICE(
			    DeviceExtension->Channel[Channel].base1_resource,
			    baseIoAddress1, targetId);
			break;
		}
	}

#ifdef DEBUG_ON
	if (!deviceResponded)
		AR_DEBUG(AR_DEBUG_FINDDEV,
		    ("No devices Found in Chan %d\n", Channel));
#endif

	return deviceResponded;
}				/* Find Devices() */

/*
 * -------------------------------------------------------------------
 * Description : ??? Input Parameters: Output Parameters:
 * -------------------------------------------------------------------
 */

BOOLEAN
GetIoMode(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR TargetId)
{
	PIDENTIFY_DATA pstCapabilities;
	ULONG sectors;
	BOOLEAN success = AR_TRUE;
	PPHYSICAL_DRIVE_INFO pPhysInfo =
	&DeviceExtension->PhysicalDrive[TargetId];

	pstCapabilities = &(pPhysInfo->IdentifyData);
	/*
	 * Calculate the total number of addressable sectors using CHS
	 * convention.
	 */
	sectors = pstCapabilities->NumberOfCylinders *
	    pstCapabilities->NumberOfHeads *
	    pstCapabilities->SectorsPerTrack;

	AR_DEBUG(AR_DEBUG_INITDEV, ("sectors%ld \n", sectors));

	if (pstCapabilities->TranslationFieldsValid &
	    IDENTIFY_FAST_TRANSFERS_SUPPORTED) {

#if 0				/* def DEBUG_ON */
		PrintCapabilities(pstCapabilities, TargetId);
#endif				/* DEBUG_ON */


		if ((pstCapabilities->Capabilities &
			IDENTIFY_CAPABILITIES_LBA_SUPPORTED)) {
			pPhysInfo->DeviceFlags |= DFLAGS_LBA;

			/*
			 * Get the total number of LBA-mode addressable
			 * sectors.
			 */
			sectors = pstCapabilities->UserAddressableSectors;

			AR_DEBUG(AR_DEBUG_INITDEV,
			    ("GetIoMode: TID %d supports LBA\n", TargetId));
		}
	}
	AR_DEBUG(AR_DEBUG_INITDEV,
	    ("##sectors%ld \n", sectors));

	DeviceExtension->PhysicalDrive[TargetId].NumberOfBlocks = sectors;

	/* ++manoj 0122 48 bit addressing support */
	if (IS_48_BIT_ENABLED(TargetId) &&
	    pstCapabilities->AddressableSectors48bit[0]) {
		pPhysInfo->NumberOfBlocks48bit =
			pstCapabilities->AddressableSectors48bit[0];
	}
	/* --manoj 0122 48 bit addressing support */

	GetTransferMode(DeviceExtension, TargetId);

	return (success);

}				/* Get IoMode() */

/*
 * -------------------------------------------------------------------
 * Description : ??? Input Parameters: Output Parameters:
 * -------------------------------------------------------------------
 */

BOOLEAN
GetTransferMode(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR TargetId)
{
	PIDENTIFY_DATA pstCapabilities;
	PPHYSICAL_DRIVE_INFO pPhysInfo =
	&DeviceExtension->PhysicalDrive[TargetId];
	pstCapabilities = &(pPhysInfo->IdentifyData);

	/* By default assume this drive is PioMode 0 */
	pPhysInfo->TransferMode = MP_PioMode0;

	if (pstCapabilities->AdvancedPioModes != 0) {

		if (pstCapabilities->AdvancedPioModes & IDENTIFY_PIO_MODE4) {
			pPhysInfo->TransferMode = MP_PioMode4;
		} else {
			if (pstCapabilities->AdvancedPioModes &
			    IDENTIFY_PIO_MODE3) {
				pPhysInfo->TransferMode = MP_PioMode3;
			} else {
				pPhysInfo->TransferMode = MP_PioMode0;
			}
		}
	}
	if (pstCapabilities->MultiWordDmaActive) {
		switch (pstCapabilities->MultiWordDmaActive) {
		case 1:
			pPhysInfo->TransferMode = MP_DmaMode0;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_DMA;
			break;

		case 2:
			pPhysInfo->TransferMode = MP_DmaMode1;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_DMA;
			break;

		case 4:
			pPhysInfo->TransferMode = MP_DmaMode2;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_DMA;
			break;
		}
	}
	if (pstCapabilities->UltraDmaActive) {

		switch (pstCapabilities->UltraDmaActive) {
		case 1:
			pPhysInfo->TransferMode = MP_UdmaMode0;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 2:
			pPhysInfo->TransferMode = MP_UdmaMode1;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 4:
			pPhysInfo->TransferMode = MP_UdmaMode2;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 8:
			pPhysInfo->TransferMode = MP_UdmaMode3;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 0x10:
			pPhysInfo->TransferMode = MP_UdmaMode4;
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			break;

		case 0x20:
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			/*
			 * manoj 1213 == to >= to take care of higher Udma
			 * speed
			 */
			if (Udma100 >= DeviceExtension->ControllerSpeed) {
				pPhysInfo->TransferMode = MP_UdmaMode5;
			} else {
				pPhysInfo->TransferMode = MP_UdmaMode4;
			}
			break;

			/* ++manoj 1213 Ultra DMA 6 support */
		case 0x40:
			pPhysInfo->DeviceFlags |= DFLAGS_USE_UDMA;
			if (Udma133 == DeviceExtension->ControllerSpeed) {
				pPhysInfo->TransferMode = MP_UdmaMode6;
			} else if (Udma100 == DeviceExtension->ControllerSpeed) {
				pPhysInfo->TransferMode = MP_UdmaMode5;
			} else {
				pPhysInfo->TransferMode = MP_UdmaMode4;
			}
			break;
			/* --manoj 1213 Ultra DMA 6 support */
		}
	}
	return AR_TRUE;

}				/* GetTransferMode() */

/*
 * program_pci = 1 means to program the pci space program_pci = 0 means not
 * to program the pci space
 */
BOOLEAN
SetTransferMode(
    PHW_DEVICE_EXTENSION DeviceExtension,
    UCHAR TargetId,
	MP_TRANSFER_MODE TransferMode)
{
	UCHAR channel = (TargetId >> 1);
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PBM_REGISTERS BaseBmAddress;
	PPHYSICAL_DRIVE_INFO pPhysInfo =
	&DeviceExtension->PhysicalDrive[TargetId];
	PCHANNEL pChannel =
	&DeviceExtension->Channel[channel];
	UCHAR statusByte = 0;

	UCHAR transfer_mode;

	/* manoj 08022002 to program drive as in enquiry Data */
	UCHAR ucEnqTransferModeTemp =
		GetInquiryTransferMode(DeviceExtension, TargetId);

	if (ucEnqTransferModeTemp >= MP_UdmaMode0) {
		pPhysInfo->TransferMode =
			min(ucEnqTransferModeTemp, pPhysInfo->TransferMode);
		TransferMode = ucEnqTransferModeTemp;
	}
	if (PCI_DEVICE_ID_CMD_3112 == GET_DEVICE_ID(DeviceExtension)) {
		/* Vani: 12/07 this change is applicable only Maxtor drives */
		if (!strncmp(DeviceExtension->PhysicalDrive[TargetId].IdentifyData.ModelNumber,
		           "aMtxro", 6))
		{
			printk_Rlsi ("megaide: FIX Drv[%d] is a Maxtor SATA drive, Programing to UDMA5 mode\n", TargetId);
			if (MP_UdmaMode6 == TransferMode) {
				TransferMode = MP_UdmaMode5;
			}
		}
	}
	switch (TransferMode) {

	case MP_DmaMode0:
		transfer_mode = STM_MULTIWORD_DMA(0);
		break;

	case MP_DmaMode1:
		transfer_mode = STM_MULTIWORD_DMA(1);
		break;

	case MP_DmaMode2:
		transfer_mode = STM_MULTIWORD_DMA(2);
		break;

	case MP_UdmaMode0:
		transfer_mode = STM_UDMA(0);
		break;

	case MP_UdmaMode1:
		transfer_mode = STM_UDMA(1);
		break;

	case MP_UdmaMode2:
		transfer_mode = STM_UDMA(2);
		break;

	case MP_UdmaMode3:
		transfer_mode = STM_UDMA(3);
		break;

	case MP_UdmaMode4:
		transfer_mode = STM_UDMA(4);
		break;

	case MP_UdmaMode5:
		transfer_mode = STM_UDMA(5);
		break;

	case MP_UdmaMode6:
		transfer_mode = STM_UDMA(6);
		break;

	default:
		/* CURRENTLY we are not programming drives to PIO mode */
		return FALSE;
		break;
	}


	baseIoAddress1 = (PIDE_REGISTERS_1)pChannel->BaseIoAddress1;
	baseIoAddress2 = (PIDE_REGISTERS_2)pChannel->BaseIoAddress2;
	BaseBmAddress = (PBM_REGISTERS)pChannel->BaseBmAddress;

	SELECT_DEVICE(
	    pChannel->base1_resource, baseIoAddress1, TargetId);

	/* Program Drives */

	ScsiPortWritePortUchar(pChannel->base1_resource,
	    (unsigned)((PUCHAR) baseIoAddress1 + 1), FEATURE_SET_TRANSFER_MODE);
	ScsiPortWritePortUchar(pChannel->base1_resource,
	    (unsigned)&(baseIoAddress1->SectorCount), transfer_mode);
	ScsiPortWritePortUchar(pChannel->base1_resource,
	    (unsigned)&(baseIoAddress1->Command), IDE_COMMAND_SET_FEATURES);
	WAIT_ON_BASE_BUSY(pChannel->base1_resource,
	    baseIoAddress1, statusByte);

	if ((statusByte & IDE_STATUS_ERROR) ||
	    (statusByte & IDE_STATUS_BUSY)) {
		printk_Rlsi("set features on TID %d failed.\n", TargetId);
		return FALSE;
	}
	/* Program PCI space */

	gaCardSpecificFunctions[DeviceExtension->ucCardType].pfnProgramChip(
	    DeviceExtension, TargetId, (UCHAR)TransferMode);

	DeviceExtension->PhysicalDrive[TargetId].TransferMode = TransferMode;

	return TRUE;

}

static unsigned char auc64xUdmaModeValues[2][9] = {
	{0x0, 0x0, 0x0, 0x31, 0x21, 0x11, 0x25, 0x15, 0x05},	/* for Master */
	{0x0, 0x0, 0x0, 0xc2, 0x82, 0x42, 0x8a, 0x4a, 0x0a}	/* for Slave */
};

/* MDMA0, MDMA1, MDMA2, UDMA0, UMDA1, UDMA2, UDMA3, UDMA4, UDMA5 */

BOOLEAN
ProgramDriveFor64xChip(
    PHW_DEVICE_EXTENSION DeviceExtension,
    UCHAR ucTargetId, UCHAR TransferMode)
{
	UCHAR channel = (ucTargetId >> 1);
	UCHAR ucOrMask, ucAndMask, ucValue, device_select;
	PBM_REGISTERS bmBase;

	if (TransferMode > MP_UdmaMode5)
		return FALSE;

	device_select = (ucTargetId & 0x01);
	ucOrMask = auc64xUdmaModeValues[device_select][TransferMode - MP_DmaMode0];

	if (device_select)
		ucAndMask = 0x35;	/* if slave do not touch master bits */
	else
		ucAndMask = 0xca;	/* if master do not touch slave bits */

	bmBase = DeviceExtension->Channel[channel].BaseBmAddress;
	ucValue =
		ScsiPortReadPortUchar(
		DeviceExtension->Channel[channel].bmaddr_resource,
		(unsigned)((PUCHAR) bmBase + 3));

	ucValue &= ucAndMask;
	ucValue |= ucOrMask;
	ScsiPortWritePortUchar(
		DeviceExtension->Channel[channel].bmaddr_resource,
		(unsigned)((PUCHAR) bmBase + 3), ucValue);

	return (TRUE);

}				/* end ProgramDriveFor64xChip() */


/*
 * Please refer to page 129 in Sil0680 Data Sheet Rev 1.1
 * for explanation of these values
 */
static unsigned char auc68xUdmaModeValues[7] = 
	{0x0F, 0x0B, 0x07, 0x05, 0x03, 0x02, 0x01};
static unsigned short auc68xDmaModeValues[3] = 
	{0x2208, 0x10C2, 0x10C1};
static unsigned short auc68xPIOModeValues[5] = 
	{0x328A, 0x2283, 0x1104, 0x10C3, 0x10C1};
static unsigned short auc68xTFTimingValues[5] = 
	{0x328A, 0x2283, 0x1281, 0x10C3, 0x10C1};


BOOLEAN
ProgramDriveFor68xChip(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucTargetId, UCHAR ucMode)
{
	u_char ucByteReg = 0;
	u_char tf_mode, pio_mode, mdma_mode, udma_mode;
	u_char ucReg80 = 0, ucReg84 = 0;
	MP_DRIVE_ID drive_id = (MP_DRIVE_ID) ucTargetId;

	/* program IDE CLK Select to 133 MHz */
	pci_read_config_byte_Rlsi(DeviceExtension, SYS_CONF_STAT_CMD_REG680 + 2, &ucByteReg);
	ucByteReg = (ucMode >= MP_UdmaMode0) ? (ucByteReg | 0x10) : (ucByteReg & 0xEF);
	pci_write_config_byte_Rlsi(DeviceExtension, SYS_CONF_STAT_CMD_REG680 + 2, ucByteReg);

	/* PIO2 = MDMA0 */
	/* PIO3 = MDMA1 */
	/* PIO4 = MDMA2(and UDMAx) */

	/* program TF Timing */
	tf_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) : (MP_DmaMode2 - MP_DmaMode0);
	tf_mode += 2;

	/* program PIO Timing */
	pio_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) : (MP_DmaMode2 - MP_DmaMode0);
	pio_mode += 2;

	/* program mdma timing */
	mdma_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) : (MP_DmaMode2 - MP_DmaMode0);

	/* program dma timing */
	udma_mode = (ucMode >= MP_UdmaMode0) ? (ucMode - MP_UdmaMode0) : 6;

	/* set TF Timing */
	if ((ucTargetId > 1) == 0) {
		pci_write_config_byte_Rlsi(DeviceExtension, IDE0_TFT_CONF_STAT_REG680 + 2,
		    auc68xTFTimingValues[tf_mode]);
	} else {
		pci_write_config_byte_Rlsi(DeviceExtension, IDE1_TFT_CONF_STAT_REG680 + 2,
		    auc68xTFTimingValues[tf_mode]);
	}

	/* get Transfer Mode registers values */
	/* bits 1:0 - Device0 { 2=mdma, 3=udma } */
	/* bits 5:4 - Device1 {2=mdma, 3=udma } */
	pci_read_config_byte_Rlsi(DeviceExtension,
		DATA_TRFR_MODE_IDE0_REG680, &ucReg80);
	pci_read_config_byte_Rlsi(DeviceExtension,
		DATA_TRFR_MODE_IDE1_REG680, &ucReg84);

	ucReg80 &= ((ucTargetId & 1) ? 0xCF : 0xFC);
	ucReg84 &= ((ucTargetId & 1) ? 0xCF : 0xFC);

	if (ucMode >= MP_UdmaMode0) {
		ucReg80 |= ((ucTargetId & 1) ? 0x30 : 0x03);
		ucReg84 |= ((ucTargetId & 1) ? 0x30 : 0x03);
	} else if (ucMode >= MP_DmaMode0) {
		ucReg80 |= ((ucTargetId & 1) ? 0x20 : 0x02);
		ucReg84 |= ((ucTargetId & 1) ? 0x20 : 0x02);
	}
	switch (drive_id) {
	case PrimaryMaster:	/* primary master */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE0_PIO_TIMING_REG680,
		    auc68xPIOModeValues[pio_mode]);	/* set PIO Timing */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE0_DMA_TIMING_REG680,
		    auc68xDmaModeValues[mdma_mode]);	/* set mdma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			IDE0_UDMA_TIMING_REG680,
		    auc68xUdmaModeValues[udma_mode]);	/* set udma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			DATA_TRFR_MODE_IDE0_REG680, ucReg80);	/* set tran mode */

		break;


	case PrimarySlave:	/* primary slave */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE0_PIO_TIMING_REG680 + 2,
		    auc68xPIOModeValues[pio_mode]);	/* set PIO Timing */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE0_DMA_TIMING_REG680 + 2,
		    auc68xDmaModeValues[mdma_mode]);	/* set mdma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			IDE0_UDMA_TIMING_REG680 + 2,
		    auc68xUdmaModeValues[udma_mode]);	/* set udma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			DATA_TRFR_MODE_IDE0_REG680, ucReg80);	/* set tran mode */

		break;


	case SecondaryMaster:	/* secondary slave */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE1_PIO_TIMING_REG680,
		    auc68xPIOModeValues[pio_mode]);	/* set PIO Timing */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE1_DMA_TIMING_REG680,
		    auc68xDmaModeValues[mdma_mode]);	/* set mdma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			IDE1_UDMA_TIMING_REG680,
		    auc68xUdmaModeValues[udma_mode]);	/* set udma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			DATA_TRFR_MODE_IDE1_REG680, ucReg84);	/* set tran mode */

		break;



	case SecondarySlave:	/* secondary slave */


		pci_write_config_word_Rlsi(DeviceExtension,
			IDE1_PIO_TIMING_REG680 + 2,
		    auc68xPIOModeValues[pio_mode]);	/* set PIO Timing */

		pci_write_config_word_Rlsi(DeviceExtension,
			IDE1_DMA_TIMING_REG680 + 2,
		    auc68xDmaModeValues[mdma_mode]);	/* set mdma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			IDE1_UDMA_TIMING_REG680 + 2,
		    auc68xUdmaModeValues[udma_mode]);	/* set udma timing */

		pci_write_config_byte_Rlsi(DeviceExtension,
			DATA_TRFR_MODE_IDE1_REG680, ucReg84);	/* set tran mode */

		break;

	default:

		break;


	}


	return (TRUE);
}				/* end ProgramDriveFor68xChip() */


/*
 * Please refer to page 19 to 23 in CSB5 Data Sheet Ver 2.0a for explanation
 * of these values
 */
static unsigned short aucCSB5PIOTimingValues[5] = {0x50, 0x47, 0x34, 0x22, 0x20};
static unsigned short aucCSB5DmaTimingModeValues[3] = {0x77, 0x21, 0x20};
static unsigned short aucCSB5UdmaModeValuesMaster[7] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x05};
static unsigned short aucCSB5UdmaModeValuesSlave[7] = {0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x50};

BOOLEAN
ProgramDriveForCSB5Chip(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucTargetId, UCHAR ucMode)
{
	UCHAR ucReg = 0;
	u_char pio_mode, mdma_mode, udma_mode;
	MP_DRIVE_ID drive_id = (MP_DRIVE_ID)ucTargetId;

	/* PIO2 = MDMA0 */
	/* PIO3 = MDMA1 */
	/* PIO4 = MDMA2(and UDMAx) */

	/* program PIO Timing */
	pio_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) : (MP_DmaMode2 - MP_DmaMode0);
	pio_mode += 2;

	/* program mdma timing */
	mdma_mode = (ucMode < MP_UdmaMode0) ?
		(ucMode - MP_DmaMode0) : (MP_DmaMode2 - MP_DmaMode0);

	/* program dma timing */
	udma_mode = (ucMode >= MP_UdmaMode0) ?
		(ucMode - MP_UdmaMode0) : 6;

	switch (drive_id) {
	case PrimaryMaster:	/* primary master */
		{
			/* PIO Timing Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x41,
				aucCSB5PIOTimingValues[pio_mode]);
			/* DMA Timing Mode Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x45,
				aucCSB5DmaTimingModeValues[mdma_mode]);
			/* PIO Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x4A, &ucReg);
			ucReg |= 0x04;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x4A, ucReg);
			/* UDMA Control Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x54, &ucReg);
			ucReg |= 0x01;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x54, ucReg);
			/* ATA Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x56, &ucReg);
			ucReg |= aucCSB5UdmaModeValuesMaster[udma_mode];
			pci_write_config_byte_Rlsi(DeviceExtension, 0x56, ucReg);
			break;
		}

	case PrimarySlave:	/* primary slave */
		{

			/* PIO Timing Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x40,
				aucCSB5PIOTimingValues[pio_mode]);
			/* DMA Timing Mode Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x44,
				aucCSB5DmaTimingModeValues[mdma_mode]);
			/* PIO Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x4A, &ucReg);
			ucReg |= 0x40;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x4A, ucReg);
			/* UDMA Control Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x54, &ucReg);
			ucReg |= 0x02;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x54, ucReg);

			/* ATA Mode Register FIX */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x56, &ucReg);
			ucReg |= aucCSB5UdmaModeValuesSlave[udma_mode];
			pci_write_config_byte_Rlsi(DeviceExtension, 0x56, ucReg);
			break;
		}

	case SecondaryMaster:	/* secondary slave */
		{
			/* PIO Timing Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x43,
				aucCSB5PIOTimingValues[pio_mode]);
			/* DMA Timing Mode Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x47,
				aucCSB5DmaTimingModeValues[mdma_mode]);
			/* PIO Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x4B, &ucReg);
			ucReg |= 0x04;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x4B, ucReg);
			/* UDMA Control Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x54, &ucReg);
			ucReg |= 0x04;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x54, ucReg);
			/* ATA Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x57, &ucReg);
			ucReg |= aucCSB5UdmaModeValuesMaster[udma_mode];
			pci_write_config_byte_Rlsi(DeviceExtension, 0x57, ucReg);
			break;
		}

	case SecondarySlave:	/* secondary slave */
		{

			/* PIO Timing Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x42,
				aucCSB5PIOTimingValues[pio_mode]);
			/* DMA Timing Mode Registers */
			pci_write_config_byte_Rlsi(DeviceExtension, 0x46,
				aucCSB5DmaTimingModeValues[mdma_mode]);
			/* PIO Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x4B, &ucReg);
			ucReg |= 0x40;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x4B, ucReg);
			/* UDMA Control Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x54, &ucReg);
			ucReg |= 0x08;
			pci_write_config_byte_Rlsi(DeviceExtension, 0x54, ucReg);
			/* ATA Mode Register */
			pci_read_config_byte_Rlsi(DeviceExtension, 0x57, &ucReg);
			ucReg |= aucCSB5UdmaModeValuesSlave[udma_mode];
			pci_write_config_byte_Rlsi(DeviceExtension, 0x57, ucReg);
			break;
		}

	default:
		AR_DEBUG(AR_DEBUG_WARNING,
			("wrong drive ID in ProgramDriveForCSB5Chip!!!!\n"));
		break;
	}			/* switch (drive_id) */
	return TRUE;
}



/*
 * -------------------------------------------------------------------
 * Description ????
 * Arguments
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_detach(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);
	int ulPhyDrvInd;
	int ControllerNumber;

	ControllerNumber = DeviceExtension->ucControllerId;

	CleanUpAdapter(DeviceExtension);

	DisableInterrupts(DeviceExtension);

	for (ulPhyDrvInd = 0;
		ulPhyDrvInd < MAX_DRIVES_PER_CONTROLLER;
		ulPhyDrvInd++) {
		if (DeviceExtension->PhysicalDrive[ulPhyDrvInd].bEnableRwCache)
			FlushCache(DeviceExtension);
	}
	EnableInterrupts(DeviceExtension);

	untimeout(mide_periodic, DeviceExtension,
		gastCardInfo[DeviceExtension->ucControllerId].mide_timeout);

	megaide_cam_detach(DeviceExtension);

	if (DeviceExtension->isr_handle != NULL)
		bus_teardown_intr(dev,
		(struct resource *)DeviceExtension->isr_resource,
		DeviceExtension->isr_handle);

	ReleaseOurIOPorts(DeviceExtension);
#ifdef	SMPNG
	mtx_destroy(gastCardInfo[DeviceExtension->ucControllerId].midemtx);
#endif
	return (0);
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

VOID
CleanUpAdapter(PHW_DEVICE_EXTENSION DeviceExtension)
{

	FreePtrsForDevExtension(DeviceExtension);
	FreeSglPtrsForChannels(DeviceExtension);

}

BOOLEAN
Is80PinCable(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucChannelId)
{
	return gaCardSpecificFunctions[DeviceExtension->ucCardType].
		pfnIs80PinCable(DeviceExtension, ucChannelId);
}

BOOLEAN
Is64x80PinCable(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucChannelId)
{
	UCHAR ucIs80PinCable;
	pci_read_config_byte_Rlsi(DeviceExtension, 0x79, &ucIs80PinCable);
	if (ucChannelId) {	/* secondary channel */
		ucIs80PinCable = (ucIs80PinCable & 0x02) >> 1;
	} else {		/* primary channel */
		ucIs80PinCable = (ucIs80PinCable & 0x01);
	}

	if (!(ucIs80PinCable & 0x01)) {
		printk_Rlsi("Is68x80PinCablenot an 80 pin cable!!!\n");
	}
	return (ucIs80PinCable);
}

BOOLEAN
Is68x80PinCable(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucChannelId)
{
	UCHAR ucIs80PinCable;

	if (ucChannelId) {	/* secondary channel */
		pci_read_config_byte_Rlsi(DeviceExtension, 0xB0, &ucIs80PinCable);
	} else {		/* primary channel */
		pci_read_config_byte_Rlsi(DeviceExtension, 0xA0, &ucIs80PinCable);
	}

	if (!(ucIs80PinCable & 0x01)) {
		printk_Rlsi("Is68x80PinCablenot an 80 pin cable!!!\n");
	}
	return (ucIs80PinCable & 0x01);
}

BOOLEAN
IsCSB580PinCable(
	PHW_DEVICE_EXTENSION DeviceExtension,
	UCHAR ucChannelId)
{
	return 0x01;
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
BOOLEAN
FoundInterrupt(
	PHW_DEVICE_EXTENSION DeviceExtension,
	PUCHAR pulChannel)
{
	BOOLEAN blReturnValue;
	BOOLEAN FireCmd = FALSE;

	blReturnValue = gaCardSpecificFunctions
		[DeviceExtension->ucCardType].pfnFoundInterrupt
		(DeviceExtension, pulChannel);

#ifdef BP_SUPPORT
	if (!blReturnValue)
	{
	/*
		idebp_check_dock_status(DeviceExtension);
	*/
		if(DeviceExtension->PMState == NORMAL)
		{
			if(idebp_check_dock_status(DeviceExtension))
				FireCmd = TRUE;
		}
		if(mide_ProcessPMChange(DeviceExtension))
			FireCmd = TRUE;
		if(FireCmd == TRUE)
			FastIDEFireCommand(DeviceExtension);


	}
#endif

	return blReturnValue;
}				/* FoundInterrupt() */


BOOLEAN
Found64xInterrupt(
	PHW_DEVICE_EXTENSION DeviceExtension,
	PUCHAR pulChannel)
{
	PBM_REGISTERS baseBm;
	UCHAR statusByte;
	ULONG ulChannel;
	ULONG ulIntChannelValue = 0;

	/*
	 * We are going to keep as a Bit Map where if one is set in one
	 * particular channel number place that means interrupt has been
	 * raised from that channel.. if no bit was set that means it is from
	 * a different card other than our PCI Card
	 */

	/*
	 * This interrupt status will show the interrupt on two
	 * channels...(ie., per PCI Space
	 */
	for (ulChannel = 0; ulChannel < MAX_CHANNELS_PER_CONTROLLER; ulChannel += 2) {
		baseBm = DeviceExtension->Channel[ulChannel].BaseBmAddress;

		statusByte = ScsiPortReadPortUchar(
		    DeviceExtension->Channel[ulChannel].bmaddr_resource,
		    (unsigned)&(baseBm->Reserved));

		if (statusByte & PRIMARY_CHANNEL_INTERRUPT) {
			ulIntChannelValue |= (1 << ulChannel);
		}
		if (statusByte & SECONDARY_CHANNEL_INTERRUPT) {
			ulIntChannelValue |= (1 << (ulChannel + 1));
		}
		if ((statusByte & PRIMARY_CHANNEL_INTERRUPT) ||
		    (statusByte & SECONDARY_CHANNEL_INTERRUPT)) {

			/*
			 * let us clear the interrupts lets make sure Read
			 * Multiple is enabled,  remember that this register
			 * is mapped to MRDMODE bit 0-1 : = 00 Memory Read =
			 * 01 Read Multiple = 1x Read Line
			 */
			statusByte |= 0x01;
			/*
			 * ScsiPortWritePortUchar((&(baseBm->Reserved)),
			 * statusByte);
			 */
		}
	}			/* fo(ucChannel */

	*pulChannel = ulIntChannelValue;

	if (ulIntChannelValue)
		return TRUE;
	else
		return FALSE;
}				/* Found64xInterrupt() */


BOOLEAN
Found68xInterrupt(
	PHW_DEVICE_EXTENSION DeviceExtension,
	PUCHAR pulChannel)
{
	PBM_REGISTERS baseBm;
	UCHAR statusByte;
	ULONG ulChannel = 0;
	ULONG ulIntChannelValue = 0;

	baseBm = DeviceExtension->Channel[ulChannel].BaseBmAddress;

	pci_read_config_byte_Rlsi(DeviceExtension, (IDE0_TFT_CONF_STAT_REG680 + 1), &statusByte);
	if (statusByte & PRIMARY_CHAN_INTERRUPT_680) {
		ulIntChannelValue |= (1 << ulChannel);
		/* Clear Interrupt */
		/*
		 * statusByte =
		 * ScsiPortReadPortUchar(
		 *	&(DeviceExtension->Channel[0].BaseIoAddress1->Command));
		 */
	}
	statusByte = 0;
	pci_read_config_byte_Rlsi(DeviceExtension, (IDE1_TFT_CONF_STAT_REG680 + 1), &statusByte);
	if (statusByte & SECONDARY_CHAN_INTERRUPT_680) {
		AR_DEBUG(AR_DEBUG_INTERUPT, ("Interrupt in IDE1\n"));
		ulIntChannelValue |= (1 << (ulChannel + 1));
		/* Clear Interrupt */
		/*
		 * statusByte =
		 * ScsiPortReadPortUchar(
		 *	&(DeviceExtension->Channel[1].BaseIoAddress1->Command));
		 */
	}
	/* manoj FIXME.. Not doing read multile setting */

	*pulChannel = ulIntChannelValue;

	if (ulIntChannelValue)
		return TRUE;
	else {
		return FALSE;
	}
}				/* Found68xInterrupt() */


BOOLEAN
FoundCSB5Interrupt(
	PHW_DEVICE_EXTENSION DeviceExtension,
	PUCHAR pulChannel)
{
	PBM_REGISTERS baseBm;
	UCHAR statusByte;
	ULONG ulChannel = 0;
	ULONG ulIntChannelValue = 0;

	baseBm = DeviceExtension->Channel[ulChannel].BaseBmAddress;


	statusByte = ScsiPortReadPortUchar(
	    DeviceExtension->Channel[ulChannel].bmaddr_resource,
	    (unsigned)(baseBm + 0x02));

	if (statusByte & CHANNEL_INTERRUPT_SERVERWORKS) {
		ulIntChannelValue |= (1 << ulChannel);
	}
	statusByte = 0;
	statusByte = ScsiPortReadPortUchar(
	    DeviceExtension->Channel[ulChannel].bmaddr_resource,
	    (unsigned)(baseBm + 0x0A));

	if (statusByte & CHANNEL_INTERRUPT_SERVERWORKS) {
		ulIntChannelValue |= (1 << (ulChannel + 1));
	}
	*pulChannel = ulIntChannelValue;

	if (ulIntChannelValue)
		return TRUE;
	else
		return FALSE;
}				/* FoundCSB5Interrupt() */


BOOLEAN
PostCommandToDrive(PHW_DEVICE_EXTENSION DeviceExtension,
    ULONG ulTargetId,
    ULONG ulStartBlock,
    ULONG ulBlockCount,
    UCHAR ucCmd,
    ULONG ulSglPhysicalAddress
)
{
	ULONG ulChannelId = ulTargetId >> 1;
	PCHANNEL pChannel = &(DeviceExtension->Channel[ulChannelId]);
	UCHAR statusByte, ucSectorNumber, ucCylinderLow, ucCylinderHigh,
	 bmStatus;
	UCHAR bmCommand = 0;
	PIDE_REGISTERS_1 baseIoAddress1 = pChannel->BaseIoAddress1;
	PBM_REGISTERS bmBase = pChannel->BaseBmAddress;
	UCHAR ucSectorNumberExt, b48bitFlag = FALSE;	/* manoj 1213 */



#ifdef TEST_RESET_PATH
	if (ulTestResetPath) {
		if (ulCommandCount > 100) {
			ulTestResetPath = FALSE;
			return SRB_STATUS_PENDING;	/* just fake it out to
							 * test the reset path */
		}
	}
#endif

	ucSectorNumber = (UCHAR)(ulStartBlock & 0x000000ff);
	ucCylinderLow = (UCHAR)((ulStartBlock & 0x0000ff00) >> 8);
	ucCylinderHigh = (UCHAR)((ulStartBlock & 0x00ff0000) >> 16);

	/* manoj 1213 FIXME, check start block */
	if (IS_48_BIT_ADDESSABLE(ulTargetId) /* && ulStartBlock >=
	        0x0FFFFFFF */ ) {

		b48bitFlag = TRUE;
		ucSectorNumberExt = (ulStartBlock & 0xff000000) >> 24;

		SELECT_LBA_DEVICE_48BIT(pChannel->base1_resource,
			baseIoAddress1, ulTargetId);

		/* Previous contents */
		/* manoj 1213 FIXME take care later */
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorCount, 0);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorNumber, ucSectorNumberExt);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderLow, 0);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderHigh, 0);

		/* New contents */
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorCount,
			(UCHAR) ulBlockCount);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorNumber,
			ucSectorNumber);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderLow,
			ucCylinderLow);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderHigh,
			ucCylinderHigh);
	} else {
		SELECT_LBA_DEVICE(DeviceExtension->Channel[ulChannelId].base1_resource, baseIoAddress1, ulTargetId, ulStartBlock);

		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorCount,
			(UCHAR) ulBlockCount);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->SectorNumber,
			ucSectorNumber);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderLow,
			ucCylinderLow);
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->CylinderHigh,
			ucCylinderHigh);
	}			/* manoj 48bit support */

	switch (ucCmd) {
	case READ:
	case READ_6:
	case READ_10:
		/* Here no need to handle WriteBack commands specially cause */
		/* this function is only used for BBW */
		/* manoj 1213 48 bit addressing */
		if (b48bitFlag == TRUE)
			ScsiPortWritePortUchar(pChannel->base1_resource,
				(unsigned)&baseIoAddress1->Command,
				IDE_COMMAND_READ_DMA_EXT);
		else
			ScsiPortWritePortUchar(pChannel->base1_resource,
				(unsigned)&baseIoAddress1->Command,
				IDE_COMMAND_READ_DMA);
		bmCommand = READ_TRANSFER;
		break;
	case WRITE:
	case WRITE_6:
	case WRITE_10:
		/* manoj 1213 48bit addressing */
		if (b48bitFlag == TRUE)
			ScsiPortWritePortUchar(pChannel->base1_resource,
				(unsigned)&baseIoAddress1->Command,
				IDE_COMMAND_WRITE_DMA_EXT);
		else
			ScsiPortWritePortUchar(pChannel->base1_resource,
				(unsigned)&baseIoAddress1->Command,
				IDE_COMMAND_WRITE_DMA);
		bmCommand = WRITE_TRANSFER;
		break;
	case VERIFY:
		ScsiPortWritePortUchar(pChannel->base1_resource,
			(unsigned)&baseIoAddress1->Command,
			IDE_COMMAND_VERIFY);
		pChannel->ExpectingInterrupt = IDE_PIO_INTERRUPT;
		return TRUE;	/* nothing else to do as verify command
				 * doesn't have any buffer transfers... */
		break;
	}

	bmStatus = ScsiPortReadPortUchar(
		pChannel->bmaddr_resource,
		(unsigned)&(bmBase->Status.AsUchar));

	/* Clear interrupt and error bits(if set) . */
	CLEAR_BM_INT(pChannel->bmaddr_resource, bmBase, statusByte);

	ScsiPortWritePortUlong(pChannel->bmaddr_resource,
		(unsigned)&(bmBase->SglAddress), ulSglPhysicalAddress);

	/* Start transfer. */
	ScsiPortWritePortUchar(pChannel->bmaddr_resource,
		(unsigned)&(bmBase->Command.AsUchar),
		(UCHAR)(bmCommand | START_TRANSFER));

	/* pChannel->ExpectingInterrupt = IDE_DMA_INTERRUPT; */

	return TRUE;
}

BOOLEAN
IsPostCommandToDriveSuccess(PHW_DEVICE_EXTENSION DeviceExtension,
    ULONG ulTargetId,
    ULONG ulStartBlock,
    ULONG ulBlockCount,
    UCHAR ucCmd,
    ULONG ulSglPhysicalAddress
)
{
	return gaCardSpecificFunctions[DeviceExtension->ucCardType].pfnIsPostCommandToDriveSuccess(DeviceExtension, ulTargetId,
	    ulStartBlock, ulBlockCount, ucCmd, ulSglPhysicalAddress);

}

BOOLEAN
IsPostCommandToDriveSuccess64x(PHW_DEVICE_EXTENSION DeviceExtension,
    ULONG ulTargetId,
    ULONG ulStartBlock,
    ULONG ulBlockCount,
    UCHAR ucCmd,
    ULONG ulSglPhysicalAddress
)
{
	UCHAR Channel = (ulTargetId >> 1);
	PCHANNEL pChannel;
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PBM_REGISTERS baseBm;
	UCHAR ucStatus = 0;
	UCHAR ucInterruptValue = (Channel ? SECONDARY_CHANNEL_INTERRUPT : PRIMARY_CHANNEL_INTERRUPT) | 0x1;

	pChannel = &(DeviceExtension->Channel[Channel]);
	baseIoAddress1 = (PIDE_REGISTERS_1)pChannel->BaseIoAddress1;
	baseIoAddress2 = (PIDE_REGISTERS_2)pChannel->BaseIoAddress2;
	baseBm = pChannel->BaseBmAddress;

	/* Clear interrupt and error bits(if set) . */
	/* CLEAR_BM_INT(baseBm, statusByte); */

	PostCommandToDrive(DeviceExtension,
		ulTargetId, ulStartBlock, ulBlockCount,
		ucCmd, ulSglPhysicalAddress);
	POLL_FOR_INTERRUPT(pChannel->bmaddr_resource);
	POLL_FOR_DMA_INTERRUPT(pChannel->bmaddr_resource, baseBm)
	ScsiPortWritePortUchar(pChannel->bmaddr_resource,
		(unsigned)&(baseBm->Command.AsUchar), STOP_TRANSFER);

	GET_BASE_STATUS(pChannel->base1_resource, baseIoAddress1, ucStatus);

	/* Clear interrupt */
	ClearPCIInterupt(DeviceExtension, Channel);

	/* if no error continue; */
	if (ucStatus & IDE_STATUS_BUSY) {
		return FALSE;
	}
	/* if no error continue; */
	if (ucStatus & IDE_STATUS_ERROR) {
		return FALSE;
	}
	return TRUE;
}


BOOLEAN
IsPostCommandToDriveSuccess68x(PHW_DEVICE_EXTENSION DeviceExtension,
    ULONG ulTargetId,
    ULONG ulStartBlock,
    ULONG ulBlockCount,
    UCHAR ucCmd,
    ULONG ulSglPhysicalAddress
)
{
	UCHAR Channel = (ulTargetId >> 1);
	PCHANNEL pChannel;
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PBM_REGISTERS baseBm;
	UCHAR ucStatus = 0;

	pChannel = &(DeviceExtension->Channel[Channel]);
	baseIoAddress1 = (PIDE_REGISTERS_1)pChannel->BaseIoAddress1;
	baseIoAddress2 = (PIDE_REGISTERS_2)pChannel->BaseIoAddress2;
	baseBm = pChannel->BaseBmAddress;

	PostCommandToDrive(DeviceExtension,
		ulTargetId, ulStartBlock, ulBlockCount,
		ucCmd, ulSglPhysicalAddress);

	POLL_FOR_INTERRUPT_68X();
	POLL_FOR_DMA_INTERRUPT(pChannel->bmaddr_resource, baseBm)
	ScsiPortWritePortUchar(pChannel->bmaddr_resource,
		(unsigned)&(baseBm->Command.AsUchar), STOP_TRANSFER);

	GET_BASE_STATUS(pChannel->base1_resource,
		baseIoAddress1, ucStatus);

	/* Clear interrupt */
	ClearPCIInterupt(DeviceExtension, Channel);

	if (ucStatus & IDE_STATUS_BUSY) {
		printk_Rlsi("680 interrupt busy \
			bit set on drive=%ld\n", ulTargetId);
		return FALSE;
	}
	/* if no error continue; */
	if (ucStatus & IDE_STATUS_ERROR) {
		UCHAR error_byte;

		printk_Rlsi("680 interrupt error bit \
			set on drive=%ld\n", ulTargetId);
		error_byte = ScsiPortReadPortUchar(
			pChannel->base1_resource,
			(unsigned)((PUCHAR)baseIoAddress1 + 1));
		printk_Rlsi("error byte = 0x%x\n", error_byte);
		return FALSE;
	}
	return TRUE;
}



BOOLEAN
IsPostCommandToDriveSuccessCSB5(PHW_DEVICE_EXTENSION DeviceExtension,
    ULONG ulTargetId,
    ULONG ulStartBlock,
    ULONG ulBlockCount,
    UCHAR ucCmd,
    ULONG ulSglPhysicalAddress
)
{
	UCHAR Channel = (ulTargetId >> 1);
	PCHANNEL pChannel;
	PIDE_REGISTERS_1 baseIoAddress1;
	PIDE_REGISTERS_2 baseIoAddress2;
	PBM_REGISTERS baseBm;
	UCHAR ucStatus = 0;

	pChannel = &(DeviceExtension->Channel[Channel]);
	baseIoAddress1 = (PIDE_REGISTERS_1)pChannel->BaseIoAddress1;
	baseIoAddress2 = (PIDE_REGISTERS_2)pChannel->BaseIoAddress2;
	baseBm = pChannel->BaseBmAddress;

	PostCommandToDrive(DeviceExtension, ulTargetId, ulStartBlock,
		ulBlockCount, ucCmd, ulSglPhysicalAddress);

	POLL_FOR_INTERRUPT_CSB5(
		DeviceExtension->Channel[Channel].bmaddr_resource,
		baseBm);
	POLL_FOR_DMA_INTERRUPT(DeviceExtension->Channel[Channel].bmaddr_resource, baseBm)
	    ScsiPortWritePortUchar(
			DeviceExtension->Channel[Channel].bmaddr_resource,
			(unsigned)&(baseBm->Command.AsUchar), STOP_TRANSFER);

	GET_BASE_STATUS(DeviceExtension->Channel[Channel].base1_resource, baseIoAddress1, ucStatus);

	if (ucStatus & IDE_STATUS_BUSY) {
		return FALSE;
	}
	/* if no error continue; */
	if (ucStatus & IDE_STATUS_ERROR) {
		return FALSE;
	}
	return TRUE;
}

static void
mide_periodic(void *data)
{
	PHW_DEVICE_EXTENSION pDE = (PHW_DEVICE_EXTENSION)data;
	ULONG ulChannelInd;
	int s;
	for (ulChannelInd = 0; ulChannelInd < MAX_CHANNELS_PER_CONTROLLER; ulChannelInd++) {
		if (pDE->Channel[ulChannelInd].ActiveQueue || (pDE->ActiveGateKeeper))
			/* if ( pDE->Channel[ulChannelInd].ActiveQueue ) */
		{
			pDE->Channel[ulChannelInd].ulActiveCmdTimeOut -= 1;
			if (!pDE->Channel[ulChannelInd].ulActiveCmdTimeOut) {
				AR_DEBUG(AR_DEBUG_DRVR_ENTRY, ("Timed Out \n"));
				xpt_freeze_devq(gastCardInfo[pDE->ucControllerId].mide_path, /* count */ 1);
				s = splcam();
				FastIDEResetBus(pDE);
				splx(s);
				xpt_release_devq(gastCardInfo[pDE->ucControllerId].mide_path, /* count */ 1, TRUE);
				AR_DEBUG(AR_DEBUG_DRVR_ENTRY, ("megaide _reset  Completed\n"));
			}
		}
	}


	/* reschedule */
	gastCardInfo[pDE->ucControllerId].mide_timeout = timeout(mide_periodic, pDE, hz);
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_shutdown(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);

	/* debug_called(1); */


	/* flush controller */
	device_printf(gastCardInfo[DeviceExtension->ucControllerId].pstDev, "shut down flushing cache...");


	/* XXX disable interrupts? */

	return (0);
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_suspend(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);


    	/*	debug_called(1); */
	mide_Suspend(DeviceExtension);
	return(0);

}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static int
mide_pci_resume(device_t dev)
{
	PHW_DEVICE_EXTENSION DeviceExtension = device_get_softc(dev);

	mide_Resume(DeviceExtension);
	return(0);
}

PHW_DEVICE_EXTENSION
GetDeviceExtension(u_char ucIndex)
{
	if (ucIndex < giNumberOfCtrlrs)
		return gastCardInfo[ucIndex].DeviceExtension;
	else
		return NULL;
}


/*
  * -------------------------------------------------------------------
  * Description ????
  *
  * Arguments
  *
  * Return Value
  * -------------------------------------------------------------------
  */

BOOLEAN
AssignMemoryGatekeeper(PHW_DEVICE_EXTENSION DeviceExtension)
{
	ULONG length = MAX_SECTORS_PER_IDE_TRANSFER * IDE_SECTOR_SIZE * MAX_CHANNELS_PER_CONTROLLER;

	length = (length > KMALLOC_MAXSIZE) ? KMALLOC_MAXSIZE : length;
	/* length = 64*1024;  */
	DeviceExtension->GateKeeperBufferLength = length;

	if (bus_dma_tag_create(
		(bus_dma_tag_t)DeviceExtension->parent_datatransfer_dmat,
		/* alignment */ 1,
		/* boundary */ 0,
		/* lowaddr */ BUS_SPACE_MAXADDR,
		/* highaddr */ BUS_SPACE_MAXADDR,
		/* filter */ NULL,
		/* filterarg */ NULL,
		/* maxsize */ DeviceExtension->GateKeeperBufferLength,
		/* nsegments */ 1,
		/* maxsegsz */ BUS_SPACE_MAXSIZE_32BIT,
		/* flags */ 0,
		((bus_dma_tag_t *)&DeviceExtension->GateKeeperBufferDmat))) {
		AR_DEBUG(AR_DEBUG_WARNING, ("can't create DMA tag for gatekeeper !!!!!!!\n"));
		return FALSE;
	}
	if (bus_dmamem_alloc(
		DeviceExtension->GateKeeperBufferDmat,
		(void **)&DeviceExtension->GateKeeperBuffer,
		BUS_DMA_NOWAIT,
		((bus_dmamap_t *)&DeviceExtension->GateKeeperBufferDmaMap))) {
		AR_DEBUG(AR_DEBUG_WARNING, ("can't allocate memory for gatekeeper !!!!!!!\n"));
		return FALSE;
	}
	bus_dmamap_load(
	    DeviceExtension->GateKeeperBufferDmat,
	    DeviceExtension->GateKeeperBufferDmaMap,
	    DeviceExtension->GateKeeperBuffer,
	    DeviceExtension->GateKeeperBufferLength,
	    mide_gatekeeper_map_helper_channel,
	    DeviceExtension,
	    0);

	bzero(DeviceExtension->GateKeeperBuffer, DeviceExtension->GateKeeperBufferLength);

	return TRUE;
}

void
FreeMemoryGatekeeper(PHW_DEVICE_EXTENSION DeviceExtension)
{

	if (DeviceExtension->GateKeeperBuffer != NULL) {
		bus_dmamap_unload(DeviceExtension->GateKeeperBufferDmat,
			DeviceExtension->GateKeeperBufferDmaMap);
		bus_dmamem_free(DeviceExtension->GateKeeperBufferDmat,
			DeviceExtension->GateKeeperBuffer,
			DeviceExtension->GateKeeperBufferDmaMap);
		bus_dma_tag_destroy(DeviceExtension->GateKeeperBufferDmat);

		DeviceExtension->GateKeeperBuffer = NULL;
		DeviceExtension->GateKeeperBufferLength = 0;
		DeviceExtension->GateKeeperBufferDmat = NULL;
		DeviceExtension->GateKeeperBufferDmaMap = NULL;

	}
}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

BOOLEAN
AssignPtrsForDevExtension(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR ulLogicalId, ulChannelId, ulTargetId;
	PCHANNEL pChannel;

	/*
	 *
	 * printk_Rlsi("channel=0x%lx\n",(ULONG)&DeviceExtension->Channel[0]);
	 * printk_Rlsi("phy=0x%lx\n",
	 *		(ULONG)&DeviceExtension->PhysicalDrive[0]);
	 * printk_Rlsi("log=0x%lx\n",
	 * 	(ULONG)&DeviceExtension->LogicalDrive[0]);
	 * printk_Rlsi("dev=0x%lx\n",(ULONG)DeviceExtension);
	 */

	/* allocating memory for base dev extension */
	/* DeviceExtension->GateKeeperBuffer */
	/* DeviceExtension->GateKeeper */
	/* DeviceExtension->GateScbExtension */
	/* DeviceExtension->scb_queue */

	AssignMemoryGatekeeper(DeviceExtension);

	DeviceExtension->GateKeeper =
		malloc(get_gatekeeper_size(), M_MIDE, M_NOWAIT | M_ZERO);
	if (DeviceExtension->GateKeeper == NULL) {
		AR_DEBUG(AR_DEBUG_WARNING,
			("malloc failed for GateKeeper !!!!!!!\n"));
		return FALSE;
	}
	DeviceExtension->GateScbExtension = malloc(get_scb_extension_size(), M_MIDE, M_NOWAIT | M_ZERO);
	if (DeviceExtension->GateScbExtension == NULL) {
		AR_DEBUG(AR_DEBUG_WARNING,
			("malloc failed for GateScbExtension !!!!!!!\n"));
		return FALSE;
	}
	DeviceExtension->scb_queue = malloc((M_IDE_MAX_COMMANDS + MAX_IOCTL_MEM_ALLOC)*get_scb_extension_size(), M_MIDE, M_NOWAIT | M_ZERO);
	if (DeviceExtension->scb_queue == NULL) {
		AR_DEBUG(AR_DEBUG_WARNING,
			("malloc failed for scb_queue !!!!!!!\n"));
		return FALSE;
	}
	/* allocating memory for logical drives */
	/* DeviceExtension->LogicalDrive[].GateKeeper */
	/* DeviceExtension->LogicalDrive[].InternalScbExtension */

	for (ulLogicalId = 0;
		ulLogicalId < MAX_LOGICAL_DRIVES;
		ulLogicalId++) {
		DeviceExtension->LogicalDrive[ulLogicalId].GateKeeper =
		    malloc(get_gatekeeper_size(), M_MIDE, M_NOWAIT | M_ZERO);
		if (DeviceExtension->LogicalDrive[ulLogicalId].GateKeeper == NULL) {
			AR_DEBUG(AR_DEBUG_WARNING,
				("malloc failed for GateKeeper !!!!!!!\n"));
			return FALSE;
		}
		DeviceExtension->LogicalDrive[ulLogicalId].InternalScbExtension =
		    malloc(get_scb_extension_size(), M_MIDE, M_NOWAIT | M_ZERO);
		if (DeviceExtension->LogicalDrive[ulLogicalId].InternalScbExtension == NULL) {
			AR_DEBUG(AR_DEBUG_WARNING,
				("malloc failed for InternalScbExtension !!!!!!!\n"));
			return FALSE;
		}
	}

	/* allocating memory for channel */
	/* DeviceExtension->Channel[].DiskConfiguration */
	/* DeviceExtension->Channel[].DiskEnquiry */
	for (ulChannelId = 0;
		ulChannelId < MAX_CHANNELS_PER_CONTROLLER;
		ulChannelId++) {
		pChannel = &DeviceExtension->Channel[ulChannelId];
		pChannel->DiskConfiguration =
		    contigmalloc(IDE_SECTOR_SIZE, M_MIDE, M_NOWAIT, 0, ~0, PAGE_SIZE, 0);
		if (pChannel->DiskConfiguration == NULL) {
			AR_DEBUG(AR_DEBUG_WARNING,
				("malloc failed for DiskConfiguration !!!!!!!\n"));
			return FALSE;
		}
		pChannel->DiskConfigurationSgl.Physical.Address =
		    (void *)vtobus(DeviceExtension->Channel[ulChannelId].DiskConfiguration);
		pChannel->DiskConfigurationSgl.Physical.Length = IDE_SECTOR_SIZE;
		pChannel->DiskConfigurationSgl.Physical.EndOfListFlag = 1;

		pChannel->DiskEnquiry =
		    contigmalloc(IDE_SECTOR_SIZE, M_MIDE, M_NOWAIT, 0, ~0, PAGE_SIZE, 0);
		if (pChannel->DiskEnquiry == NULL) {
			AR_DEBUG(AR_DEBUG_WARNING, ("malloc failed for DiskEnquiry !!!!!!!\n"));
			return FALSE;
		}
		pChannel->DiskEnquirySgl.Physical.Address =
		    (void *)vtobus(DeviceExtension->Channel[ulChannelId].DiskEnquiry);
		pChannel->DiskEnquirySgl.Physical.Length = IDE_SECTOR_SIZE;
		pChannel->DiskEnquirySgl.Physical.EndOfListFlag = 1;

	}

	/* allocating memory for Physical drive */
	/* DeviceExtension->PhysicalDrive[].ErrorLogSector */

	for (ulTargetId = 0; ulTargetId < MAX_DRIVES_PER_CONTROLLER; ulTargetId++) {
		DeviceExtension->PhysicalDrive[ulTargetId].ErrorLogSector =
		    contigmalloc(IDE_SECTOR_SIZE, M_MIDE, M_NOWAIT, 0, ~0, PAGE_SIZE, 0);
		if (DeviceExtension->PhysicalDrive[ulTargetId].ErrorLogSector == NULL) {
			AR_DEBUG(AR_DEBUG_WARNING, ("malloc failed for ErrorLogSector !!!!!!!\n"));
			return FALSE;
		}
		DeviceExtension->PhysicalDrive[ulTargetId].DiskErrorLogSgl.Physical.Address =
		    (void *)vtobus(DeviceExtension->PhysicalDrive[ulTargetId].ErrorLogSector);
		DeviceExtension->PhysicalDrive[ulTargetId].DiskErrorLogSgl.Physical.Length = IDE_SECTOR_SIZE;
		DeviceExtension->PhysicalDrive[ulTargetId].DiskErrorLogSgl.Physical.EndOfListFlag = 1;

	}

	return TRUE;
}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
FreePtrsForDevExtension(PHW_DEVICE_EXTENSION DeviceExtension)
{
	UCHAR ulLogicalId, ulChannelId, ulTargetId;

	/*
	 * free'ing memory for base dev extension
	 * DeviceExtension->GateKeeperBuffer DeviceExtension->GateKeeper
	 * DeviceExtension->GateScbExtension DeviceExtension->scb_queue
	 */


	FreeMemoryGatekeeper(DeviceExtension);

	if (DeviceExtension->GateKeeper != NULL) {
		free(DeviceExtension->GateKeeper, M_MIDE);
	}
	if (DeviceExtension->GateScbExtension != NULL) {
		free(DeviceExtension->GateScbExtension, M_MIDE);
	}
	if (DeviceExtension->scb_queue == NULL) {
		free(DeviceExtension->scb_queue, M_MIDE);
	}
	/*
	 * free'ing memory for logical drives
	 * DeviceExtension->LogicalDrive[].GateKeeper
	 * DeviceExtension->LogicalDrive[].InternalScbExtension
	 */

	for (ulLogicalId = 0; ulLogicalId < MAX_LOGICAL_DRIVES; ulLogicalId++) {
		if (DeviceExtension->LogicalDrive[ulLogicalId].GateKeeper != NULL) {
			free(DeviceExtension->LogicalDrive[ulLogicalId].GateKeeper, M_MIDE);
		}
		if (DeviceExtension->LogicalDrive[ulLogicalId].InternalScbExtension == NULL) {
			free(DeviceExtension->LogicalDrive[ulLogicalId].InternalScbExtension, M_MIDE);
		}
	}

	/*
	 * allocating memory for channel
	 * DeviceExtension->Channel[].DiskConfiguration
	 * DeviceExtension->Channel[].DiskEnquiry
	 */
	for (ulChannelId = 0;
	    ulChannelId < MAX_CHANNELS_PER_CONTROLLER; ulChannelId++) {
		if (DeviceExtension->Channel[ulChannelId].DiskConfiguration != NULL)
			contigfree(DeviceExtension->Channel[ulChannelId].DiskConfiguration,
			    IDE_SECTOR_SIZE,
			    M_MIDE);

		if (DeviceExtension->Channel[ulChannelId].DiskEnquiry != NULL)
			contigfree(DeviceExtension->Channel[ulChannelId].DiskEnquiry,
			    IDE_SECTOR_SIZE,
			    M_MIDE);
	}

	/*
	 * allocating memory for Physical drive
	 * DeviceExtension->PhysicalDrive[].ErrorLogSector
	 */

	for (ulTargetId = 0; ulTargetId < MAX_DRIVES_PER_CONTROLLER; ulTargetId++) {
		if (DeviceExtension->PhysicalDrive[ulTargetId].ErrorLogSector != NULL)
			contigfree(DeviceExtension->PhysicalDrive[ulTargetId].ErrorLogSector,
			    IDE_SECTOR_SIZE,
			    M_MIDE);
	}

}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

BOOLEAN
AssignSglPtrsForChannels(PHW_DEVICE_EXTENSION DeviceExtension)
{
	size_t segsize;
	PCHANNEL pChannel;
	ULONG ulChannel;


	/*
	 * Create a tag for mapping buffers into adapter-addressable
	 */

	if (bus_dma_tag_create(
		/* parent */ NULL,
		/* alignment */ 1,
		/* boundary */ 0,
		/* lowaddr */ BUS_SPACE_MAXADDR_32BIT,
		/* highaddr */ BUS_SPACE_MAXADDR,
		/* filter */ NULL,
		/* filterarg */ NULL,
		/* maxsize */ MAX_BYTES_PER_IDE_TRANSFER,
		/* nsegments */ M_IDE_MAX_SGLIST,
		/* maxsegsize */ BUS_SPACE_MAXSIZE_32BIT,
		/* flags */ BUS_DMA_ALLOCNOW,
		((bus_dma_tag_t *) & DeviceExtension->parent_datatransfer_dmat))) {
		AR_DEBUG(AR_DEBUG_WARNING, ("can't allocate DMA tag for parent !!!!!!!\n"));
		return FALSE;
	}
	if (bus_dma_tag_create(
		(bus_dma_tag_t) DeviceExtension->parent_datatransfer_dmat,
		/* alignment */ 1,
		/* boundary */ 0,
		/* l owaddr */ BUS_SPACE_MAXADDR,
		/* highaddr */ BUS_SPACE_MAXADDR,
		/* filter */ NULL,
		/* filterarg */ NULL,
		/* maxsize */ MAX_BYTES_PER_IDE_TRANSFER,
		/* nsegments */ M_IDE_MAX_SGLIST,
		/* maxsegsz */ BUS_SPACE_MAXSIZE_32BIT,
		/* f lags */ 0,
		((bus_dma_tag_t *) & DeviceExtension->mide_buffer_dmat))) {
		AR_DEBUG(AR_DEBUG_WARNING,
		    ("can't allocate DMA tag for per request buffers !!!!!!!\n"));
		return FALSE;
	}
	/*
	 * Create a tag per channel sglist
	 */

	if (bus_dma_tag_create(
		NULL, /* parent */
		1, /* alignment */
		0, /* boundary */
		BUS_SPACE_MAXADDR_32BIT, /* lowaddr */
		BUS_SPACE_MAXADDR, /* highaddr */
		NULL, /* filter */
		NULL, /* filterarg */
		MAXBSIZE, /* maxsize */
		M_IDE_MAX_SGLIST, /* nsegments */
		BUS_SPACE_MAXSIZE_32BIT, /* maxsegsize */
		BUS_DMA_ALLOCNOW, /* flags */
		((bus_dma_tag_t *) & DeviceExtension->parent_sglist_dmat))) {
		AR_DEBUG(AR_DEBUG_WARNING, ("can't allocate DMA tag for parent !!!!!!!\n"));
		return FALSE;
	}
	segsize = MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY);

	for (ulChannel = 0; ulChannel < MAX_CHANNELS_PER_CONTROLLER; ulChannel++) {
		pChannel = &(DeviceExtension->Channel[ulChannel]);

		if (bus_dma_tag_create(
			(bus_dma_tag_t) DeviceExtension->parent_sglist_dmat,
			/* alignment */ 1,
			/* boundary */ 0,
			/* lowaddr */ BUS_SPACE_MAXADDR,
			/* highaddr */ BUS_SPACE_MAXADDR,
			/* filter */ NULL,
			/* filterarg */ NULL,
			/* maxsize */ segsize,
			/* nsegments */ 1,
			/* maxsegsz */ BUS_SPACE_MAXSIZE_32BIT,
			/* flags */ 0,
			((bus_dma_tag_t *) & pChannel->mide_sg_dmat))) {
			AR_DEBUG(AR_DEBUG_WARNING,
			    ("can't create DMA tag for sglist !!!!!!!\n"));
			return FALSE;
		}
		if (bus_dmamem_alloc(
			pChannel->mide_sg_dmat,
			(void **)&pChannel->pSglEntry,
			BUS_DMA_NOWAIT,
			((bus_dmamap_t *) & pChannel->mide_sg_dmamap))) {
			AR_DEBUG(AR_DEBUG_WARNING,
			    ("can't allocate memory for sglist !!!!!!!\n"));
			return FALSE;
		}
		bus_dmamap_load(
		    pChannel->mide_sg_dmat,
		    pChannel->mide_sg_dmamap,
		    pChannel->pSglEntry,
		    segsize,
		    mide_sglist_map_helper_channel,
		    pChannel,
		    0);
	}

	return TRUE;
}

/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
FreeSglPtrsForChannels(PHW_DEVICE_EXTENSION DeviceExtension)
{
	PCHANNEL pChannel;
	ULONG ulChannel;

	for (ulChannel = 0; ulChannel < MAX_CHANNELS_PER_CONTROLLER; ulChannel++) {
		/* free and destroy DMA memory and tag for s/g lists */
		pChannel = &(DeviceExtension->Channel[ulChannel]);

		if (!pChannel->mide_sg_dmat)
			continue;

		bus_dmamap_unload(pChannel->mide_sg_dmat,
		    pChannel->mide_sg_dmamap);
		bus_dmamem_free(pChannel->mide_sg_dmat,
		    pChannel->pSglEntry, pChannel->mide_sg_dmamap);
		bus_dma_tag_destroy(pChannel->mide_sg_dmat);
	}

	if (DeviceExtension->mide_buffer_dmat)
		bus_dma_tag_destroy(DeviceExtension->mide_buffer_dmat);

	/* destroy the parent DMA tag */
	if (DeviceExtension->parent_sglist_dmat)
		bus_dma_tag_destroy(DeviceExtension->parent_sglist_dmat);

	if (DeviceExtension->parent_datatransfer_dmat)
		bus_dma_tag_destroy(DeviceExtension->parent_datatransfer_dmat);

}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
mide_sglist_map_helper_channel(void *arg,
    bus_dma_segment_t * segs, int nseg, int error)
{
	PCHANNEL pChannel = (PCHANNEL) arg;
	PSGL_ENTRY sgl;
	ULONG sglBasePhysicalAddress, length;
	ULONG ulStartAddress, ulEndAddress, ulStartPage, ulEndPage;
	ULONG ulLengthReduced = 0;
	ULONG maxSglEntries;
	LONG lengthInNextBoundary;
	LONG lengthLeftInBoundary;

	if (nseg != 1) {
		printk_Rlsi(
		    "mide_sglist_map_helper_channel: nseg=%d\n", nseg);
		return;
	}
	memset_Rlsi(pChannel->pSglEntry, 0, segs->ds_len);

	sglBasePhysicalAddress = segs->ds_addr;
	sgl = pChannel->pSglEntry;

	if (sglBasePhysicalAddress % SGL_HW_BOUNDARY) {
		length = (((sglBasePhysicalAddress / SGL_HW_BOUNDARY) + 1) *
		    SGL_HW_BOUNDARY) - sglBasePhysicalAddress;
	} else {
		length = SGL_HW_BOUNDARY;
	}

	if (length >=
	    (MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY))) {
		length = MAX_SGL_ENTRIES_PER_PHYSICAL_DRIVE * sizeof(SGL_ENTRY);
	}
	ulStartAddress = (ULONG) sglBasePhysicalAddress;
	ulEndAddress = ulStartAddress + length;


	/*
	 * Check to see if SGL Pointer is DWORD Aligned and make it so.
	 * Remember we allocated an extra SGL entry for this purpose.
	 */

	if ((ULONG) sglBasePhysicalAddress & 3) {
		sglBasePhysicalAddress =
		    (((ULONG) sglBasePhysicalAddress & 0xfffffffcL) + 4L);
		ulLengthReduced =
		    sglBasePhysicalAddress - ulStartAddress;
		ulStartAddress = (ULONG) sglBasePhysicalAddress;
		length -= ulLengthReduced;
		sgl += ulLengthReduced;
	}
	ulStartPage = (ULONG) (ulStartAddress / SGL_HW_BOUNDARY);
	ulEndPage = (ULONG) (ulEndAddress / SGL_HW_BOUNDARY);

	if (ulStartPage == ulEndPage) {
		/* all entries are in the same page */
		maxSglEntries = (length / sizeof(SGL_ENTRY));
	} else {
		/*
		 * Make sure the SGL does NOT cross a 4K boundary.
		 */

		lengthLeftInBoundary = SGL_HW_BOUNDARY -
		    ((ULONG) sglBasePhysicalAddress & (SGL_HW_BOUNDARY - 1));
		lengthInNextBoundary = length - lengthLeftInBoundary;

		if (lengthInNextBoundary > lengthLeftInBoundary) {
			/*
			 * SGL crosses 4KB boundary and top portion is larger
			 * than bottom. Use top portion.
			 */

			sglBasePhysicalAddress = sglBasePhysicalAddress + lengthLeftInBoundary;
			sgl = (PSGL_ENTRY) ((ULONG) sgl + lengthLeftInBoundary);
			maxSglEntries = lengthInNextBoundary / sizeof(SGL_ENTRY);

		} else {
			maxSglEntries = lengthLeftInBoundary / sizeof(SGL_ENTRY);
		}
	}

	pChannel->SglBaseVirtualAddress = (PSGL_ENTRY) sgl;
	pChannel->SglBasePhysicalAddress = sglBasePhysicalAddress;
	pChannel->MaxSglEntries =
	    (maxSglEntries < M_IDE_MAX_SGLIST) ?
	    maxSglEntries : M_IDE_MAX_SGLIST;
}
#ifdef BP_SUPPORT
ULONG megaide_get_bp_opcode(ULONG opcode)
{
	return bp_values[opcode];
}
#endif
