/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/mide_scsi.c,v emoore Exp$
 */

char OemProductId[16] = " LD XY  MegaRAID";
char OemVendorId[8] = "LSILogic";
u_int8_t waitflag = 0;
u_int8_t waitgateflag = 0;

static void
megaide_cam_action(struct cam_sim * sim, union ccb * ccb)
{
	PHW_DEVICE_EXTENSION DeviceExtension = cam_sim_softc(sim);
	int s;

	ccb->ccb_h.spriv_ptr1 = xpt_done;

	switch (ccb->ccb_h.func_code) {
		/* execute an I/O transaction */
	case XPT_SCSI_IO:
		{

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_SCSI_IO\n"));
			s = splcam();
			FastIDEProcessAndQueueRequest(DeviceExtension, &ccb->csio);
			splx(s);
			break;
		}

		/* calculate logical geometry of the disk */
	case XPT_CALC_GEOMETRY:
		{
			int heads, sectors, cylinders;
			struct ccb_calc_geometry *ccg = &ccb->ccg;

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_CALC_GEOMETRY\n"));

			if (ccg->volume_size < 1032192) {	/* size less than 504MB */
				heads = 16;
				sectors = 63;
				cylinders = (int)ccg->volume_size / (heads * sectors);
			} else if (ccg->volume_size < 2064384) {	/* size less than 1008MB */
				heads = 32;
				sectors = 63;
				cylinders = (int)ccg->volume_size / (heads * sectors);
			} else if (ccg->volume_size < 4128768) {	/* size less than 2016MB */
				heads = 64;
				sectors = 63;
				cylinders = (int)ccg->volume_size / (heads * sectors);
			} else if (ccg->volume_size < 8257536) {	/* size less than 4032MB */
				heads = 128;
				sectors = 63;
				cylinders = (int)ccg->volume_size / (heads * sectors);
			} else {/* size greater than 4032MB */
				heads = 255;
				sectors = 63;
				cylinders = (int)ccg->volume_size / (heads * sectors);
			}

			/* return result */
			ccg->heads = heads;
			ccg->secs_per_track = sectors;
			ccg->cylinders = cylinders;

			ccg->ccb_h.status = CAM_REQ_CMP;
			scsi_done((void *)ccb);
			break;
		}


		/* path inquiry of host bus adapter */
	case XPT_PATH_INQ:
		{
			struct ccb_pathinq *cpi = &ccb->cpi;
			int controlller_speed = 0;

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_PATH_INQ\n"));

			cpi->version_num = 1;	/* XXX  */
			cpi->hba_inquiry = PI_TAG_ABLE | PI_WIDE_16;
			cpi->hba_misc = PIM_NOBUSRESET;
			cpi->target_sprt = 0;
			cpi->hba_eng_cnt = 0;
			cpi->max_target = M_IDE_MAX_LOGICAL_DRIVES;
			cpi->max_lun = 0;
			cpi->initiator_id = M_IDE_MAX_LOGICAL_DRIVES;	/* XXX variable */
			strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
			strncpy(cpi->hba_vid, "LSI", HBA_IDLEN);
			strncpy(cpi->dev_name, cam_sim_name(sim), DEV_IDLEN);
			cpi->unit_number = cam_sim_unit(sim);
			cpi->bus_id = cam_sim_bus(sim);

			switch (DeviceExtension->ControllerSpeed) {
			case Udma100:
				controlller_speed = 100;
				break;

			case Udma133:
				controlller_speed = 133;
				break;
			case Udma66:
				controlller_speed = 66;
			default:
				break;
			}

			cpi->base_transfer_speed = (1000 * controlller_speed) / 2;
			cpi->ccb_h.status = CAM_REQ_CMP;
			scsi_done((void *)ccb);
			break;

		}

		/* abort respective ccb */
	case XPT_ABORT:
		{
			struct ccb_abort *abort_ccb = &ccb->cab;

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_ABORT\n"));
			ccb->ccb_h.status = megiade_abort(DeviceExtension, abort_ccb);
			scsi_done((void *)ccb);
			break;
		}

		/* bus reset */
	case XPT_RESET_BUS:
		{
			struct ccb_resetbus *resetbus_ccb = &ccb->crb;

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_RESET_BUS\n"));
			ccb->ccb_h.status = megiade_reset_bus(DeviceExtension, resetbus_ccb);
			scsi_done((void *)ccb);
			break;
		}

		/* bus device reset */
	case XPT_RESET_DEV:
		{
			struct ccb_resetdev *resetdev_ccb = &ccb->crd;

			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_RESET_DEV\n"));
			ccb->ccb_h.status = megiade_reset_dev(DeviceExtension, resetdev_ccb);
			scsi_done((void *)ccb);
			break;
		}

		/* get values of SCSI transfer settings */
	case XPT_GET_TRAN_SETTINGS:
		{
			struct ccb_trans_settings	*cts = &ccb->cts;
			PLOGICAL_DRIVE_INFO ldInfo;
			int ControllerNumber, bus;

			ControllerNumber = DeviceExtension->ucControllerId;

			bus = cam_sim_bus(gastCardInfo[ControllerNumber].mide_sim);

			ldInfo = GetLDInfo(DeviceExtension, bus, cts->ccb_h.target_id, cts->ccb_h.target_lun);
			AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_GET_TRAN_SETTINGS\n"));

			if (!ldInfo) {
				cts->ccb_h.status = CAM_DEV_NOT_THERE;
				scsi_done((void *)ccb);
				return ;
			}

			cts->valid = 0;
			cts->valid |= CCB_TRANS_TQ_VALID;
			cts->valid |= CCB_TRANS_DISC_VALID;
			cts->valid |= CCB_TRANS_BUS_WIDTH_VALID;
			cts->flags |= (!ldInfo->TagDisable) ?  CCB_TRANS_TAG_ENB:  ~CCB_TRANS_TAG_ENB;
			cts->flags |= CCB_TRANS_DISC_ENB;

			cts->bus_width = MSG_EXT_WDTR_BUS_16_BIT;
			cts->ccb_h.status = CAM_REQ_CMP;
			scsi_done((void *)ccb);
			break;
		}

		/* set values of SCSI transfer settings */
	case XPT_SET_TRAN_SETTINGS:
		{
			struct ccb_trans_settings	*cts = &ccb->cts;
			PLOGICAL_DRIVE_INFO ldInfo;
			int ControllerNumber,bus;

			AR_DEBUG(AR_DEBUG_SCSI_CMD,("XPT_SET_TRAN_SETTINGS\n"));
			ControllerNumber = DeviceExtension->ucControllerId;
			bus = cam_sim_bus(gastCardInfo[ControllerNumber].mide_sim);

			ldInfo = GetLDInfo(DeviceExtension, bus, cts->ccb_h.target_id, cts->ccb_h.target_lun);

			if (!ldInfo) {
				cts->ccb_h.status = CAM_DEV_NOT_THERE;
				scsi_done((void *)ccb);
				return ;
			}
			if (!(cts->flags & CCB_TRANS_USER_SETTINGS)) {
				cts->ccb_h.status = CAM_REQ_INVALID;
				scsi_done((void *)ccb);
				break;
			}
			ldInfo->TagDisable = (!(cts->flags & CCB_TRANS_TAG_ENB))? 1: 0;

			cts->ccb_h.status = CAM_REQ_CMP;
			scsi_done((void *)ccb);
			break;
		}
	case XPT_TERM_IO:	/* Terminate the I/O process */
		/* XXX Implement */
		AR_DEBUG(AR_DEBUG_SCSI_CMD, ("XPT_TERM_IO\n"));
		ccb->ccb_h.status = CAM_REQ_INVALID;
		scsi_done((void *)ccb);
		break;

	default:
		{
			/* we can't do this */
			AR_DEBUG(AR_DEBUG_SCSI_CMD, (
                        	"NOT_SUPPORTED_CMD_XPT=0x%x\n",
                                ccb->ccb_h.func_code));
			ccb->ccb_h.status = CAM_REQ_INVALID;
			scsi_done((void *)ccb);
			break;
		}
	}

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value TRUE if success
 * -------------------------------------------------------------------
 */

void
FastIDEProcessAndQueueRequest(PHW_DEVICE_EXTENSION DeviceExtension, struct ccb_scsiio * csio)
{
	u_int32_t ccb_status = CAM_REQ_INVALID;
	PLOGICAL_DRIVE_INFO ldInfo;
	PFAST_IDE_SCB_EXTENSION scbExtension;
	int ControllerNumber, bus;
	UCHAR cdb_data[16];

	ControllerNumber = DeviceExtension->ucControllerId;

	bus = cam_sim_bus(gastCardInfo[ControllerNumber].mide_sim);

	ldInfo = GetLDInfo(DeviceExtension, bus, csio->ccb_h.target_id, csio->ccb_h.target_lun);

	if (!ldInfo) {
		csio->ccb_h.status = CAM_DEV_NOT_THERE;
		scsi_done((void *)csio);
		return;
	}
	if (ldInfo->Status == LDS_OffLine) {
		AR_DEBUG(AR_DEBUG_WARNING, ("Array is Offline \n"));
		csio->ccb_h.status = CAM_SEL_TIMEOUT;
		scsi_done((void *)csio);
		return;
	}
	/* driver does not support commands  > 10 bytes */
	if (csio->cdb_len > 10) {
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("command too large (%d > %d)\n",
			csio->cdb_len, 10));
		csio->ccb_h.status = CAM_REQ_CMP_ERR;
		scsi_done((void *)csio);
		return;
	}
	/* check that the CDB pointer is not to a physical address */
	if ((csio->ccb_h.flags & CAM_CDB_POINTER) && (csio->ccb_h.flags & CAM_CDB_PHYS)) {
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("CDB pointer is to physical address"));
		csio->ccb_h.status = CAM_REQ_CMP_ERR;
		scsi_done((void *)csio);
		return;
	}
	/* check that SGLIST is not to a physical address */
	if ((csio->ccb_h.flags & CAM_SCATTER_VALID) && (csio->ccb_h.flags & CAM_SG_LIST_PHYS)) {
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("SGLIST pointer is to physical address"));
		csio->ccb_h.status = CAM_REQ_CMP_ERR;
		scsi_done((void *)csio);
		return;
	}
	if (csio->ccb_h.flags & CAM_CDB_POINTER) {
		bcopy(csio->cdb_io.cdb_ptr, cdb_data, csio->cdb_len);
	} else {
		bcopy(csio->cdb_io.cdb_bytes, cdb_data, csio->cdb_len);
	}

	check_for_config_update(DeviceExtension);

	switch (cdb_data[0]) {

	case VERIFY:
	case READ_6:
	case READ_10:
	case WRITE_6:
	case WRITE_10:
	case WRITE_VERIFY:
		{

			if (ldInfo->Status == LDS_Deleted) {
				csio->ccb_h.status = CAM_DEV_NOT_THERE;
				scsi_done((void *)csio);
				return;
			}
			AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("READ/WRITE - bus=%x,target=%x,lun=%x\n",
				bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));

			/*
			 *
			 * printk_Rlsi("cmd=0x%x,len=%d\n",cdb_data[0],csio->d
			 * xfer_len);
			 */

			/*
			 * Work around races.
			 */
			if ((csio->ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_INPROG) {
				scsi_done((void *)csio);
				return;
			}
			scbExtension = scsi_malloc(DeviceExtension, (void *)csio, FALSE);


			if (scbExtension == NULL) {
				/*
				 * printk_Rlsi("FREEZE -
				 * bus=%x,target=%x,lun=%x\n",
				 * bus,csio->ccb_h.target_id,csio->ccb_h.targe
				 * t_lun);
				 *
				 * printk_Rlsi("1 pend
				 * cmds=%ld\n",DeviceExtension->NumberOfPendin
				 * gScb);
				 */
				if (DeviceExtension->resource_shortage == FALSE) {
					DeviceExtension->resource_shortage = TRUE;
					xpt_freeze_simq(gastCardInfo[ControllerNumber].mide_sim, 1);
					DeviceExtension->FreezeCnt++;
				}
				csio->ccb_h.status = CAM_REQUEUE_REQ;
				scsi_done((void *)csio);
				FastIDEFireCommand(DeviceExtension);	/* feed queue */
				return;
			}
			if (DeviceExtension->NumberOfPendingScb >= M_IDE_MAX_COMMANDS) {
				/*
				 * printk_Rlsi("2 pend
				 * cmds=%ld\n",DeviceExtension->NumberOfPendin
				 * gScb);
				 */
				if (DeviceExtension->resource_shortage == FALSE) {
					DeviceExtension->resource_shortage = TRUE;
					xpt_freeze_simq(gastCardInfo[ControllerNumber].mide_sim, 1);
					DeviceExtension->FreezeCnt++;
				}
				csio->ccb_h.status |= CAM_REQUEUE_REQ;
				scsi_done((void *)csio);
				FastIDEFireCommand(DeviceExtension);	/* Vani: 0911 feed queue */
				return;
			}
			/*
			 * os specific scsi packet stuff copied to srb
			 * extension
			 */
			scbExtension->pstScsiCmd = (VOID *) csio;
			scbExtension->target = csio->ccb_h.target_id;
			scbExtension->lun = csio->ccb_h.target_lun;
			scbExtension->channel = bus;
			memcpy_Rlsi(scbExtension->scsi_cdb, cdb_data, csio->cdb_len);
			scbExtension->cdb_len = csio->cdb_len;
			/* sglist */
			if (csio->ccb_h.flags & CAM_SCATTER_VALID) {
				megaide_request_map_helper((void *)scbExtension,
				    (bus_dma_segment_t *) csio->data_ptr,
				    csio->sglist_cnt, 0);
			}
			/* single buffer */
			else {
				if (csio->ccb_h.flags & CAM_DATA_PHYS)
					scbExtension->request_buffer = (VOID *) csio->data_ptr;
				else {
					megaide_map_request(scbExtension);
				}
			}
			if (!megaide_atapi_startio(DeviceExtension, scbExtension)) {
				megaide_unmap_request(scbExtension);
				scsi_free(DeviceExtension, (void *)csio);
				csio->ccb_h.status = CAM_REQ_CMP_ERR;
				scsi_done((void *)csio);
			}
			return;
		}
		break;

	case READ_CAPACITY:
		{

			ULONG bytesPerBlock = IDE_SECTOR_SIZE;
			PREAD_CAPACITY_DATA readCapacity;
			ULONG lsize;

			AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("READ_CAPACITY - bus=%x,target=%x,lun=%x\n",
				bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
			memset_Rlsi(csio->data_ptr, 0, csio->dxfer_len);

			readCapacity = (PREAD_CAPACITY_DATA) csio->data_ptr;
			lsize = ldInfo->Sectors;
			readCapacity->BytesPerBlock = BIG_ENDIAN_ULONG(bytesPerBlock);
			readCapacity->LogicalBlockAddress = BIG_ENDIAN_ULONG(lsize);
			ccb_status = CAM_REQ_CMP;

		}

		break;


	case TEST_UNIT_READY:
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("TEST_UNIT_READY - bus=%x,target=%x,lun=%x\n",
			bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
		ccb_status = CAM_REQ_CMP;
		break;


	case MODE_SENSE:
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("MODE_SENSE - bus=%x,target=%x,lun=%x\n",
			bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
		ccb_status = CAM_DEV_NOT_THERE;
		break;

	case INQUIRY:
		{
			PINQUIRYDATA pInq;

			AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("INQUIRY - bus=%x,target=%x,lun=%x\n",
				bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
			memset_Rlsi(csio->data_ptr, 0, csio->dxfer_len);
			pInq = (PINQUIRYDATA) csio->data_ptr;

			pInq->Versions = 2;
			pInq->ResponseDataFormat = 2;
			((PUCHAR) pInq)[7] = 0x32;
			pInq->AdditionalLength = 0xFA;

			/*
			 * Fill Actual Vendor ID depending on
			 * SubsystemVendorID
			 */
			memmove_Rlsi(pInq->ProductId, OemProductId, 16);

			/*
			 * Fill Actual Product ID depending on
			 * SubsystemVendorID
			 */
			memmove_Rlsi(pInq->VendorId, OemVendorId, 8);

			pInq->ProductId[4] = ' ';	/* higher digit */
			pInq->ProductId[5] = (UCHAR) ldInfo->LogicalDriveNumber + '0';	/* lower digit */

			ccb_status = CAM_REQ_CMP;

		}

		break;

	case REQUEST_SENSE:
		{
			PUCHAR sense_data = (PUCHAR) & csio->sense_data;

			AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("REQUEST_SENSE - bus=%x,target=%x,lun=%x\n",
				bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
			memset_Rlsi(csio->data_ptr, 0, csio->dxfer_len);

			sense_data[0] = 0x70;
			sense_data[7] = 0x18;

			sense_data[8] = 0xFF;
			sense_data[9] = 0xFF;
			sense_data[10] = 0xFF;
			sense_data[11] = 0xFF;
			sense_data[19] = 0xFF;
			sense_data[20] = 0xFF;
			sense_data[21] = 0xFF;
			sense_data[22] = 0xFF;
			sense_data[23] = 0xFF;

			ccb_status = CAM_REQ_CMP;

		}

		break;


	case FORMAT_UNIT:

		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("FORMAT_UNIT - bus=%x,target=%x,lun=%x\n",
			bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
		ccb_status = CAM_DEV_NOT_THERE;

		break;


	case SYNCHRONIZE_CACHE:

		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("SYNCHRONIZE_CACHE - bus=%x,target=%x,lun=%x\n",
			bus, csio->ccb_h.target_id, csio->ccb_h.target_lun));
		/* FlushCache(DeviceExtension); */
		ccb_status = CAM_REQ_CMP;

		break;

	default:

		/* printk_Rlsi("NOT_SUPPORTED_CMD=0x%x\n",cdb_data[0]); */
		AR_DEBUG(AR_DEBUG_SCSI_CMD1, ("NOT_SUPPORTED_CMD=0x%x\n", cdb_data[0]));
		ccb_status = CAM_REQ_INVALID;
		break;

	}

	csio->ccb_h.status = ccb_status;
	scsi_done((void *)csio);
}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
u_int32_t
megiade_abort(PHW_DEVICE_EXTENSION DeviceExtension, struct ccb_abort * abort_ccb)
{
	u_int32_t ccb_status = CAM_REQ_INVALID;

	/* FIXME */

	return ccb_status;
}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
u_int32_t
megiade_reset_bus(PHW_DEVICE_EXTENSION DeviceExtension, struct ccb_resetbus * resetbus_ccb)
{
	u_int32_t ccb_status = CAM_REQ_CMP;
	int s;

	/* FIXME */
	s = splcam();
	FastIDEResetBus(DeviceExtension);
	splx(s);

	return ccb_status;

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
u_int32_t
megiade_reset_dev(PHW_DEVICE_EXTENSION DeviceExtension, struct ccb_resetdev * resetdev_ccb)
{
	u_int32_t ccb_status = CAM_REQ_CMP;

	/* FIXME */

	return ccb_status;

}



/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
megaide_intr(void *data)
{
	 AR_DEBUG(AR_DEBUG_INTERUPT, ("megaide_intr Called\n")); 
	/* spin_lock_irqsave(&io_request_lock, flags); */
	FastIDEInterrupt((PHW_DEVICE_EXTENSION) data);
	/* spin_unlock_irqrestore(&io_request_lock, flags); */
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
megaide_cam_poll(struct cam_sim * sim)
{
	PHW_DEVICE_EXTENSION DeviceExtension = cam_sim_softc(sim);
	AR_DEBUG(AR_DEBUG_INTERUPT, ("megaide_cam_poll Called\n"));

	FastIDEInterrupt(DeviceExtension);

}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

int
megaide_cam_attach(PHW_DEVICE_EXTENSION DeviceExtension)
{
	struct cam_devq *devq;
	int ControllerNumber;

	ControllerNumber = DeviceExtension->ucControllerId;

	/* debug_called(1); */

	if ((devq = cam_simq_alloc(M_IDE_MAX_COMMANDS)) == NULL)
		return (ENOMEM);

	/* allocate a sim */
	if ((gastCardInfo[ControllerNumber].mide_sim =
		cam_sim_alloc(megaide_cam_action,
		    megaide_cam_poll,
		    "mide",
		    DeviceExtension,
		    device_get_unit(gastCardInfo[ControllerNumber].pstDev),
		    1,
		    M_IDE_MAX_COMMANDS,
		    devq)) == NULL) {
		cam_simq_free(devq);
		printk_Rlsi("CAM SIM attach failed\n");
		return (ENOMEM);
	}
	/* register the bus ID so we can get it later */
	if (xpt_bus_register(gastCardInfo[ControllerNumber].mide_sim, 0)) {
		printk_Rlsi("CAM XPT bus registration failed\n");
		return (ENXIO);
	}
	megiade_cam_rescan_target(DeviceExtension, 0);

	return (0);
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
megaide_cam_detach(PHW_DEVICE_EXTENSION DeviceExtension)
{
	int ControllerNumber;

	ControllerNumber = DeviceExtension->ucControllerId;

	if (gastCardInfo[ControllerNumber].mide_sim != NULL) {
		xpt_free_path(gastCardInfo[ControllerNumber].mide_path);
		xpt_bus_deregister(cam_sim_path(gastCardInfo[ControllerNumber].mide_sim));
		cam_sim_free(gastCardInfo[ControllerNumber].mide_sim, TRUE);
	}
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
megiade_cam_rescan_target(PHW_DEVICE_EXTENSION DeviceExtension, int target)
{
	union ccb *ccb;
	int ControllerNumber;

	ControllerNumber = DeviceExtension->ucControllerId;

	/* debug_called(1); */

	if ((ccb = malloc(sizeof(union ccb), M_TEMP, M_WAITOK | M_ZERO)) == NULL) {
		printk_Rlsi("rescan failed (can't allocate CCB)\n");
		return;
	}
	if (xpt_create_path(&gastCardInfo[ControllerNumber].mide_path, xpt_periph,
		cam_sim_path(gastCardInfo[ControllerNumber].mide_sim),
		target,
		0) != CAM_REQ_CMP) {
		printk_Rlsi("CAM XPT create path failed\n");
		return;
	}
	xpt_setup_ccb(&ccb->ccb_h, gastCardInfo[ControllerNumber].mide_path, 5 /* priority (low) */ );
	ccb->ccb_h.func_code = XPT_SCAN_BUS;
	ccb->ccb_h.cbfcnp = megiade_cam_rescan_callback;
	ccb->crcn.flags = CAM_FLAG_NONE;
	xpt_action(ccb);

	/* scan is now in progress */
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
static void
megiade_cam_rescan_callback(struct cam_periph * periph, union ccb * ccb)
{
	free(ccb, M_TEMP);
}

void PassthruIOCTL_buffer_For_Map(PFAST_IDE_SCB_EXTENSION scbExtension, PUCHAR dataBuffer, ULONG dataSize, UCHAR Dir )
{
	struct ccb_scsiio *csio = (struct ccb_scsiio *) scbExtension->pstScsiCmd;

	/* Vani: 1114 :  Passthru UDMA change */
	csio->data_ptr =  (void *) dataBuffer;
	csio->dxfer_len = dataSize;
	if(Dir == DIR_IN)
		csio->ccb_h.flags |= CAM_DIR_IN;
	else if(Dir == DIR_OUT)
		csio->ccb_h.flags |= CAM_DIR_OUT;
	scbExtension->IoctlFlag = 1;
	megaide_map_request(scbExtension);

}
/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

static void
megaide_map_request(PFAST_IDE_SCB_EXTENSION scbExtension)
{
	struct ccb_scsiio *csio = (struct ccb_scsiio *) scbExtension->pstScsiCmd;
	bus_dmasync_op_t op;
	int s;
	int error;

	waitflag = 1;

	s = splsoftvm();

	error = bus_dmamap_load((bus_dma_tag_t) (scbExtension->mide_buffer_dmat),
	    (bus_dmamap_t) (scbExtension->mide_buffer_dmamap),
	    csio->data_ptr,
	    csio->dxfer_len,
	    megaide_request_map_helper,
	    scbExtension,
	    0);

	if (error == EINPROGRESS) {
		while (waitflag);
		printk_Rlsi("Awakened from wait \n");
		waitflag = 1;
	}
	splx(s);

	if ((csio->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN)
		op = BUS_DMASYNC_PREREAD;
	else
		op = BUS_DMASYNC_PREWRITE;

	bus_dmamap_sync((bus_dma_tag_t) (scbExtension->mide_buffer_dmat),
	    (bus_dmamap_t) (scbExtension->mide_buffer_dmamap),
	    op);

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
static void
megaide_request_map_helper(void *arg, bus_dma_segment_t * segs, int nseg, int error)
{
	PFAST_IDE_SCB_EXTENSION scbExtension = (PFAST_IDE_SCB_EXTENSION) arg;
	PSGL_ENTRY pSglEntry = scbExtension->SglEntry;
	int i;

	if (error != 0) {
		printk_Rlsi("Unexepected error 0x%x returned from bus_dmamap_load\n", error);
		waitflag = 0;
		return;
	}
	for (i = 0; i < nseg; i++) {
		pSglEntry[i].Physical.Address = (PVOID) segs[i].ds_addr;
		pSglEntry[i].Physical.Length = segs[i].ds_len;
	}

        /* Vani: 1114 :  Passthru UDMA change, when ioctl is issued the request buffer pointer is already used */
        if(!scbExtension->IoctlFlag)
		scbExtension->request_buffer = pSglEntry;
  	else
		scbExtension->ioctl_buffer = pSglEntry;

	scbExtension->sglem = nseg;
	scbExtension->buffer_is_mapped = TRUE;

	waitflag = 0;

}

void Generic_Unmap_Request(PFAST_IDE_SCB_EXTENSION scbExtension)
{
	megaide_unmap_request(scbExtension);
}
/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
static void
megaide_unmap_request(PFAST_IDE_SCB_EXTENSION scbExtension)
{
	struct ccb_scsiio *csio = (struct ccb_scsiio *) scbExtension->pstScsiCmd;
	bus_dmasync_op_t op;

	/* check that unmapping is necessary */
	if (scbExtension->buffer_is_mapped != TRUE)
		return;

	if ((csio->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN)
		op = BUS_DMASYNC_POSTREAD;
	else
		op = BUS_DMASYNC_POSTWRITE;

	bus_dmamap_sync((bus_dma_tag_t) scbExtension->mide_buffer_dmat,
	    (bus_dmamap_t) scbExtension->mide_buffer_dmamap,
	    op);

	bus_dmamap_unload((bus_dma_tag_t) scbExtension->mide_buffer_dmat,
	    (bus_dmamap_t) scbExtension->mide_buffer_dmamap);

	scbExtension->buffer_is_mapped = FALSE;

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
void
presync_sglist(PCHANNEL pChannel)
{
	bus_dmamap_sync((bus_dma_tag_t) pChannel->mide_sg_dmat,
	    (bus_dmamap_t) pChannel->mide_sg_dmamap,
	    BUS_DMASYNC_PREWRITE);
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
void
postsync_sglist(PCHANNEL pChannel)
{
	bus_dmamap_sync((bus_dma_tag_t) pChannel->mide_sg_dmat,
	    (bus_dmamap_t) pChannel->mide_sg_dmamap,
	    BUS_DMASYNC_POSTWRITE);
}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
void
presync_gatekeeper(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR command)
{
	bus_dmasync_op_t op;

	/*
	 * int s; int error;
	 *
	 * s = splsoftvm();
	 *
	 * waitgateflag = 1;
	 *
	 * error = bus_dmamap_load( DeviceExtension->GateKeeperBufferDmat,
	 * DeviceExtension->GateKeeperBufferDmaMap,
	 * DeviceExtension->GateKeeperBuffer,
	 * DeviceExtension->GateKeeperBufferLength,
	 * mide_gatekeeper_map_helper_channel, DeviceExtension, 0);
	 *
	 * if (error == EINPROGRESS) { while (waitgateflag);
	 * printk_Rlsi("Awakened from wait \n"); waitgateflag = 1; }
	 *
	 * splx(s);
	 *
	 */
	op = (command == READ) ? BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE;

	bus_dmamap_sync((bus_dma_tag_t) DeviceExtension->GateKeeperBufferDmat,
	    (bus_dmamap_t) DeviceExtension->GateKeeperBufferDmaMap,
	    op);
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
void
postsync_gatekeeper(PHW_DEVICE_EXTENSION DeviceExtension, UCHAR command)
{
	bus_dmasync_op_t op;

	op = (command == READ) ? BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE;

	bus_dmamap_sync((bus_dma_tag_t) DeviceExtension->GateKeeperBufferDmat,
	    (bus_dmamap_t) DeviceExtension->GateKeeperBufferDmaMap,
	    op);

	/*
	 * bus_dmamap_unload(
	 * DeviceExtension->GateKeeperBufferDmat,DeviceExtension->GateKeeperBu
	 * fferDmaMap);
	 */

}


/*
 * -------------------------------------------------------------------
 * Description ????
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
static void
mide_gatekeeper_map_helper_channel(void *arg, bus_dma_segment_t * segs, int nseg, int error)
{
	PHW_DEVICE_EXTENSION DeviceExtension = (PHW_DEVICE_EXTENSION) arg;

	DeviceExtension->GateKeeperPhysicalBuffer = segs->ds_addr;
	waitgateflag = 0;
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */


PFAST_IDE_SCB_EXTENSION
scsi_malloc(PHW_DEVICE_EXTENSION DeviceExtension, VOID * pstScsiCmd, BOOLEAN AllocIoctl)
{
	PFAST_IDE_SCB_EXTENSION scbExtension = NULL;
	int i = 0;
	struct ccb_hdr *ccb_h;
	int searchScb;

	ccb_h = (struct ccb_hdr *) pstScsiCmd;

	if (DeviceExtension->NumberOfPendingScb > M_IDE_MAX_COMMANDS)
		return NULL;

	if (AllocIoctl) {
		for (i = M_IDE_MAX_COMMANDS; i < (M_IDE_MAX_COMMANDS + MAX_IOCTL_MEM_ALLOC); i++) {

			if (DeviceExtension->scb_queue_in_use[i] == TRUE)
				continue;

			scbExtension = (PFAST_IDE_SCB_EXTENSION) ((ULONG) DeviceExtension->scb_queue +
			    (i * get_scb_extension_size()));
			break;
		}
	} else {
		DeviceExtension->last_free_scb = (DeviceExtension->last_free_scb >= M_IDE_MAX_COMMANDS) ? 0 : DeviceExtension->last_free_scb;
		for (i = 0, searchScb = DeviceExtension->last_free_scb; i < M_IDE_MAX_COMMANDS; i++, searchScb++) {

			searchScb = (searchScb == M_IDE_MAX_COMMANDS) ? 0 : searchScb;
			if (searchScb > M_IDE_MAX_COMMANDS)
				printk_Rlsi("Serious Error in Scb Index\n");

			if (DeviceExtension->scb_queue_in_use[searchScb] == TRUE)
				continue;

			scbExtension = (PFAST_IDE_SCB_EXTENSION) ((ULONG) DeviceExtension->scb_queue +
			    (searchScb * get_scb_extension_size()));
			break;
		}
		if (scbExtension) {
			i = searchScb;
			DeviceExtension->last_free_scb = i;
		}
	}

	if (scbExtension) {
		memset_Rlsi(scbExtension, 0, get_scb_extension_size());
		ccb_h->spriv_ptr0 = scbExtension;
		ccb_h->status = CAM_REQ_INPROG;
		ccb_h->status |= CAM_SIM_QUEUED;
		DeviceExtension->scb_queue_in_use[i] = TRUE;
		scbExtension->scb_queue_index = i;
		scbExtension->mide_buffer_dmat = DeviceExtension->mide_buffer_dmat;
	}
	return (scbExtension);
}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
scsi_free(PHW_DEVICE_EXTENSION DeviceExtension, VOID * pstScsiCmd)
{
	PFAST_IDE_SCB_EXTENSION scbExtension;
	struct ccb_hdr *ccb_h;


	ccb_h = (struct ccb_hdr *) pstScsiCmd;

	if (!ccb_h->spriv_ptr0)
		return;
	scbExtension = (PFAST_IDE_SCB_EXTENSION) ccb_h->spriv_ptr0;
	DeviceExtension->scb_queue_in_use[scbExtension->scb_queue_index] = FALSE;

	ccb_h->spriv_ptr0 = 0;

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
scsi_done(VOID * pstScsiCmd)
{
	struct ccb_hdr *ccb_h;
	XPT_DONE megaide_xpt_done;

	ccb_h = (struct ccb_hdr *) pstScsiCmd;
	megaide_xpt_done = (XPT_DONE) ccb_h->spriv_ptr1;
	megaide_xpt_done((union ccb *) ccb_h);
	ccb_h->spriv_ptr1 = 0;
}


/*
 * XPT_DONE
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
megaide_complete_command(PHW_DEVICE_EXTENSION DeviceExtension, PFAST_IDE_SCB_EXTENSION ScbExtension, int result)
{
	struct ccb_hdr *ccb_h;
	struct ccb_scsiio *csio;
	u_int32_t ccb_status;

	ccb_h = (struct ccb_hdr *) ScbExtension->pstScsiCmd;
	csio = (struct ccb_scsiio *) ScbExtension->pstScsiCmd;

	if (!ccb_h) {
		return;
	}
	megaide_unmap_request(ScbExtension);

	csio->scsi_status = (u_int8_t) (result & 0xFFFF);
	ccb_status = (u_int32_t) (result >> 16);

	switch (ccb_status) {
	case MIDE_STATUS_OK:
		ccb_h->status = CAM_REQ_CMP;
		break;
	case MIDE_STATUS_BUS_BUSY:
		ccb_h->status = CAM_SCSI_BUSY;
		break;
	case MIDE_STATUS_BAD_TARGET:
		ccb_h->status = CAM_PATH_INVALID;
		break;
	case MIDE_STATUS_ABORT:
		ccb_h->status = CAM_REQ_ABORTED;
		break;
	case MIDE_STATUS_ERROR:
		ccb_h->status = CAM_REQ_CMP_ERR;
		break;
	case MIDE_STATUS_RESET:
		ccb_h->status = CAM_REQUEUE_REQ;
		break;
	}

	if ((DeviceExtension->resource_shortage == TRUE) && (ScbExtension->scb_queue_index < M_IDE_MAX_COMMANDS)) {
		/* ccb_h->status |= CAM_RELEASE_SIMQ; */
		if (DeviceExtension->FreezeCnt > 1)
			printk_Rlsi("Freeze count  %d \n", DeviceExtension->FreezeCnt);
		/*
		 *
		 * xpt_release_simq(gastCardInfo[DeviceExtension->ucController
		 * Id].mide_sim, TRUE);
		 */
		ccb_h->status |= CAM_RELEASE_SIMQ;
		DeviceExtension->FreezeCnt--;
		DeviceExtension->resource_shortage = FALSE;

	}
	scsi_free(DeviceExtension, ccb_h);
	scsi_done(ccb_h);

}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
fill_sense_buffer(PFAST_IDE_SCB_EXTENSION ScbExtension, UCHAR error_byte, BOOLEAN blReadWriteError, ULONG error_sector)
{
	struct ccb_scsiio *csio;
	PUCHAR sense_data;

	csio = (struct ccb_scsiio *) ScbExtension->pstScsiCmd;
	sense_data = (PUCHAR) & csio->sense_data;

	if (error_byte & (ABRT_ERR | ID_ERR | MCR_ERR)) {
		if (sense_data) {
			sense_data[0] = 0x70;
			sense_data[2] = ABORTED_COMMAND;
		}
	}
	if (error_byte & (ECC_ERR | ICRC_ERR)) {
		if (sense_data) {
			sense_data[0] = blReadWriteError ? 0x70 : 0xF0;
			sense_data[2] = blReadWriteError ? 0 : MEDIUM_ERROR;

			/* set up LBA address */
			sense_data[3] = (error_sector >> 24) & 0xff;
			sense_data[4] = (error_sector >> 16) & 0xff;
			sense_data[5] = (error_sector >> 8) & 0xff;
			sense_data[6] = error_sector & 0xff;
			sense_data[7] = 6;


			if (blReadWriteError == 0) {	/* read error */
				sense_data[12] = 0x11;
				sense_data[13] = 0x0;
			}
			if (blReadWriteError == 1) {	/* write error */
				sense_data[12] = 0x0C;
				sense_data[13] = 0x0;
			}
		}
	}
}

/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */
ULONG
GetPhysicalAddressAndLength(void *request_buffer, unsigned short sglem, ULONG ulDataOffset, ULONG * pulContiguousLength)
{
	ULONG physical_address = 0;
	PSGL_ENTRY pSglEntry = (PSGL_ENTRY) request_buffer;
	ULONG ulCurLength, ulSglInd;

	*pulContiguousLength = 0;
	physical_address = 0;

	for (ulSglInd = 0, ulCurLength = 0; ulSglInd < sglem; ulSglInd++) {
		if ((ulCurLength + pSglEntry[ulSglInd].Physical.Length) > ulDataOffset) {	/* This is the SGL
												 * Element */
			*pulContiguousLength = pSglEntry[ulSglInd].Physical.Length - (ulDataOffset - ulCurLength);
			physical_address = (ULONG) pSglEntry[ulSglInd].Physical.Address + (ulDataOffset - ulCurLength);
			break;
		}
		ulCurLength += pSglEntry[ulSglInd].Physical.Length;
	}
	return physical_address;
}


/*
 * -------------------------------------------------------------------
 * Description
 *
 * Arguments
 *
 * Return Value
 * -------------------------------------------------------------------
 */

void
print_sglist(void *request_buffer, unsigned short sglem, int tag)
{
	int i;
	PSGL_ENTRY pstSgList = (PSGL_ENTRY) request_buffer;

	if (sglem == 0)
		return;
	printk_Rlsi("b-%d\n", tag);
	for (i = 0; i < sglem; i++) {
		printk_Rlsi("i=%d,ds_addr=%lx,ds_len=%x\n", i,
		    (ULONG) pstSgList[i].Physical.Address, (ULONG) pstSgList[i].Physical.Length);
	}
	printk_Rlsi("a-%d\n", tag);
}

/*
 * -------------------------------------------------------------------
 * Description : Make the data Transfer length available to closed source.
 *
 * Arguments: Pointer to FAST_IDE_SCB_EXTENSION structure.
 *
 * Return Value: ULONG Author : Vani
 * -------------------------------------------------------------------
 */
ULONG
Get_DataTransfer_Length(PFAST_IDE_SCB_EXTENSION scbExtension)
{
	struct ccb_scsiio *csio = (struct ccb_scsiio *) scbExtension->pstScsiCmd;
	return csio->dxfer_len;
}
