#ifndef __BP_IOCTLS_H__
#define __BP_IOCTLS_H__

/*------------- application to Megaide interface structures - start ------------*/
/*BP IOCTL OPCODE's*/
#define AR_GET_BPD_INFO			101
#define AR_BP_STATUS_INFO 		102
#define AR_GET_NVRAM_EVENT_INFO		103
#define AR_CLEAR_NVRAM_EVENT_LOG	104
#define AR_GET_NVRAM_EVENT_LOG		105
#define AR_BP_OPTIONS			106
#define AR_BP_BEEP_SILENCE		107
#define AR_ADD_NVRAM_EVENT_LOG	        108
#define AR_BLINK_LED_0			109
#define AR_ON_LED_0			110
#define AR_OFF_LED_0			111
#define AR_BLINK_LED_1			112
#define AR_ON_LED_1			113
#define AR_OFF_LED_1			114
#define AR_IDENTIFY_DEVICE_LED		115
#define AR_RESET_DEVICE_LED		116

/*Success Return Values*/
#define BPD_ERROR_STATUS_SUCCESS	0x00    /*No Error*/
#define BPD_ERROR_DEVICE_NOT_DOCKED	0x01    /*HDD Not Docked*/
#define BPD_ERROR_BUFFER_INSUFFICIENT	0x02    /*Insufficient Buffer*/
                                                /* - App can resend the request with more buffer*/
#define BPD_ERROR_EVENT_LOG_NOT_PRESENT	0x03    /* The requested Event log is not present / No more event log
                                                 * - This is an indication for the apps to stop requesting
                                                 * - more event logs. So this is not really an error.
                                                 */

/*Error Return Values*/
#define BPD_ERROR_UNKNOWN_ERROR			0x80    /*Unknown Error*/
#define BPD_ERROR_FUNCTION_NOT_IMPLEMENTED	0x81    /*Function Not Implemented*/
#define BPD_ERROR_DEVICE_ADDRESS_NOT_PRESENT	0x82    /* Illegal Register Address*/
#define BPD_ERROR_EVENT_RECORD_MISMATCH		0x83    /* Event Record Mismatch - to be returned for ClearEventLog fn.*/
#define BPD_ERROR_EVENT_LOG_FULL		0x84    /* All event logs are used up.*/
#define BPD_ERROR_BPD_INITIALIZING		0x85    /* BP driver is now on initialization.*/
#define BPD_ERROR_DRIVER_NOT_EXIST		0x86    /* BP driver is not loaded.*/


/*Macros to find out if a return value is success or error*/
#define BPD_SUCCESS(status) (!(status & 0x80))


# ifndef MAX_CHANNELS_PER_CONTROLLER
# define MAX_CHANNELS_PER_CONTROLLER 2
#endif


/*Capabilities Definitions*/
#define BPD_CAP_NONE						0x0000
#define BPD_CAP_NVRAM_EVENT_LOG		0x0001
#define BPD_CAP_LED			0x0002
#define BPD_CAP_BUZZER			0x0004
#define BPD_CAP_FAN			0x0008
#define BPD_CAP_TEMPERATURE		0x0010
#define BPD_CAP_BPLAYER_EVENT_LOG	0x0020

/*NVRAM failure codes*/
#define NVRAM_HDD_STATUS_GOOD			0
#define NVRAM_HDD_STATUS_TIMEOUT        	1
#define NVRAM_HDD_STATUS_BUSY			2
#define NVRAM_HDD_STATUS_LOST_POWER		3
#define NVRAM_HDD_STATUS_EXCEED_ERROR_COUNT     4
#define NVRAM_HDD_STATUS_FORCE_FAIL_APP		5
#define NVRAM_HDD_STATUS_FORCE_ONLINE_APP	6
#define NVRAM_HDD_STATUS_FORCE_BY_SOFTWARE	8
#define NVRAM_HDD_STATUS_MEDIA_ERROR		9

/*NVRAM Command Failed Status*/
#define NVRAM_CMD_FAILED_FIRST_TRY			0
#define NVRAM_CMD_FAILED_AT_RETRY			1
#define NVRAM_CMD_FAILED_AT_FIRST_TRY_ON_MIRROR		2
#define NVRAM_CMD_FAILED_AT_RETRY_ON_MIRROR		3
#define NVRAM_CMD_FAILED_WRITE_RECOVERED_DATA		4
#define NVRAM_CMD_FAILED_RECOVERED_WRITE_BUT_READ_PASSED	5
#define NVRAM_CMD_FAILED_RECOVERED_WRITE_AND_READ_FAILED	6

/*NVRAM Command Succeed Status*/
#define NVRAM_CMD_SUCCEED_AFTER_RETRY			1
#define NVRAM_CMD_SUCCEED_MIRROR_FIRST_TRY		2
#define NVRAM_CMD_SUCCEED_MIRROR_RETRY			3
#define NVRAM_CMD_SUCCEED_RECOVERED_DATA_ON_TARGET	4

/*NVRAM Command Type*/
#define NVRAM_CMD_READ					0
#define NVRAM_CMD_WRITE_FOR_REASSIGNMENT		1
#define NVRAM_CMD_WRITE					2
#define NVRAM_CMD_IOCTL					3
#define NVRAM_CMD_UNKNOWN 				0xff


typedef enum
{
	BP_BEEP_OFF=0,
	BP_BEEP_ON=1,
	BP_BEEP_BOOT_NA=3
}
BP_BEEP;
typedef enum
{
	BP_LED_OFF=0,
	BP_LED_ON=1,
	BP_LED_BLINK=2
}
BP_LED;

typedef struct _BP_GET_BPD_INFO
{
  USHORT MajorVersion;
  USHORT MinorVersion;
  ULONG Capability;
}
__attribute__((packed))
BP_GET_BPD_INFO, *PBP_GET_BPD_INFO;

typedef struct _BP_STATUS_INFO
{
	BP_BEEP		BeepStatus;
	BP_LED		LedStatus[MAX_CHANNELS_PER_CONTROLLER];
	UCHAR		NewEvent_Notify;

}
__attribute__((packed))
BP_STATUS_INFO, *PBP_STATUS_INFO;

typedef struct _BP_OPTIONS
{
	ULONG	BeepEnable;
}
__attribute__((packed))
BP_OPTIONS, *PBP_OPTIONS;

typedef struct _BP_GET_NVRAM_EVENT_INFO
{
  USHORT UsedEventLogRecords;
  USHORT FreeEventLogRecords;
  UCHAR  MajorVersion;
  UCHAR  MinorVersion;
  UCHAR  Reserved[2];
}
__attribute__((packed))
BP_GET_NVRAM_EVENT_INFO, *PBP_GET_NVRAM_EVENT_INFO;

typedef struct _REGISTER_ACCESS
{
  ULONG RegisterAddress;
  UCHAR RegisterValue;
  UCHAR Reserved[3];
}
__attribute__((packed))
REGISTER_ACCESS, *PREGISTER_ACCESS;

/*LED Control Buffer*/
typedef struct _LED_CONTROL
{
	ULONG  LEDNumber;
	USHORT LEDStatus;
}
__attribute__((packed))
LED_CONTROL, *PLED_CONTROL;

typedef struct _IDE_BUS_ISOLATION {
	USHORT	Channel;
	USHORT	Isolation;
}
__attribute__((packed))
IDE_BUS_ISOLAION, *PIDE_BUS_ISOLATION;

typedef struct _EVENT_LOG_INFO
{
	USHORT UsedEventLogRecords;
	USHORT FreeEventLogRecords;
	UCHAR  MajorVersion;
	UCHAR  MinorVersion;
	UCHAR  Reserved[2];
}
__attribute__((packed))
EVENT_LOG_INFO, *PEVENT_LOG_INFO;


/* --------------- common defines------ */
/* The IOCTL function codes from 0x800 to 0xFFF are for customer use.
 * Device type           -- in the "User Defined" range."
 */
#define BPD_TYPE 80000

#define IOCTL_BPD_INIT_BACKPLANE	_IOWR( BPD_TYPE, 0x00, sizeof(struct _INIT_BACKPLANE))
#define IOCTL_BPD_READ_REGISTER		_IOWR( BPD_TYPE, 0x01, sizeof(struct _REGISTER_ACCESS))
#define IOCTL_BPD_WRITE_REGISTER	_IOW ( BPD_TYPE, 0x02, sizeof(struct _REGISTER_ACCESS))
#define IOCTL_BPD_LED_CONTROL		_IOW ( BPD_TYPE, 0x03, sizeof(struct _LED_CONTROL))
#define IOCTL_BPD_BEEP_CONTROL		_IOW ( BPD_TYPE, 0x04, sizeof(ULONG))
#define IOCTL_BPD_DOCK_STATUS		_IOWR( BPD_TYPE, 0x05, sizeof(ULONG))
#define IOCTL_BPD_IDE_BUS_ISOLATION	_IOW ( BPD_TYPE, 0x06, sizeof(struct _IDE_BUS_ISOLATION))

/*NVRAM Event Log Interface*/
#define IOCTL_BPD_GET_EVENT_LOG_INFO _IOWR( BPD_TYPE, 0x07, sizeof(struct _EVENT_LOG_INFO))
#define IOCTL_BPD_CLEAR_EVENTS     	_IOW ( BPD_TYPE, 0x08, sizeof(ULONG))
#define IOCTL_BPD_ADD_EVENT 		_IOW ( BPD_TYPE, 0x09, sizeof(struct _ADD_EVENT))
#define IOCTL_BPD_GET_EVENT 		_IOWR( BPD_TYPE, 0x0A, sizeof(struct _EVENT_LOG_ENTRY))

/* DEBUG purpose*/
#define IOCTL_BPD_INITIALIZE		_IO  ( BPD_TYPE, 0x80 )
#define IOCTL_BPD_GET_ERROR_CODE	_IO( BPD_TYPE, 0x81 )
#define IOCTL_BPD_BEEP_STATUS		_IOR ( BPD_TYPE, 0x82, sizeof(ULONG))
#define IOCTL_READ_EEPROM		_IOWR( BPD_TYPE, 0x83, 1024 )
#define IOCTL_WRITE_EEPROM		_IOW ( BPD_TYPE, 0x84, 1024 )
#define IOCTL_ERASE_EEPROM		_IOWR( BPD_TYPE, 0x85, sizeof(struct BPD_WRITE_BUFFER_IN))
#define IOCTL_SET_FAULT_LED		_IOW ( BPD_TYPE, 0x86, sizeof(ULONG))
#define IOCTL_GET_VERSION		_IOR ( BPD_TYPE, 0x87, DRIVER_VERSION_LENGTH )

/*Bit Masks for Event Message ID*/
#define BPD_EVENT_MESSAGE_SEVERITY_MASK     0xC0
#define BPD_EVENT_MESSAGE_ORIGINATOR_MASK   0x20
#define BPD_EVENT_MESSAGE_ID_MASK           0x1F

/*Macros to Convert Mask to Integer values*/

#define BPD_SEVERITY_FROM_MESSAGE ( UCHAR Message ) \
 ( (Message & BPD_EVENT_MESSAGE_SEVERITY_MASK )  >> 6 )

#define BPD_ORIGINATOR_FROM_MESSAGE ( UCHAR Message )  \
    ( (Message & BPD_EVENT_MESSAGE_ORIGINATOR_MASK )  >> 5 )

#define BPD_ID_FROM_MESSAGE ( UCHAR Message )  \
    ( (Message & BPD_EVENT_MESSAGE_ID_MASK ) )

#define	BPD_EVENT_ID_CLEAR_BY_BIOS			0x00
#define	BPD_EVENT_ID_CLEAR_BY_DRIVER		0x20

#define	BPD_EVENT_ID_DOCK_DETECT_BY_BIOS	0x11
#define	BPD_EVENT_ID_DOCK_DETECT_BY_DRIVER	0x31

/* Fail Reason definitions*/
#define HDF_NO_ERROR              0x00
#define HDF_TIMEOUT               0x01
#define HDF_BUSY                  0x02
#define HDF_LOST_POWER            0x03
#define HDF_ERROR                 0x04
#define HDF_APP_FORCED_FAIL       0x05
#define HDF_APP_FORCED_ONLINE     0x06
#define HDF_MEDIUM_ERROR          0x09
#define HDF_OTHER_ERROR           0x0A

/*Command Status*/
#define HDCS_FIRST_TRY            0x00
#define HDCS_RETRY                0x01
#define HDCS_MIRROR_FIRST_TRY     0x02
#define HDCS_MIRROR_RETRY         0x03

/*Command Type*/
#define HDCT_READ			0x00
#define HDCT_WRITE_FOR_REASSIGNMENT	0x01
#define HDCT_WRITE			0x02
#define HDCT_IOCTL			0x03

/* ADD Event Log Entry Structure
 * Follows Event Log Specification Version 1.0
 */
#define CLEAR_EVENT_ID				0x20
#define ATA_COMMAND_FAILURE_CRITICAL		0xE1
#define ATA_COMMAND_FAILURE_INFORMATION		0x21
#define IDE_REGISTERS				0x22
#define BACKPLANE_API_CALL_FAILURE		0xE3
#define POWER_STATE_CHANGE			0xA4
#define RAID_STATUS_CHANGE			0xA5
#define SMART_STATUS_CHANGE			0xA6
#define ATA_COMMAND_SUCCESS			0x27

#pragma pack(1)

/* Each Structure inside the union
 * shall be exactly 11 bytes
 */
typedef union _EVENT_LOG_DETAIL
{
    struct _CLEAR_EVENT
    {
      UCHAR  Reserved;
      USHORT LastCriticalErrorCount;
      USHORT LastWarningErrorCount;
      USHORT LastInfoErrorCount;
      ULONG  LastClearEventTimeStamp;
    } ClearEvent;

    struct _ATA_CMD_FAILURE
    {
      UCHAR  HDDNumber;
      ULONG  SectorNumber;
      UCHAR  FailReason;
      UCHAR  LastSMARTStatus;
      UCHAR  CommandStatus;
      UCHAR  CommandType;
      UCHAR  Reserved[2];
    } ATACmdFailure;

    struct _IDE_REGISTERS
    {
      UCHAR  Error;
      UCHAR  SectorCount;
      UCHAR  Sector;
      UCHAR  CylinderHigh;
      UCHAR  CylinderLow;
      UCHAR  Head;
      UCHAR  AlternateStatus;
      UCHAR  Command;
      UCHAR  Reserved[3];
    } IDERegisters;

    struct _BIOS_BP_API_FAILURE	/*Never called from Driver.*/
    {
      USHORT ax;
      USHORT bx;
      USHORT cx;
      USHORT dx;
      USHORT axe; /*Error Returned from OEM Layer*/
      UCHAR  Reserved;
    } BIOSBPAPIFailure;

    struct _DRIVER_BP_API_FAILURE
    {
      UCHAR  IOCTLCode;
      ULONG  Parameter1;
      ULONG  Parameter2;
      USHORT OEMError;
    } DriverBPAPIFailure;

    struct _POWER_STATE_CHANGE
    {
      UCHAR  ControllerNumber;
      UCHAR  HDDNumber;
      UCHAR  OldPowerState;
      UCHAR  NewPowerState;
      UCHAR  ErrorRegister;
      UCHAR  StatusRegister;
      UCHAR  Reserved[5];

    } PowerManagementFailure;

    struct _RAID_STATUS_CHANGE
    {
      UCHAR  HDDNumber;
      UCHAR  DriveStatus;
      UCHAR  DriveErrorCount;
      UCHAR  FailReason;
      UCHAR  ArrayID;
      UCHAR  RAIDLevel;
      UCHAR  ArrayStatus;
      UCHAR  Reserved[4];
    } RAIDStatusChange;

    struct _SMART_STATUS_CHANGE
    {
      UCHAR  HDDNumber;
      UCHAR  CylinderHigh;
      UCHAR  CylinderLow;
      UCHAR  StatusRegister;
      UCHAR  Reserved[7];
    } SMARTStatusChange;

   struct _ATA_CMD_SUCCESS_STATUS
    {
      UCHAR  HDDNumber;
      ULONG  SectorNumber;
      UCHAR  CommandStatus;
      UCHAR  CommandType;
      UCHAR  Reserved[4];
    } ATACmdSuccessStatus;

}EVENT_LOG, *PEVENT_LOG;

typedef struct _ADD_EVENT
{
  UCHAR MessageID;     /*Formed according to specification*/
  EVENT_LOG EventLog;
}
ADD_EVENT, *PADD_EVENT;

/* Get Event Log Entry Structure
 * Follows Event Log Specification Version 1.0
 */
typedef struct _EVENT_LOG_ENTRY
{

  ULONG TimeStamp;
  ADD_EVENT Event;

}
EVENT_LOG_ENTRY, *PEVENT_LOG_ENTRY;

typedef struct _BP_LOG_ENTRY
{
     ULONG index;
     EVENT_LOG_ENTRY  EventLogEntry;
}
BP_LOG_ENTRY, *PBP_LOG_ENTRY;

#pragma pack()

#endif /* __BP_IOCTLS_H__*/
/*vi set ts=8:*/

