/* MegaIDE driver test utility */

#define SUPPORT_IRCD3		0xFF
#define BACKPLANE_SUPPORT	0xFF 

#include "ideutil.h"
#include "mp_hd_ioctls.h"
#ifdef SUPPORT_IRCD3
#include "../lib/inc/megaide_ircd3.h"
#endif
#include "bp_ioctl.h"
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define CHAR_DEVICE_FILE		"/dev/mide"

static int select_target(void);
static void print_ircd_info(PNEW_IRCD IRCDData);

#define INITIALIZE_IOCTL(ptr, inlen, outlen, opcode) \
{ \
		u_long size; \
		size = (outlen > inlen) ? outlen : inlen; \
		ptr = malloc( (size + sizeof(ioctl_header) ) ); \
		if (!ptr) { \
			printf("Memory allocation Failed\n"); \
			exit(0); \
		} \
}
void
initialize_ioctl(char_megaide_ioctl * megaide_ioctl, u_long inlen, u_long outlen, u_long opcode)
{
	u_long size;
	pchar_ioctl_data pwdata;

	INITIALIZE_IOCTL(megaide_ioctl->wdata, inlen, outlen, opcode);
	size = (outlen > inlen) ? outlen : inlen;
	bzero((void *)megaide_ioctl->wdata, (size + sizeof(ioctl_header)));
	megaide_ioctl->inlen = inlen;
	megaide_ioctl->outlen = outlen;
	pwdata = (pchar_ioctl_data) megaide_ioctl->wdata;
	pwdata->sic.header_len = sizeof(ioctl_header);
	pwdata->sic.data_len = size;
	pwdata->sic.opcode = opcode;
	memcpy(megaide_ioctl->acSignature, "MEGANIT", 7);
	megaide_ioctl->iAdapNumber = 1;
}

#define FREE_MEM(ptr)		free(ptr)

int
select_source(void)
{
	char read_buffer[20];
	int return_selection = 0;
	do {
		printf(" \n\n");
		printf(" Select your Source Drive\n");
		printf(" (0) - Primary Master \n");
		printf(" (1) - Primary Slave \n");
		printf(" (2) - Secondary Master \n");
		printf(" (3) - Secondary Slave \n");
		write(1, " Selection(0-4)     :", 21);
		read(0, read_buffer, 21);
		return_selection = atoi(read_buffer);
	} while (return_selection > 3);
	return (return_selection);
}

int
select_target(void)
{
	char read_buffer[20];
	int return_selection = 0;
	do {
		printf(" \n\n");
		printf(" Select your Target Drive\n");
		printf(" (0) - Primary Master \n");
		printf(" (1) - Primary Slave \n");
		printf(" (2) - Secondary Master \n");
		printf(" (3) - Secondary Slave \n");
		write(1, " Selection(0-4)  : ", 21);
		read(0, read_buffer, 21);
		return_selection = atoi(read_buffer);
	} while (return_selection > 3);
	return (return_selection);
}

u_char
select_log_drive(void)
{
	char read_buffer[20];
	u_char return_selection = 0;

	do {
		printf(" \n\n");
		printf(" Select your Logical Drive\n");
		write(1, " Selection(0-7)  : ", 21);
		read(0, read_buffer, 21);
		return_selection = (u_char) atoi(read_buffer);
	} while (return_selection > 7);

	return (return_selection);
}

UCHAR 
select_smart_action(void)
{
	char read_buffer[20];
	int return_selection = 0;

	do {
		printf(" \n\n");
		printf(" Select Smart Action\n");
		printf(" (0) - Smart Enable \n");
		printf(" (1) - Smart Disable \n");
		printf(" (2) - Smart Read Data \n");
		printf(" (3) - Smart Read Status \n");
		write(1, " Selection(0-4)  : ", 21);
		read(0, read_buffer, 21);
		return_selection = atoi(read_buffer);
	} while (return_selection > 3);

	switch (return_selection) {
	case 0:
		return 0xD8;
	case 1:
		return 0xD9;
	case 2:
		return 0xD0;
	case 3:
		return 0xDA;
	}
	return 0xD8;
}

int 
select_option(void)
{
	char read_buffer[20];
	int return_selection = 0;

	do {
		printf(" \n\n");
		printf(" Select your option\n");
		printf(" (0) - Online \n");
		printf(" (1) - Offline \n");
		printf(" (2) - Hotspare \n");
		printf(" (3) - Fail\n");
		write(1, " Selection(0-4)     :", 21);
		read(0, read_buffer, 21);
		return_selection = atoi(read_buffer);
	} while (return_selection > 3);


	return (return_selection);
}

//set_options
void 
set_options(int file_handle)
{
	PMP_RAID_OPTIONS driver_options;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_RAID_OPTIONS), 0, AR_SET_OPTIONS);

	pCharIoctl = megaide_ioctl.wdata;
	driver_options = (PMP_RAID_OPTIONS) pCharIoctl->data_buffer;
	driver_options->ulRebuildRate = 50;
	driver_options->ulIdentical = 1;
	driver_options->ulAutoRebuild = 0;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Set Driver Options is Success\n");
	} else {
		printf("Set Driver Options call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* set_options() */


//driver_probe
void
driver_probe(int file_handle)
{
	PMP_PROBE_AMIRAID probe_data;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_PROBE_AMIRAID), AR_PROBE_AMIRAID);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = megaide_ioctl.wdata;
		probe_data = (PMP_PROBE_AMIRAID) pCharIoctl->data_buffer;
		printf("Probe Information => %s, Controller Index = %d\n",
		    (char *)probe_data->acData, probe_data->usControllerId);
	} else {
		printf("DriverProbe call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);

}				/* driver_probe */

//get_driver_version
void
get_driver_version(int file_handle)
{
	PMP_IDE_VERSION megaide_version;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;


	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_IDE_VERSION), AR_GET_VERSION);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = megaide_ioctl.wdata;
		megaide_version = (PMP_IDE_VERSION) pCharIoctl->data_buffer;
		printf("Driver Version is %1d.%1d-%1ld\n",
		    megaide_version->usMajorVer,
		    megaide_version->usMinorVer,
		    megaide_version->ulBuild);
	} else {
		printf("GetDriverVersion call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_driver_version() */

//get_raid_status
void 
get_raid_status(int file_handle)
{
	PMP_RAID_STATUS_INFO_EX raid_status;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	int i;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_RAID_STATUS_INFO_EX), AR_RAID_STATUS_INFO_EX);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = megaide_ioctl.wdata;
		raid_status = (PMP_RAID_STATUS_INFO_EX) pCharIoctl->data_buffer;
		printf("LOG DRV Status -   ");
		for (i = 0; i < 4; i++) {
			printf("[%d]=0x%x  ", i, raid_status->ArrayStatus[i]);
		}
		printf("\n\t\t   ");

		for (i = 4; i < MAX_LOGICAL_DRIVES; i++) {
			printf("[%d]=0x%x  ", i, raid_status->ArrayStatus[i]);
		}

		printf("\nPhysical Drive Status  - ");
		for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
			printf("[%d]=0x%x  ", i, raid_status->PhyStatus[i]);

		}
		printf("\nCheck Error Log Status - ");
		for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
			printf("[%d]=0x%lx  ", i, (ULONG) raid_status->CheckErrorLog[i]);

		}
		printf("\nLOG DRV Activity - ");
		for (i = 0; i < MAX_DRIVES_PER_CONTROLLER; i++) {
			printf("[%d]=0x%x  ", i, raid_status->CurrentArrayActivity[i]);
		}

		printf("\n\t\t   ");

		for (i = 4; i < MAX_LOGICAL_DRIVES; i++) {
			printf("[%d]=0x%x  ", i, raid_status->CurrentArrayActivity[i]);
		}

		printf("\n");

	} else {
		printf("Get RAID Status call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_raid_status() */

void 
get_physical_drv_info(int file_handle)
{
	PIDE_PHYSICAL_DRIVE_INFO raid_info;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	u_char i;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(IDE_PHYSICAL_DRIVE_INFO), AR_GET_PHYSICAL_DRIVE_INFO);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {

		pCharIoctl = megaide_ioctl.wdata;
		raid_info = (PIDE_PHYSICAL_DRIVE_INFO) pCharIoctl->data_buffer;

		printf("Phy Drv Present -  ");
		for (i = 0; i < 4; i++) {
			printf("[%d]=0x%x  ", i, raid_info->PhysicalDrives[i].ucIsPowerConnected);
		}
		printf("\n");

	} else {
		printf("Get Physical Drive Information call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_physical_drv_info() */

//get_controller_info
void 
get_controller_info(int file_handle)
{
	PMP_CONTROLLER_DATA controller_data;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_CONTROLLER_DATA), AR_GET_CONTROLLER_INFO);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		controller_data = (PMP_CONTROLLER_DATA) pCharIoctl->data_buffer;
		printf("Get Controller Information is Success\n");
	} else {
		printf("Get Controller Information call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_controller_info() */


//get_raid_info
void 
get_raid_info(int file_handle)
{
	PMP_IDE_RAID_INFO raid_info;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_IDE_RAID_INFO), AR_RAID_INFO);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		raid_info = (PMP_IDE_RAID_INFO) pCharIoctl->data_buffer;

		printf("Get Raid Information is Success\n");
	} else {
		printf("Get Raid Information call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_raid_info() */


//get_sparepool_info
void 
get_sparepool_info(int file_handle)
{
	PMP_SPAREPOOL_DATA sparepool_info;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0, sizeof(MP_SPAREPOOL_DATA), AR_GET_SPAREPOOL_INFO);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		sparepool_info = (PMP_SPAREPOOL_DATA) pCharIoctl->data_buffer;
		printf("Get Spare Pool Information is Success\n");
	} else {
		printf("Get Spare Pool Information call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* get_sparepool_info() */


u_long 
send_task(int file_handle, MP_ACTION_TYPES action, u_char sel_log_drive, MP_DRIVE_ID drive_source, MP_DRIVE_ID drive_target, ULONG * ulBlockCount)
{
	PMP_SEND_ACTIVITY_EX send_activity_ex;
	u_long return_code;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_SEND_ACTIVITY_EX), sizeof(MP_SEND_ACTIVITY_EX), AR_SEND_ACTIVITY_EX);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	send_activity_ex = (PMP_SEND_ACTIVITY_EX) pCharIoctl->data_buffer;
	send_activity_ex->Action = action;
	send_activity_ex->SourceDriveId = drive_source;
	send_activity_ex->TargetDriveId = drive_target;
	send_activity_ex->ulBlockCount = 0xffffffff;
	send_activity_ex->ucLogicalDriveId = sel_log_drive;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		send_activity_ex = (PMP_SEND_ACTIVITY_EX) pCharIoctl->data_buffer;
		if (ulBlockCount)
			*ulBlockCount = send_activity_ex->ulBlockCount;
		return_code = 1;
		printf("Send task Success\n");
	} else {
		if (ulBlockCount)
			*ulBlockCount = 0;
		return_code = 0;
		printf("Send task failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
	return (return_code);
}

void 
do_rebuild(int file_handle, MP_DRIVE_ID drive_source, MP_DRIVE_ID drive_target)
{
	ULONG ulBlockCountMax;
	ULONG ulBlockCount;
	ULONG ulPrevBlockCount = 0;

	//when you issue start rebuild, ioctl will return the max number of blocks
	    // as we issue a next block, ioctl will return the number of blocks that have been completed

	    set_options(file_handle);

	if (send_task(file_handle, START_REBUILDING, 0, drive_source, drive_target, &ulBlockCountMax) &&
	    ulBlockCountMax) {
		do {
			if (!send_task(file_handle, DO_NEXT_BLOCK, 0, drive_source, drive_target, &ulBlockCount)) {
				printf("failed DO_NEXT_BLOCK\n");
				break;
			}
			printf("max blocks=%ld, completed=%ld, count=%ld\n", ulBlockCountMax, ulBlockCount, ulBlockCount - ulPrevBlockCount);
			ulPrevBlockCount = ulBlockCount;
			sleep(1);

		} while (ulBlockCount != ulBlockCountMax);
	} else {
		printf("failed START_REBUILD \n");
	}

	printf("Rebuilding completed \n");

}

void 
do_check_consistency(int file_handle, u_char sel_log_drive, MP_DRIVE_ID drive_source, MP_DRIVE_ID drive_target)
{
	ULONG ulBlockCountMax;
	ULONG ulBlockCount;
	ULONG ulPrevBlockCount = 0;

	//when you issue start consisteny, ioctl will return the max number of blocks
	    // as we issue a next block, ioctl will return the number of blocks that have been completed

	    if (send_task(file_handle, CONSISTENCY_CHECK, sel_log_drive, drive_source, drive_target, &ulBlockCountMax) &&
	    ulBlockCountMax) {
		do {

			if (!send_task(file_handle, DO_NEXT_BLOCK, sel_log_drive, drive_source, drive_target, &ulBlockCount)) {
				printf("failed DO_NEXT_BLOCK\n");
				break;
			}
			printf("max blocks=%ld, completed=%ld, count=%ld\n", ulBlockCountMax, ulBlockCount, ulBlockCount - ulPrevBlockCount);
			ulPrevBlockCount = ulBlockCount;
			sleep(1);

		} while (ulBlockCount != ulBlockCountMax);
	} else {
		printf("failed CHECK_CONSISTENCY \n");
	}

	printf("Check Consistency completed \n");

}

void 
get_activity(file_handle, sel_log_drive)
{
	PMP_GET_ACTIVITY_EX get_activity_ex;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	ULONG ulBlockCountMax = 0;
	ULONG ulBlockCount = 0;
	ULONG ulPrevBlockCount = 0;

	while (1) {

		initialize_ioctl(&megaide_ioctl, sizeof(MP_GET_ACTIVITY_EX), sizeof(MP_GET_ACTIVITY_EX), AR_GET_ACTIVITY_EX);

		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		get_activity_ex = (PMP_GET_ACTIVITY_EX) pCharIoctl->data_buffer;
		get_activity_ex->ulLogicalDriveId = sel_log_drive;

		if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
			get_activity_ex = (PMP_GET_ACTIVITY_EX) pCharIoctl->data_buffer;
		} else {
			printf("Get Activity failed\n");
			FREE_MEM(megaide_ioctl.wdata);
			return;
		}

		if (get_activity_ex->Action != MP_OP_NOP) {
			ulBlockCountMax = get_activity_ex->ulTotBlockCount;
			ulBlockCount = get_activity_ex->ulDoneBlockCount;

			printf("opcode=0x%x max blocks=%ld, completed=%ld, count=%ld\n",
			    get_activity_ex->Action,
			    ulBlockCountMax,
			    ulBlockCount,
			    ulBlockCount - ulPrevBlockCount);

			ulPrevBlockCount = ulBlockCount;
			sleep(1);

		} else {
			printf("No background action on this logical drive\n");
			FREE_MEM(megaide_ioctl.wdata);
			return;
		}
	}
	FREE_MEM(megaide_ioctl.wdata);
}




void 
dump_error_log(int file_handle, MP_DRIVE_ID DriveId)
{
	PMP_ERROR_LOG error_log;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_ERROR_LOG), sizeof(MP_ERROR_LOG), AR_GET_ERROR_LOG);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	error_log = (PMP_ERROR_LOG) pCharIoctl->data_buffer;
	error_log->DriveID = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		error_log = (PMP_ERROR_LOG) pCharIoctl->data_buffer;
		printf("Max Errors=%d, Number of Errors=%d, Head=%d, Tail=%d\n",
		    error_log->errorlog_hdr.MaxErrorCount,
		    error_log->errorlog_hdr.TotalErrorCount,
		    error_log->errorlog_hdr.Head,
		    error_log->errorlog_hdr.Tail);

	} else {
		printf("Dump Error Log call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}

void 
init_error_log(int file_handle, MP_DRIVE_ID DriveId)
{
	PMP_FAIL_DRIVE error_log;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_FAIL_DRIVE), 0, AR_INIT_ERRORLOG);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	error_log = (PMP_FAIL_DRIVE) pCharIoctl->data_buffer;
	error_log->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Init Error Log is Success\n");
	} else {
		printf("Init Error Log call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}


void 
create_new_entry_error_log(int file_handle, MP_DRIVE_ID DriveId)
{
	PMP_FAIL_DRIVE error_log;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_FAIL_DRIVE), 0, AR_LOG_ERROR);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	error_log = (PMP_FAIL_DRIVE) pCharIoctl->data_buffer;
	error_log->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Create new Log error is Success\n");
	} else {
		printf("Create new Log error call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}



void 
issue_reset(int file_handle)
{
	char_megaide_ioctl megaide_ioctl;

	initialize_ioctl(&megaide_ioctl, 0, 0, AR_RESET_CONTROLLER);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Issue Reset ");
		//display_return_code(megaide_ioctl->wdata.sic.return_code);
	} else {
		printf("Issue reset call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}

void 
issue_identify(int file_handle, MP_DRIVE_ID drive_source)
{

	char_megaide_ioctl megaide_ioctl;
	PPASS_THRU_DATA passthru_data;
	IDENTIFY_DATA *ident;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(PASS_THRU_DATA), sizeof(PASS_THRU_DATA), AR_EXECUTE_PASS_THRU_COMMAND);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	passthru_data = (PPASS_THRU_DATA) pCharIoctl->data_buffer;
	passthru_data->uchTargetID = (UCHAR) drive_source;
	passthru_data->ulSize = 512;
	passthru_data->bIsPIO = 1;
	passthru_data->bIsIn = 1;
	passthru_data->aucTaskFile[7] = IDE_COMMAND_IDENTIFY;
	passthru_data->aucTaskFile[6] = ((drive_source & 1) << 4) | IDE_CHS_MODE;
	passthru_data->ucTaskFileBitMap = 0xC0;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {

		passthru_data = (PPASS_THRU_DATA) pCharIoctl->data_buffer;

		ident = (IDENTIFY_DATA *) & passthru_data->aucBuffer;
		printf("Model=%s\n", ident->ModelNumber);
		printf("Serial=%s\n", ident->SerialNumber);
		printf("Firmware=%s\n", ident->FirmwareRevision);
		printf("Identify ");
		display_return_code(pCharIoctl->sic.return_code);

	} else {
		printf("Identify call failed\n");
	}
	FREE_MEM(megaide_ioctl.wdata);
}


void 
smart_ioctl(int file_handle, MP_DRIVE_ID drive_source,
    UCHAR smart_action)
{

	char_megaide_ioctl megaide_ioctl;
	PMP_SMART_ACTIVITY smart_activity;
	pchar_ioctl_data pCharIoctl;
	int ret = -1;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_SMART_ACTIVITY), sizeof(MP_SMART_ACTIVITY), AR_EXECUTE_SMART_COMMAND);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	smart_activity = (PMP_SMART_ACTIVITY) pCharIoctl->data_buffer;
	smart_activity->TargetID = (UCHAR) drive_source;
	smart_activity->Action = smart_action;

	ret = ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl);
	if (!ret) {

		smart_activity = (PMP_SMART_ACTIVITY) pCharIoctl->data_buffer;
		printf("SMART ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf("SMART call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}

void 
ircd_ioctl(int file_handle, MP_DRIVE_ID drive_source)
{

	char_megaide_ioctl megaide_ioctl;
	PMP_IRCD_ACTIVITY ircd_activity;
	pchar_ioctl_data pCharIoctl;
	char read_buffer[10], fpath[15];
	int ret = -1, selection, fd = -1;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_IRCD_ACTIVITY), sizeof(MP_IRCD_ACTIVITY), AR_EXECUTE_IRCD_COMMAND);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	ircd_activity = (PMP_IRCD_ACTIVITY) pCharIoctl->data_buffer;
	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n(0) - READ IRCD \n");
		printf(" (1) - WRITE IRCD \n");
		write(1, " Selection(0-1)  : ", 21);
		read(0, read_buffer, 21);
		selection = atoi(read_buffer);
	} while (selection > 1);

	memset(read_buffer, 0, sizeof(read_buffer));
	memset(fpath, 0, sizeof(fpath));
	printf("\nenter file name to read/write ircd: ");
	scanf("%s", read_buffer);
	sprintf(fpath, "/tmp/%s", read_buffer);
	fd = open(fpath, O_RDWR | O_CREAT);
	if (fd < 1) {
		printf("Error in opening file\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}
	if (selection == 1) {
		ret = read(fd, ircd_activity->ucData, 512);
		if (ret < 512) {
			printf("Error in reading file\n");
			close(fd);
			FREE_MEM(megaide_ioctl.wdata);
			return;
		}
	}
	ircd_activity->TargetID = (UCHAR) drive_source;
	ircd_activity->Action = selection;

	ret = ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl);
	if (!ret) {

		ircd_activity = (PMP_IRCD_ACTIVITY) pCharIoctl->data_buffer;
		printf("IRCD ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf("IRCD call failed\n");
	}
	if (selection == 0) {
		ret = write(fd, ircd_activity->ucData, 512);
		if (ret < 512) {
			printf("Error in writing to file\n");
		}
	}
	print_ircd_info((PNEW_IRCD) ircd_activity->ucData);

	close(fd);
	FREE_MEM(megaide_ioctl.wdata);
}

void 
execute_pass_thru_command(int file_handle)
{

	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	char read_buffer[10], fpath[15];
	int ret = -1, command_type = 0xFF, fd = -1, sel_target;
	int sector_count = 0;
	int start_sector = 0;
	int pattern = 0xFF;
	PPASS_THRU_DATA_EX passthru_data;

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n (0) - IDE_COMMAND_WRITE \n");
		printf(" (1) - IDE_COMMAND_READ\n");
		write(1, " Selection(0-1)  : ", 21);
		read(0, read_buffer, 21);
		command_type = atoi(read_buffer);
	} while (command_type > 1);

	sel_target = select_target();

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		if (command_type) {
			printf("Enter Number of Sectors to Read\n");
		} else {
			printf("Enter Number of Sectors to Write\n");
		}
		write(1, " Selection(1-254)  : ", 21);
		read(0, read_buffer, 21);
		sector_count = atoi(read_buffer);
	} while (sector_count == 0 || sector_count > 254);

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		write(1, "Enter Start Sector  :", 21);
		read(0, read_buffer, 21);
		start_sector = atoi(read_buffer);
	} while (start_sector == 0);

	if (command_type) {
		memset(read_buffer, 0, sizeof(read_buffer));
		memset(fpath, 0, sizeof(fpath));
		printf("\nenter file name to read data");
		scanf("%s", read_buffer);
		sprintf(fpath, "/tmp/%s", read_buffer);
		fd = open(fpath, O_RDWR | O_CREAT);
		if (fd < 1) {
			printf("Error in opening file\n");
			return;
		}
	} else {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\nEnter pattern to write 0-255\n");
		write(1, "Selection(1-256)  : ", 21);
		read(0, read_buffer, 21);
		pattern = atoi(read_buffer);
	}

	if (command_type) {
		//read
		    initialize_ioctl(&megaide_ioctl,
		    sizeof(PASS_THRU_DATA_EX),
		    (sizeof(PASS_THRU_DATA_EX) + (sector_count * 512)),
		    AR_EXECUTE_PASS_THRU_COMMAND);
	} else {
		initialize_ioctl(&megaide_ioctl,
		    (sizeof(PASS_THRU_DATA_EX) + (sector_count * 512)),
		    sizeof(PASS_THRU_DATA_EX),
		    AR_EXECUTE_PASS_THRU_COMMAND);
	}

	printf("command_type %d\n", command_type);
	printf("sel_target %d\n", sel_target);
	printf("sector_count %d\n", sector_count);
	printf("start_sector %d\n", start_sector);
	printf("pattern %d\n", pattern);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	passthru_data = (PPASS_THRU_DATA_EX) pCharIoctl->data_buffer;

	if (!command_type)
		memset(passthru_data->aucBuffer, pattern, (sector_count * 512));

	passthru_data->uchTargetID = (UCHAR) sel_target;
	passthru_data->ulSize = (sector_count * 512);
	passthru_data->bIsPIO = 1;
	passthru_data->aucVersionNumber = 2;
	passthru_data->aucTaskFile[2] = sector_count;
	passthru_data->aucTaskFile[3] = (start_sector & 0x000000ff);
	passthru_data->aucTaskFile[4] = (start_sector & 0x0000ff00) >> 8;
	passthru_data->aucTaskFile[5] = (start_sector & 0x00ff0000) >> 16;
	passthru_data->aucTaskFile[6] = ((sel_target & 1) << 4) | IDE_LBA_MODE | ((start_sector & 0x0f000000) >> 24);
	if (command_type) {
		passthru_data->bIsIn = 1;
		//Reading From device
		    passthru_data->aucTaskFile[7] = IDE_COMMAND_READ;
	} else {
		passthru_data->bIsIn = 0;
		//writing to device
		    passthru_data->aucTaskFile[7] = IDE_COMMAND_WRITE;
	}

	passthru_data->ucTaskFileBitMap = 0xFC;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {

		passthru_data = (PPASS_THRU_DATA_EX) pCharIoctl->data_buffer;
		display_return_code(pCharIoctl->sic.return_code);

		if (command_type) {
			ret = write(fd, passthru_data->aucBuffer, (sector_count * 512));
			if (ret < (sector_count * 512)) {
				printf("Error in writing to file\n");
			}
		}
	} else {
		printf("Execute Passthru call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
	if (command_type) {
		close(fd);
	}
}


void 
diag_command(int file_handle)
{
	PDIAG_PASS_THRU_DATA pDiagPassThru;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	char read_buffer[10], fpath[15];
	int ret = -1, sel_target, fd = -1;
	char sub_opcode;
	int pattern = 0xFF;
	ULONG sectorOffset = 0, dataTransferLen = 0;


	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n Enter Sub Op Code\n");
		printf(" (0) - DIAG_WRITE_DMA\n");
		printf(" (1) - DIAG_WRITE_MULTIPLE_PIO\n");
		printf(" (2) - DIAG_WRITE_SECTOR_PIO\n");
		printf(" (3) - DIAG_READ_DMA\n");
		printf(" (4) - DIAG_READ_MULTIPLE_PIO\n");
		printf(" (5) - DIAG_READ_VERIFY_SECTOR_PIO\n");
		printf(" (6) - DIAG_READ_SECTOR_PIO\n");
		write(1, " Selection(0-6)  : ", 21);
		read(0, read_buffer, 21);
		sub_opcode = atoi(read_buffer);
	} while (sub_opcode > 6);

	sel_target = select_target();

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n Enter Sector Offset\n");
		write(1, " Selection(0-256)  : ", 21);
		read(0, read_buffer, 21);
		sectorOffset = atoi(read_buffer);
	} while (sectorOffset > DIAG_TOTAL_SECTORS);

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n Enter Transfer Lenth In Bytes\n");
		write(1, " Selection(0-(256*512))  : ", 27);
		read(0, read_buffer, 21);
		dataTransferLen = atoi(read_buffer);
	} while ((dataTransferLen + (sectorOffset * 512)) > (DIAG_TOTAL_SECTORS * 512));

	if (sub_opcode > 2) {
		//Read Commands
		    memset(read_buffer, 0, sizeof(read_buffer));
		memset(fpath, 0, sizeof(fpath));
		printf("\nenter file name to read data");
		scanf("%s", read_buffer);
		sprintf(fpath, "/tmp/%s", read_buffer);
		fd = open(fpath, O_RDWR | O_CREAT);
		if (fd < 1) {
			printf("Error in opening file\n");
			return;
		}
	} else {
		//Write
		    memset(read_buffer, 0, sizeof(read_buffer));
		printf("\nEnter pattern to write 0-255\n");
		write(1, "Selection(1-256)  : ", 21);
		read(0, read_buffer, 21);
		pattern = atoi(read_buffer);
	}

	if (sub_opcode > 2) {
		//Read Commands
		    initialize_ioctl(&megaide_ioctl,
		    sizeof(DIAG_PASS_THRU_DATA), //In Len
		    (sizeof(DIAG_PASS_THRU_DATA) + dataTransferLen), //Out Len
		    AR_DIAG_PASSTHRU);
	} else {
		//write Commands
		    initialize_ioctl(&megaide_ioctl,
		    (sizeof(DIAG_PASS_THRU_DATA) + dataTransferLen), //In len
		    sizeof(DIAG_PASS_THRU_DATA), //Out len
		    AR_DIAG_PASSTHRU);
	}

	printf("sub_opcode %d\n", sub_opcode);
	printf("sel_target %d\n", sel_target);
	printf("sectorOffset %ld\n", sectorOffset);
	printf("dataTransferLen %ld\n", dataTransferLen);
	printf("pattern %d\n", pattern);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	pDiagPassThru = (PDIAG_PASS_THRU_DATA) pCharIoctl->data_buffer;

	if (sub_opcode < 3) {
		//write commands only
		    memset(pDiagPassThru->passthru_data.aucBuffer, pattern, dataTransferLen);
	}
	pDiagPassThru->DriveID = sel_target;
	pDiagPassThru->diag_sub_code = sub_opcode;
	pDiagPassThru->sector = sectorOffset;
	pDiagPassThru->data_transfer_length = dataTransferLen;
#if 1
	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {

		pDiagPassThru = (PDIAG_PASS_THRU_DATA) pCharIoctl->data_buffer;
		display_return_code(pCharIoctl->sic.return_code);

		if (sub_opcode > 2) {
			ret = write(fd, pDiagPassThru->passthru_data.aucBuffer, dataTransferLen);
			if (ret < dataTransferLen) {
				printf("Error in writing to file\n");
			}
		}
	} else {
		printf("Execute Passthru call failed\n");
	}
#endif
	FREE_MEM(megaide_ioctl.wdata);
	if (sub_opcode > 2) {
		close(fd);
	}
}

void 
error_recovery(int file_handle)
{

	char_megaide_ioctl megaide_ioctl;

	initialize_ioctl(&megaide_ioctl, 0, 0, AR_ERROR_RECOVERY);

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Issue error recovery Success\n");
	} else {
		printf("Issue error recovery call failed\n");
	}
	FREE_MEM(megaide_ioctl.wdata);
}

void 
change_drive_status(int file_handle, MP_DRIVE_ID DriveId, MP_PHYSICAL_DRIVE_STATUS status)
{
	PMP_SET_PDRV_STATUS phy_drv_state_info;
	pchar_ioctl_data pCharIoctl;
	char_megaide_ioctl megaide_ioctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_SET_PDRV_STATUS), 0, AR_SET_DRIVE_STATUS);
	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	phy_drv_state_info = (PMP_SET_PDRV_STATUS) pCharIoctl->data_buffer;

	phy_drv_state_info->DriveId = DriveId;
	phy_drv_state_info->DriveStatus = status;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Set Drive Status");
	} else {
		printf("Set Drive Status call failed\n");
	}
	display_return_code(pCharIoctl->sic.return_code);

	FREE_MEM(megaide_ioctl.wdata);

}
void 
set_drive_status(int file_handle)
{
	int operation;
	int selection;
	operation = select_option();
	selection = select_source();

	switch (operation) {
	case 0:
		change_drive_status(file_handle, selection, PDS_Online);
		break;
	case 1:
		change_drive_status(file_handle, selection, PDS_OffLine);
		break;
	case 2:
		change_drive_status(file_handle, selection, PDS_HotSpare);
		break;
	case 3:
		change_drive_status(file_handle, selection, PDS_Failed);
		break;
	default:
		break;
	}
}

//fail_drive
void 
fail_drive(int file_handle, MP_DRIVE_ID DriveId)
{
	PMP_FAIL_DRIVE fail_drive_info;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_FAIL_DRIVE), 0, AR_FAIL_DRIVE);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	fail_drive_info = (PMP_FAIL_DRIVE) pCharIoctl->data_buffer;
	fail_drive_info->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Fail drive");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf("Fail Drive call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}				/* fail_drive() */

void 
make_hot_spare(int file_handle, MP_DRIVE_ID DriveId)
{
	PMP_FAIL_DRIVE error_log;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_FAIL_DRIVE), 0, AR_MAKE_HOTSPARE);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	error_log = (PMP_FAIL_DRIVE) pCharIoctl->data_buffer;
	error_log->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		printf("Make hotspare");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf("Make hot spare call failed\n");
	}

	FREE_MEM(megaide_ioctl.wdata);
}

//send_megaide_ioctl
void 
send_megaide_ioctl(int file_handle)
{
	driver_probe(file_handle);
	get_driver_version(file_handle);
	get_controller_info(file_handle);
	get_raid_info(file_handle);
	get_sparepool_info(file_handle);
	get_raid_status(file_handle);
}				/* send_megaide_ioctl */


/***********************************************************
 * New IRCD Related IOCTLS
 ***********************************************************/
#ifdef SUPPORT_IRCD3
void 
DisplayConfig(int file_handle)
{
	PREAD_WRITE_IRCD pReadWriteIRCD;
	PNEW_IRCD_HEADER pRaidHeader;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(READ_WRITE_IRCD), sizeof(READ_WRITE_IRCD), AR_READ_WRITE_IRCD_INFO);

	//printf("Correct this IOCTL and use it!!\n");
	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
	pReadWriteIRCD->usGet = 0;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
		printf("Read IRCD ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf(" Read IRCD failed\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}

	pRaidHeader = (PNEW_IRCD_HEADER) & pReadWriteIRCD->IRCD.Header;

	printf("Signature %s, MajorVer %1d, MinorVer %1d, No:LD %1d \n",
	    pRaidHeader->acSignature,
	    (USHORT) pRaidHeader->ucMajorVersion,
	    (USHORT) pRaidHeader->ucMinorVersion, (USHORT) pRaidHeader->ucNumOfLDs);
	FREE_MEM(megaide_ioctl.wdata);

}

void 
WriteIRCD(int file_handle)
{
	PREAD_WRITE_IRCD pReadWriteIRCD;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(READ_WRITE_IRCD), sizeof(READ_WRITE_IRCD), AR_READ_WRITE_IRCD_INFO);
	printf("Correct this IOCTL and use it!!\n");
	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
	pReadWriteIRCD->usGet = 1;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		//pIRCDInfo = (PNEW_IRCD) megaide_ioctl->wdata.data_buffer;
		printf("Write IRCD ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf(" Write IRCD failed\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}
	FREE_MEM(megaide_ioctl.wdata);

}

void 
ReadEnquiry(int file_handle)
{
	PENQUIRY_INFO pEnquiryInfo;
	PADP_INFO pAdpInfo;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(ENQUIRY_INFO), sizeof(ENQUIRY_INFO), AR_ENQUIRY_INFO);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	pEnquiryInfo = (PENQUIRY_INFO) pCharIoctl->data_buffer;
	pEnquiryInfo->usGet = 0;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pEnquiryInfo = (PENQUIRY_INFO) pCharIoctl->data_buffer;
		printf("Read Enquiry ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf(" Read Enquiry failed\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}

	pAdpInfo = (PADP_INFO) & pEnquiryInfo->enquiry_buffer.AdpInfo;

	printf("ucRbldRate %1d, ucDiskCoercion %1d, ucBiosState %1d, ucTwoMBReserve %1d \n",
	    (USHORT) pAdpInfo->ucRbldRate,
	    (USHORT) pAdpInfo->ucDiskCoercion,
	    (USHORT) pAdpInfo->ucBiosState, (USHORT) pAdpInfo->ucTwoMBReserve);
	FREE_MEM(megaide_ioctl.wdata);
}


void 
WriteEnquiry(int file_handle)
{
	ENQUIRY_INFO EnquiryInfo;
	PENQUIRY pEnquiry;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(ENQUIRY_INFO), sizeof(ENQUIRY_INFO), AR_ENQUIRY_INFO);
	printf("Correct this IOCTL and use!!!!!!!!!!!!!!!!!!!!!\n");
	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	EnquiryInfo.usGet = 1;
	memcpy((PUCHAR) & EnquiryInfo.enquiry_buffer, (PUCHAR) pEnquiry, IRCD_SECTOR_SIZE);
	memcpy(pCharIoctl->data_buffer, (PUCHAR) pEnquiry, sizeof(ENQUIRY_INFO));

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		//pIRCDInfo = (PNEW_IRCD) megaide_ioctl->wdata.data_buffer;
		printf("Write Enquiry ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf(" Write Enquiry failed\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}

	FREE_MEM(megaide_ioctl.wdata);

}

void 
add_delete_log_drive(int file_handle)
{

	/* read IRCD */
	PREAD_WRITE_IRCD pReadWriteIRCD;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;
	char read_buffer[10];
	int selection;
	PNEW_IRCD pNewIrcd;

	initialize_ioctl(&megaide_ioctl, sizeof(READ_WRITE_IRCD), sizeof(READ_WRITE_IRCD), AR_READ_WRITE_IRCD_INFO);
	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
	pReadWriteIRCD->usGet = 0;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
		printf("Read IRCD ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf(" Read IRCD failed\n");
		FREE_MEM(megaide_ioctl.wdata);
		return;
	}

	pNewIrcd = &pReadWriteIRCD->IRCD;

	print_ircd_info(pNewIrcd);

	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n(0) - Add Logical Drive\n");
		printf(" (1) - Delete Logical Drive\n");
		write(1, " Selection(0-1)  : ", 21);
		read(0, read_buffer, 21);
		selection = atoi(read_buffer);
	} while (selection > 1);

	/* delete logical drive */
	if (selection) {
		pNewIrcd->LDInfo[pNewIrcd->Header.ucNumOfLDs].ucLDState = LDS_Deleted;
		pNewIrcd->Header.ucNumOfLDs--;
		pReadWriteIRCD->usGet = 1;
		if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
			pReadWriteIRCD = (PREAD_WRITE_IRCD) pCharIoctl->data_buffer;
			printf("Write IRCD ");
			display_return_code(pCharIoctl->sic.return_code);
		} else {
			printf("WriteIRCD failed\n");
			FREE_MEM(megaide_ioctl.wdata);
			return;
		}
		print_ircd_info(pNewIrcd);
	}
	FREE_MEM(megaide_ioctl.wdata);

}

void 
print_ircd_info(PNEW_IRCD IRCDData)
{
	UCHAR ucNoLogDrv, i;

	ucNoLogDrv = IRCDData->Header.ucNumOfLDs;

	printf("\nHeader");
	printf("\n\t|--- Signature %s", IRCDData->Header.acSignature);
	printf("\n\t|--- Version %2c:%2c", IRCDData->Header.ucMajorVersion, IRCDData->Header.ucMinorVersion);
	printf("\n\t|--- NoOfLDS %dn", IRCDData->Header.ucNumOfLDs);

	for (i = 0; i < ucNoLogDrv; i++) {
		UCHAR j;

		printf("\nLogical Drive %d", i + 1);
		printf("\n\t|--- Raid Level %d", IRCDData->LDInfo[i].ucRaidLevel);
		printf("\n\t|--- Status: %d [Onln-0/Undef-1/Deg-2/Off-3/Deleted=4]", IRCDData->LDInfo[i].ucLDState);
		printf("\n\t|--- No Of Stripes %d", IRCDData->LDInfo[i].ucNumStripes);
		printf("\n\t|--- Stripe Size %d", IRCDData->LDInfo[i].uwStripeSize);
		printf("\n\t|--- Span depth %d", IRCDData->LDInfo[i].ucSpanDepth);

		for (j = 0; j < IRCDData->LDInfo[i].ucSpanDepth; j++) {
			printf("\n\t\t|--- Start sec %ld", IRCDData->LDInfo[i].Span[j].ulStartLBA);
			printf("\n\t\t|--- No Of Secs %ld", IRCDData->LDInfo[i].Span[j].ulNumberofSectors);
			printf("\n\t\t\t|--- Drive %x", IRCDData->LDInfo[i].Span[j].aucDriveList[0]);
			printf("\n\t\t\t|--- Drive %x", IRCDData->LDInfo[i].Span[j].aucDriveList[1]);
			printf("\n\t\t\t|--- Drive %x", IRCDData->LDInfo[i].Span[j].aucDriveList[2]);
			printf("\n\t\t\t|--- Drive %x", IRCDData->LDInfo[i].Span[j].aucDriveList[3]);
		}
	}

	for (i = 0; i < 4; i++) {
		printf("\nPhysical Drive %d", i + 1);
		printf("\n\t|--- Size=%ld, State [%2x]\n", IRCDData->PDInfo[i].ulSize, IRCDData->PDInfo[i].ucState);
	}
	return;

}


#endif				/* SUPPORT_IRCD3 */

/***********************************************************
 * BackPlane Related IOCTLS
 ***********************************************************/
#ifdef BACKPLANE_SUPPORT

void 
get_backplane_info(int file_handle)
{
	PBP_GET_BPD_INFO pbpd_data;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0,
	    sizeof(BP_GET_BPD_INFO), AR_GET_BPD_INFO);
	printf("get backplane info %d\n", AR_GET_BPD_INFO);
	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		pbpd_data = (PBP_GET_BPD_INFO) pCharIoctl->data_buffer;
		printf("MajorVersion = %1d, MinorVersion = %1d  , Capability = %ld \n", pbpd_data->MajorVersion, pbpd_data->MinorVersion,
		    pbpd_data->Capability);
		printf("AR_GET_BPD_INFO call Success\n");
	} else {
		printf("AR_GET_BPD_INFO call failed\n");
	}
	//display_return_code(megaide_ioctl->wdata.sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}				/* get_backplane_info */

void 
get_bp_controls_status(int file_handle)
{
	PBP_STATUS_INFO pbp_controlstatus_data;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, 0,
	    sizeof(BP_STATUS_INFO), AR_BP_STATUS_INFO);
	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
		pbp_controlstatus_data = (PBP_STATUS_INFO) pCharIoctl->data_buffer;
		printf("BeepStatus = %d,  LedStatus-0 = %d, LedStatus-1 = %d \n", (int)pbp_controlstatus_data->BeepStatus,
		    (int)pbp_controlstatus_data->LedStatus[0], (int)pbp_controlstatus_data->LedStatus[1]);
		printf("AR_BP_STATUS_INFO call Success\n");
	} else {
		printf("AR_BP_STATUS_INFO call failed\n");
	}
	//display_return_code(megaide_ioctl->wdata.sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}				/* get_bp_controls_status */

int 
select_led_action(void)
{
	char read_buffer[20];
	int return_selection = 0;
	do {
		printf(" \n\n");
		printf(" Select your Led Action\n");
		printf(" (0) - led 0 BLINK \n");
		printf(" (1) - led 0 ON \n");
		printf(" (2) - led 0 OFF \n");
		printf(" (3) - LED 1 BLINK \n");
		printf(" (4) - LED 1 ON \n");
		printf(" (5) - LED 1 OFF \n");
		printf(" (6) - Set Identify Device LED\n");
		printf(" (7) - Reset Device LED \n");
		write(1, " Selection ( 0-6 ) : ", 21);
		read(0, read_buffer, 21);
		return_selection = atoi(read_buffer);
	} while (return_selection > 6);
/*#define AR_BLINK_LED_0			109
#define AR_ON_LED_0			110
#define AR_OFF_LED_0			111
#define AR_BLINK_LED_1			112
#define AR_ON_LED_1			113
#define AR_OFF_LED_1			114
*/
	return ((return_selection + AR_BLINK_LED_0));
}

void 
set_identify_device_led(int file_handle, int DriveId)
{
	PMP_DETECT_DRIVE drive;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_DETECT_DRIVE),
	    0, AR_IDENTIFY_DEVICE_LED);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;

	drive->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;
		printf("set_identify_device_led ");
	} else {
		printf("set_identify_device_led failed\n");
	}
	display_return_code(pCharIoctl->sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}

void 
set_device_led(int file_handle, int DriveId)
{
	PMP_DETECT_DRIVE drive;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_DETECT_DRIVE),
	    0, AR_RESET_DEVICE_LED);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;

	drive->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;
		printf("set_identify_device_led ");
	} else {
		printf("set_identify_device_led failed\n");
	}
	display_return_code(pCharIoctl->sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}

void 
bp_set_led_action(int file_handle, int action)
{
	char_megaide_ioctl megaide_ioctl;
	int DriveId;
        
	if ((action == AR_IDENTIFY_DEVICE_LED) ||
	    (action == AR_RESET_DEVICE_LED)) {
		DriveId = select_source();
		if (action == AR_IDENTIFY_DEVICE_LED) {
			set_identify_device_led(file_handle, DriveId);
		} else if (action == AR_RESET_DEVICE_LED) {
		}
	} else {
		initialize_ioctl(&megaide_ioctl, 0, 0, action);
		if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
			printf("Backplane call success\n");
		} else {
			printf("Backplane call failed \n");
		}
		//display_return_code(megaide_ioctl->wdata.sic.return_code);
		FREE_MEM(megaide_ioctl.wdata);
	}
}


void 
detect_drive(int file_handle, int DriveId)
{
	PMP_DETECT_DRIVE drive;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_DETECT_DRIVE),
	    0, AR_DETECT_DRIVE);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;

	drive->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;
		printf("Detect drive drive ");
	} else {
		printf("Detect drive failed\n");
	}
	display_return_code(pCharIoctl->sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}

void 
remove_drive(int file_handle, int DriveId)
{
	PMP_DETECT_DRIVE drive;
	char_megaide_ioctl megaide_ioctl;
	pchar_ioctl_data pCharIoctl;

	initialize_ioctl(&megaide_ioctl, sizeof(MP_DETECT_DRIVE),
	    0, AR_REMOVE_DRIVE);

	pCharIoctl = (pchar_ioctl_data) megaide_ioctl.wdata;
	drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;

	drive->DriveId = DriveId;

	if (!ioctl(file_handle, MEGAIDE_DEV_IOCTL, &megaide_ioctl)) {
		drive = (PMP_DETECT_DRIVE) pCharIoctl->data_buffer;
		printf("Remove drive ");
		display_return_code(pCharIoctl->sic.return_code);
	} else {
		printf("Remove drive failed\n");
	}
	display_return_code(pCharIoctl->sic.return_code);
	FREE_MEM(megaide_ioctl.wdata);
}

void 
detect_or_remove_drive(int file_handle)
{
	char read_buffer[10];
	int selection;
	int driveId;
	do {
		memset(read_buffer, 0, sizeof(read_buffer));
		printf("\n(0) - Detect Drive\n");
		printf(" (1) - Remove Drive\n");
		write(1, " Selection(0-1)  : ", 21);
		read(0, read_buffer, 21);
		selection = atoi(read_buffer);
	} while (selection > 1);

	driveId = select_source();

	if (selection) {
		remove_drive(file_handle, driveId);
	} else {
		detect_drive(file_handle, driveId);
	}
}

#endif


void 
display_return_code(u_long return_code)
{
	switch (return_code) {
		case MP_IOCTL_SUCCESS:
		printf("Command return Code: MP_IOCTL_SUCCESS\n");
		break;

	case MP_IOCTL_FAIL:
		printf("Command return Code: MP_IOCTL_FAIL\n");
		break;

	case MP_IOCTL_TIMEOUT:
		printf("Command return Code: MP_IOCTL_TIMEOUT\n");
		break;

	case MP_IOCTL_PENDING:
		printf("Command return Code: MP_IOCTL_PENDING\n");
		break;

	case MP_IOCTL_DATALEN_MISMATCH:
		printf("Command return Code: MP_IOCTL_DATALEN_MISMATCH\n");
		break;

	case MP_IOCTL_BAD_OPCODE:
		printf("Command return Code: MP_IOCTL_BAD_OPCODE\n");
		break;

	case MP_IOCTL_PARAM_MISMATCH:
		printf("Command return Code: MP_IOCTL_PARAM_MISMATCH\n");
		break;

	case MP_IOCTL_SMART_STATUS_OK:
		printf("Command return Code: MP_IOCTL_SMART_STATUS_OK\n");
		break;

	case MP_IOCTL_SMART_STATUS_BAD:
		printf("Command return Code: MP_IOCTL_SMART_STATUS_BAD\n");
		break;

	case MP_IOCTL_DATA_LOCKED:
		printf("Command return Code: MP_IOCTL_DATA_LOCKED\n");
		break;

	case MP_IOCTL_POWER_CONNECTED:
		printf("Command return Code: MP_IOCTL_POWER_CONNECTED\n");
		break;

	case MP_IOCTL_NO_POWER_CONNECTED:
		printf("Command return Code: MP_IOCTL_NO_POWER_CONNECTED\n");
		break;

	case MP_IOCTL_NO_48_BIT_SUPPORT:
		printf("Command return Code: MP_IOCTL_NO_48_BIT_SUPPORT\n");
		break;

	case BPD_ERROR_FUNCTION_NOT_IMPLEMENTED:
		printf("Command return Code: BPD_ERROR_FUNCTION_NOT_IMPLEMENTED\n");
		break;

	case BPD_ERROR_DEVICE_ADDRESS_NOT_PRESENT:
		printf("Command return Code: BPD_ERROR_DEVICE_ADDRESS_NOT_PRESENT\n");
		break;

	case BPD_ERROR_EVENT_RECORD_MISMATCH:
		printf("Command return Code: BPD_ERROR_EVENT_RECORD_MISMATCH\n");
		break;

	case BPD_ERROR_EVENT_LOG_FULL:
		printf("Command return Code: BPD_ERROR_EVENT_LOG_FULL\n");
		break;

	case BPD_ERROR_BPD_INITIALIZING:
		printf("Command return Code: BPD_ERROR_BPD_INITIALIZING\n");
		break;

	case BPD_ERROR_DRIVER_NOT_EXIST:
		printf("Command return Code: BPD_ERROR_DRIVER_NOT_EXIST\n");
		break;

	default:
		printf("Command return Code: UNKNOWN!!\n");
		break;
	}
}

int 
dump_menu(void)
{
	char read_buffer[20];
	int return_selection;

	printf(" \n\n");
	printf(" Megaide IOCTL Options\n");
	printf(" [ 0] - Info                [ 1] - Rebuild From Source To Target \n");
	printf(" [ 2] - Check Consistency   [ 3] - Get Activity(EX)\n");
	printf(" [ 4] - Abort               [ 5] - Dump Error Log\n");
	printf(" [ 6] - Erase Error Log     [ 7] - Create Error Log Entry\n");
	printf(" [ 8] - Issue Reset         [ 9] - Identify(PassThru cmd)\n");
	printf(" [10] - Smartioctl(PasThru) [11] - IRCD ioctl(PasThru)\n");
	printf(" [12] - Pass thru Cmd       [13] - Diagnostic PassThru \n");
	printf(" [14] - Enable Er Recovery  [15] - Set Drive Status\n");
	printf(" [16] - Reset Controller\n");

#ifdef SUPPORT_IRCD3
	printf(" New IRCD Options---------------------------------------\n");
	printf(" [20] - Read IRCD           [21] - Write IRCD\n");
	printf(" [22] - Read Enquiry        [23] - Write Enquiry\n");
	printf(" [24] - Add/Delete LD       \n");
#endif
#ifdef BACKPLANE_SUPPORT
	printf(" Backplane Options--------------------------------------\n");
	printf(" [40] - Get BP Info         [41] - Get BP controls status \n");
#if 0
	printf(" [42] - Get NVRAM INFO      [43] - CLEAR NVRAM \n");
	printf(" [44] - Get NV EVENT INFO   [45] - Enable/disable Beeper \n");
	printf(" [46] - Silence beeper      [47] - Add event \n");
#endif
	printf(" [48] - Change Led status   [49] - Detect/Remove PD\n");
#endif
	printf(" [60] -  EXIT\n");
	write(1, " Selection(0-20)   :", 20);
	read(0, read_buffer, 22);
	return_selection = atoi(read_buffer);
	return (return_selection);
}




int 
main(int argc, char **argv)
{
	int fd;
	char devname[30];
	int selection;
	int sel_source, sel_target;
	u_char sel_log_drive;
	UCHAR smart_action;

	printf("ideutil - version %s\n", VERSION);
	strcpy(devname, CHAR_DEVICE_FILE);

	fd = open(devname, O_RDWR | O_FSYNC);
	if (fd < 0) {
		perror(devname);
		exit(errno);
	}
	printf("%s: device sucessfully opened\n\n", devname);

try_again:
	selection = dump_menu();
	switch (selection) {
		//get_controller_info(fd) // AR_GET_CONTROLLER_INFO
		    // get_raid_info() // AR_RAID_INFO
		    // get_sparepool_info // AR_GET_SPAREPOOL_INFO

	case 0:
		driver_probe(fd);
		//AR_PROBE_AMIRAID
		    get_driver_version(fd);
		//AR_GET_VERSION
		    get_raid_status(fd);
		//AR_RAID_STATUS_INFO_EX
		    get_physical_drv_info(fd);
		//AR_GET_PHYSICAL_DRIVE_INFO
		    break;
	case 1:
		sel_source = select_source();
		sel_target = select_target();
		do_rebuild(fd, sel_source, sel_target);
		//AR_SEND_ACTIVITY_EX
		    break;
		//START_REBUILDING

	case 2:
		sel_log_drive = select_log_drive();
		sel_source = select_source();
		sel_target = select_target();
		do_check_consistency(fd, sel_log_drive, //AR_SEND_ACTIVITY_EX
		    sel_source, sel_target);
		//CONSISTENCY_CHECK
		    break;

	case 3:
		sel_log_drive = select_log_drive();
		get_activity(fd, sel_log_drive);
		//AR_GET_ACTIVITY_EX
		    break;

	case 4:
		sel_log_drive = select_log_drive();
		sel_source = select_source();
		send_task(fd, ABORT_COMMAND, sel_log_drive, sel_source, 0, NULL);
		break;

	case 5:
		selection = select_source();
		dump_error_log(fd, selection);
		//AR_GET_ERROR_LOG
		    break;

	case 6:
		selection = select_source();
		init_error_log(fd, selection);
		//AR_INIT_ERRORLOG
		    break;

	case 7:
		selection = select_source();
		create_new_entry_error_log(fd, selection);
		//AR_LOG_ERROR
		    break;

	case 8:
		issue_reset(fd);
		break;
		//AR_RESET_CONTROLLER

	case 9:
		selection = select_source();
		issue_identify(fd, selection);
		//AR_EXECUTE_PASS_THRU_COMMAND
		    break;

	case 10:
		smart_action = select_smart_action();
		selection = select_source();
		smart_ioctl(fd, selection, smart_action);
		//AR_EXECUTE_SMART_COMMAND
		    break;

	case 11:
		selection = select_source();
		//AR_EXECUTE_IRCD_COMMAND
		    ircd_ioctl(fd, selection);
		break;

	case 12:
		execute_pass_thru_command(fd);
		break;
		//AR_EXECUTE_PASS_THRU_COMMAND

	case 13:
		diag_command(fd);
		break;
		//AR_DIAG_PASSTHRU

	case 14:
		error_recovery(fd);
		break;
		//AR_ERROR_RECOVERY

	case 15:
		set_drive_status(fd);
		break;

	case 16:
		issue_reset(fd);
		break;

#ifdef SUPPORT_IRCD3
	case 20:
		DisplayConfig(fd);
		break;

	case 21:
		WriteIRCD(fd);
		break;

	case 22:
		ReadEnquiry(fd);
		break;

	case 23:
		WriteEnquiry(fd);
		break;

	case 24:
		add_delete_log_drive(fd);
		break;
#endif				/* SUPPORT_IRCD3 */

#ifdef BACKPLANE_SUPPORT

	case 40:
		get_backplane_info(fd);
		break;
		//AR_GET_BPD_INFO

	case 41:
		get_bp_controls_status(fd);
		break;
		//AR_BP_STATUS_INFO
#if 0
	case 42:
		get_nvram_info(fd);
		break;

	case 43:
		clear_nvram_action(fd);
		break;

	case 44:
		get_bp_eventlog_data(fd);
		break;

	case 45:
		set_bp_options_action(fd);
		break;

	case 46:
		bp_beep_silence_action(fd);
		break;

	case 47:
		add_nvram_action(fd);
		break;
#endif
	case 48:{
			int return_selection = 0;
			return_selection = select_led_action();
			bp_set_led_action(fd, return_selection);
			break;
		}
	case 49:
		detect_or_remove_drive(fd);
		break;
		//AR_DETECT_DRIVE
		    // AR_REMOVE_DRIVE

#endif
	case 60:
		close(fd);
		exit(0);
		return (0);
	default:
		printf(" Unsupported Option!!! \n");
		break;
	}
	goto try_again;
}				/* main() */

/*vi set ts=8:*/
