/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/util/ideutil.h,v emoore Exp$
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
/*
 * #include <linux/hdreg.h> #include <linux/fs.h> #include <linux/major.h>
 */
#define		VERSION			"00.01a"
#define		DEVICE_FILE		"/dev/sda"
#define		MAX_REQ_SENSE_LEN	0x20
#define		MAX_CDB_LEN		0x0A

/* Maximum supported Drives and channels and controllers */
#define		MAX_CONTROLLERS		2
#define		MAX_CHANNELS_PER_CONTROLLER 2
#define		MAX_DRIVES_PER_CHANNEL	2

#define MAX_DRIVES_PER_CONTROLLER      \
	(MAX_CHANNELS_PER_CONTROLLER * MAX_DRIVES_PER_CHANNEL)


#define PHYSICAL_DRIVE_MODEL_LENGTH 	80
#define PHYSICAL_DRIVE_SERIAL_NO_LENGTH	40
#define PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH 10


#define		SEL_INFO		0
#define		SEL_REBUILD		1
#define		SEL_CHECK_CON		2
#define		SEL_GET_ACTIVITY	3
#define		SEL_ABORT		4
#define		SEL_FAIL		5
#define		SEL_HOT_SPARE		6
#define		SEL_DUMP_ERROR_LOG	7
#define		SEL_INIT_ERROR_LOG	8
#define		SEL_CREATE_ERROR_LOG	9
#define		SEL_RESET		10
#define		SEL_IDENTIFY_PASS_THRU	11
#define		SEL_SMART_IOCTL		12
#define		SEL_IRCD_IOCTL		13
#define		SEL_DELETE_LOG_DRIVE	14
#define		SEL_READ_ENQUIRY	15
#define		SEL_WRITE_ENQUIRY	16
#define		SEL_READ_IRCD		17
#define		SEL_WRITE_IRCD		18
#define		SEL_PASS_THRU_COMMAND	19
#define		SEL_DIAG_COMMAND	20
#define		SEL_EXIT		25

#ifdef BACKPLANE_SUPPORT
#define		SEL_GET_BPD_INFO	0
#define		SEL_GET_BP_CONTROLS_STATUS 1
#define		SEL_GET_NVRAM_INFO	2
#define		SEL_CLEAR_NVRAM		3
#define		SEL_GET_NVRAM_EVENTLOG_DATA 4
#define		SEL_SET_OPTIONS		5
#define		SEL_SILENCE_BEEP	6
#define		SEL_ADD_ENTRY		7
#define		SEL_LED_ACTION		8
#define		SEL_DRIVE_INSERT	9
#define		SEL_ERROR_RECOVERY	10
#define		SEL_DRIVE_REMOVE	11
#define		SEL_BP_RAID_STATUS	12
#define		SEL_BP_RESET		13
#define		SEL_EXIT_BACKPLANE	14
#endif


// IDE command definitions
#define IDE_COMMAND_ATAPI_RESET		0x08
#define IDE_COMMAND_RECALIBRATE		0x10
#define IDE_COMMAND_READ		0x20
#define IDE_COMMAND_WRITE		0x30
#define IDE_COMMAND_VERIFY		0x40
#define IDE_COMMAND_SEEK		0x70
#define IDE_COMMAND_SET_DRIVE_PARAMETERS 0x91
#define IDE_COMMAND_ATAPI_PACKET	0xA0
#define IDE_COMMAND_ATAPI_IDENTIFY	0xA1
#define IDE_COMMAND_EXECUTE_SMART	0xB0
#define IDE_COMMAND_READ_MULTIPLE	0xC4
#define IDE_COMMAND_WRITE_MULTIPLE	0xC5
#define IDE_COMMAND_SET_MULTIPLE	0xC6
#define IDE_COMMAND_READ_DMA		0xC8
#define IDE_COMMAND_WRITE_DMA		0xCA
#define IDE_COMMAND_GET_MEDIA_STATUS	0xDA
#define IDE_COMMAND_ENABLE_MEDIA_STATUS	0xEF
#define IDE_COMMAND_IDENTIFY		0xEC
#define IDE_COMMAND_MEDIA_EJECT		0xED
#define IDE_COMMAND_SET_FEATURES	0xEF
#define IDE_COMMAND_FLUSH_CACHE		0xE7
// 48 bit command Definition
#define IDE_COMMAND_FLUSH_CACHE_EXT	0xEA
#define IDE_COMMAND_READ_DMA_EXT	0x25
#define IDE_COMMAND_READ_DMA_QUEUED_EXT	0x26
#define IDE_COMMAND_READ_MULTIPLE_EXT	0x29
#define IDE_COMMAND_READ_NATIVE_MAX_ADDRESS_EXT	0x27
#define IDE_COMMAND_READ_SECTORS_EXT	0x24
#define IDE_COMMAND_READ_VERIFY_SECTORS	0x40
#define IDE_COMMAND_SET_MAX_ADDRESS_EXT	0x37
#define IDE_COMMAND_WRITE_DMA_EXT	0x35
#define IDE_COMMAND_WRITE_DMA_QUEUED_EXT 0x36
#define IDE_COMMAND_WRITE_MULTIPLE_EXT	0x39
#define IDE_COMMAND_WRITE_SECTORS_EXT	0x34


#ifndef LONG
typedef long LONG;
#endif

#ifndef ULONG
typedef unsigned long ULONG;
#endif

#ifndef PULONG
typedef unsigned long *PULONG;
#endif

#ifndef CHAR
typedef char CHAR;
#endif

#ifndef UCHAR
typedef unsigned char UCHAR;
#endif

#ifndef WORD
typedef unsigned int WORD;
typedef unsigned int UINT;
#endif

#ifndef BYTE
typedef unsigned char BYTE;
#endif

#ifndef BOOLEAN
typedef long BOOLEAN;
#endif

#ifndef PVOID
typedef void *PVOID;
#endif

#ifndef SCSI_IOCTL_SEND_COMMAND
#define SCSI_IOCTL_SEND_COMMAND		0x01
#endif

#ifndef PUCHAR
typedef unsigned char *PUCHAR;
#endif

#ifndef USHORT
typedef unsigned short USHORT;
typedef USHORT *PUSHORT;
#endif

#define IDE_LBA_MODE			0xE0
#define IDE_CHS_MODE			0xA0

/*
 * IDENTIFY data.
 */
#pragma pack (1)
typedef struct _IDENTIFY_DATA {
    USHORT	GeneralConfiguration;	/* 00 00 */
    USHORT	NumberOfCylinders;	/* 02 1a */
    USHORT	Reserved1;		/* 04 2 */
    USHORT	NumberOfHeads;		/* 06 3 */
    USHORT	UnformattedBytesPerTrack; /* 08 4 */
    USHORT	UnformattedBytesPerSector; /* 0A 5 */
    USHORT	SectorsPerTrack;	/* 0C 6 */
    USHORT	VendorUnique1[3];	/* 0E  7 - 9 */
    UCHAR 	SerialNumber[20];	/* 14 10 - 19 */
    USHORT	BufferType;		/* 28 20 */
    USHORT	BufferSectorSize;	/* 2A 21 */
    USHORT	NumberOfEccBytes;	/* 2C 22 */
    UCHAR	FirmwareRevision[8];	/* 2E  23 - 26 */
    UCHAR	ModelNumber[40];	/* 36 27 - 46 */
    UCHAR	MaximumBlockTransfer;	/* 5E  47 */
    UCHAR	VendorUnique2;		/* 5F */
    USHORT	DoubleWordIo;		/* 60 48 */
    USHORT	Capabilities;		/* 62 49 */
    USHORT	Reserved2;		/* 64 50 */
    UCHAR	VendorUnique3;		/* 66 51 */
    UCHAR	PioCycleTimingMode;	/* 67 */
    UCHAR	VendorUnique4;		/* 68 52 */
    UCHAR	DmaCycleTimingMode;	/* 69 */
    USHORT	TranslationFieldsValid;	/* 6A 53 */
    USHORT	NumberOfCurrentCylinders; /* 6C 54 */
    USHORT	NumberOfCurrentHeads;	/* 6E  55 */
    USHORT	CurrentSectorsPerTrack; /* 70 56 */
    ULONG	CurrentSectorCapacity;	/* 72 57 - 58 */
    USHORT	CurrentMultiSectorSetting; /* 59 */
    ULONG	UserAddressableSectors;	/* 60 - 61 */
    UCHAR	SingleWordDmaActive;
    UCHAR	SingleWordDmaSupport;	/* 62 */
    UCHAR	MultiWordDmaActive;
    UCHAR	MultiWordDmaSupport;	/* 63 */
    UCHAR	Reserved4;
    UCHAR	AdvancedPioModes;	/* 64 */
    USHORT	MinimumMwXferCycleTime;	/* 65 */
    USHORT	RecommendedMwXferCycleTime; /* 66 */
    USHORT	MinimumPioCycleTime;	/* 67 */
    USHORT	MinimumPioCycleTimeIordy; /* 68 */
    USHORT	Reserved5[11];		/* 69 - 79 */
    USHORT	MajorRevision;		/* 80 */
    USHORT	MinorRevision;		/* 81 */
    USHORT	CmdSupported1;		/* 82 */
    USHORT	CmdSupported2;		/* 83 */
    USHORT	FtrSupported;		/* 84 */
    USHORT	CmdEnabled1;		/* 85 */
    USHORT	CmdEnabled2;		/* 86 */
    USHORT	FtrEnabled;		/* 87 */
    UCHAR	UltraDmaSupport;
    UCHAR	UltraDmaActive;		/* 88 */
    USHORT	Reserved7[37];		/* 89 - 125 */
    USHORT	LastLun;		/* 126 */
    UCHAR	MediaStatusNotification;
    USHORT	DeviceWriteProtect;
    USHORT	Reserved11[128];	/* 128 - 255 */
}

 IDENTIFY_DATA, *PIDENTIFY_DATA;

#pragma pack ()
void DisplayConfig(int file_handle);
void WrieIRCD(int file_handle);
void ReadEnquiry(int file_handle);
void WriteEnquiry(int file_handle);
void display_return_code(u_long return_code);

#ifdef BACKPLANE_SUPPORT
void get_backplane_info(int file_handle);
void get_bp_controls_status(int file_handle);
/*
void get_nvram_info(int file_handle);
ulong add_nvram_action(int file_handle);
ulong clear_nvram_action(int file_handle);
ulong get_bp_eventlog_data(int file_handle);
ulong set_bp_options_action(int file_handle);
ulong bp_beep_silence_action(int file_handle);
void bp_detect_drive(int file_handle, int DriveId);
void bp_remove_drive(int file_handle, int DriveId);
*/
void  bp_set_led_action(int file_handle, int action);
int select_led_action(void);
void set_identify_device_led(int file_handle, int DriveId);
void set_device_led(int file_handle, int DriveId);
void detect_drive(int file_handle, int DriveId);
void remove_drive(int file_handle, int DriveId);
void detect_or_remove_drive(int file_handle);


int dump_menu(void);
#endif
/* vi set ts=8: */
