/*
 * Copyright (c) 2001,2002 Eric Moore
 * Copyright (c) 2001,2002 Vanishree Sabapathi
 * Copyright (c) 2002 LSI Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The party using or redistributing the source code and binary forms
 *    agrees to the disclaimer below and the terms and conditions set forth
 *    herein.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *	$FreeBSD: src/sys/contrib/dev/mide/util/mp_hd_ioctls.h,v emoore Exp$
 */

#ifndef __HD_IOCTLS_H__
#define __HD_IOCTLS_H__

/*
*-------------------------------------------------------------------
* Macros Definition
*-------------------------------------------------------------------
*/

#define MAX_LOGICAL_DRIVES		8


#define INVALID_CONNECTION_ID		0xff
#define INVALID_DRIVE_ID		0xff
#define INVALID_ARRAY_ID		0xff
#define INVALID_CHANNEL_ID		0xff



#define PHYSICAL_DRIVE_MODEL_LENGTH 	80
#define PHYSICAL_DRIVE_SERIAL_NO_LENGTH	40
#define PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH 10

/*#define IDE_RESERVED_SECTORS		4096  // Last 2 MB is reserved for Config, Error Log and Bad Block Management*/
#define IDE_RESERVED_SECTORS		2  /* Last 2 sectors Config and Error Log*/

/*
*-------------------------------------------------------------------
* Structure Definition
*-------------------------------------------------------------------
*/



#define MEGAIDE_IOCTL			0x80
#define MEGAIDE_IOCTL_NEW		0xE0




/* opcodes*/
#define AR_PROBE_AMIRAID		1
#define AR_RAID_INFO			2
#define AR_SET_OPTIONS			3
#define AR_GET_ERROR_LOG		4
#define AR_READ_BLK_CMD			5
#define AR_WRITE_BLK_CMD		6
#define AR_GET_EVENT			7
#define AR_SEND_ACTIVITY		8
#define AR_GET_ACTIVITY			9

/* +++ edm March 13, 2002*/
#define AR_GET_ACTIVITY_EX		10		/* new ioctl for IRCD ver 3*/
#define AR_SEND_ACTIVITY_EX		11		/* new ioctl for IRCD ver 3*/
#define AR_DELETE_LOG_DRIVE		13		/* new ioctl for IRCD ver 3*/
#define AR_RAID_STATUS_INFO_EX		14		/* new ioctl for IRCD ver 3*/
#define AR_GET_PHYSICAL_DRIVE_INFO	15		/* new ioctl for IRCD ver 3*/
#define AR_GET_VERSION_EX		16
/* --- edm March 13, 2002*/

#define AR_GET_CONTROLLER_INFO		18
#define AR_GET_SPAREPOOL_INFO		19
#define AR_GET_FAILDRIVE_INFO		20
#define AR_GET_VERSION			21
#define AR_EXECUTE_SMART_COMMAND	22
#define AR_CLEAR_ERROR_LOG		40
#define AR_RAID_STATUS_INFO		41
#define AR_FAIL_DRIVE			42
#define AR_RESET_CONTROLLER		43
#define AR_INIT_ERRORLOG		44
#define AR_LOG_ERROR			45
#define AR_MAKE_HOTSPARE		46
#define AR_EXECUTE_PASS_THRU_COMMAND	47

/*++Nagendra 12/12/01*/
#define AR_EXECUTE_IDENTIFY_COMMAND	48
#define AR_EXECUTE_FEATURES_COMMAND	49
#define AR_GET_GEOMETRY_COMMAND		50
#define AR_INTERNAL_PASS_THRU_COMMAND	51
#define AR_EXECUTE_IRCD_COMMAND		52
/*--Nagendra 12/12/01*/

#define AR_ERROR_RECOVERY		53
#define AR_DIAG_PASSTHRU		54/*manoj 053102*/
#define AR_SET_DRIVE_STATUS		55/*Vani 1017 Linux Fixes */

/* ioctls for backplane*/
#define AR_BACKPLANE_ACTIVITY		100
#define AR_DETECT_DRIVE			120
#define AR_REMOVE_DRIVE			121

/* ++ Vani: 0319*/
#define AR_READ_WRITE_IRCD_INFO		125
#define AR_ENQUIRY_INFO			126
/* -- Vani: 0319*/





#ifndef MAX_COMMAND_SIZE
#define	MAX_COMMAND_SIZE 		12
#endif

/* return_code*/
#define	MP_IOCTL_SUCCESS		0
#define	MP_IOCTL_FAIL			1
#define	MP_IOCTL_TIMEOUT		2
#define	MP_IOCTL_PENDING		4
#define	MP_IOCTL_DATALEN_MISMATCH	5
#define	MP_IOCTL_BAD_OPCODE		6
#define	MP_IOCTL_PARAM_MISMATCH		7
#define	MP_IOCTL_SMART_STATUS_OK	8 /*Nagendra 12/12/01*/
#define	MP_IOCTL_SMART_STATUS_BAD	9 /*Nagendra 12/12/01*/
#define	MP_IOCTL_DATA_LOCKED		10
#define	MP_IOCTL_POWER_CONNECTED	11
#define	MP_IOCTL_NO_POWER_CONNECTED	12
#define	MP_IOCTL_NO_48_BIT_SUPPORT	13/*manoj 052802*/


/* even more return codes*/
#define IOCTL_OK			0
#define IOCTL_ERROR			1
#define IOCTL_QUEUE			2

/* ++ Vani: Defines for IRCD, Enquiry and errorlog update*/
#define NO_HOLD				0
#define UPDATE_IRCD			1
#define UPDATE_ERRORLOG			2
#define UPDATE_ENQUIRY			3
#define UPDATE_ENQ_IRCD			4
/* -- Vani: New IRCD*/
/* ++ Vani: Defines for Enquiry Update*/
#define	BOTH_CHANNELS			0x11
#define FIRST_CHANNEL			0x10
#define SECOND_CHANNEL			0x01
/* -- Vani: Defines for Enquiry Update*/


#define MAX_DRIVES_PER_CHANNEL 		2

typedef struct _ioctl_header {
    u_long	header_len;	/* length of this header*/
    u_long	timeout;	/* timeout for this command (1 - no timeout)*/
    u_long	opcode;		/* requested opcode */
    u_long	return_code;	/* return code */
    u_long	data_len;	/* length of data area */
} ioctl_header;

/* the driver will handle copying the data_buffer[] to and from user space*/
typedef struct _ioctl_data {
    ioctl_header sic;		/* header */
    u_char 	data_buffer[( 4*1024) - sizeof(ioctl_header)];		/* ptr to data body , driver will*/
}
__attribute__ ((packed))
ioctl_data, *pioctl_data;

/* this is the structure passed to scsi mid layer */
typedef struct _megaide_ioctl_in_new{
    u_long	inlen;		/* length of data to be written to device*/
    u_long	outlen;		/* length of data to be read from device */
    u_char	cmnd[10];	/* SCSI Command ( copied to scsi_cmnd->cmnd  )*/
    ioctl_data	wdata;		/* Data written to device starts here ( coped to scsi_cmnd->request_buffer ) */
}
__attribute__ ((packed))
megaide_ioctl_in_new , *pmegaide_ioctl_in_new ;

typedef struct _megaide_ioctl_in{
    u_long	inlen;		/* length of data to be written to device */
    u_long	outlen;		/* length of data to be read from device */
    u_char	cmnd[12];	/* SCSI Command ( copied to scsi_cmnd->cmnd )*/
    ioctl_data	wdata;		/* Data written to device starts here ( coped to scsi_cmnd->request_buffer )*/
}
__attribute__ ((packed))
megaide_ioctl_in , *pmegaide_ioctl_in ;

/* this is the structure returned from scsi mid layer */
typedef struct _megaide_ioctl_out{
    u_long	inlen;		/* length of data to be written to device */
    u_long	outlen;		/* length of data to be read from device*/
    ioctl_data	wdata;		/* Data written to device starts here ( coped to scsi_cmnd->request_buffer )*/
}
__attribute__ ((packed))
megaide_ioctl_out, *pmegaide_ioctl_out;


/*++ manoj 0520 for New IOCTL Implementaion*/

/* the driver will handle copying the data_buffer[] to and from user space*/
typedef struct _char_ioctl_data {
    ioctl_header sic; 		/* IOCTL header*/
    u_char 	data_buffer[1];	/* ptr to data body , driver will*/
}
__attribute__ ((packed))
char_ioctl_data, *pchar_ioctl_data;

/*
* New IOCTL Implementaion it does not requre cmnd field
*/
typedef struct _char_megaide_ioctl {
    u_long	inlen;		/* length of data to driver = sizeof (ioctl_data)+data_length)*/
    u_long	outlen;		/* length of data from driver = sizeof (ioctl_data)+data_length)*/
    int		iAdapNumber;	/* Adapter number starts From 1*/
    char	acSignature[7];	/* Must contain "MEGANIT" without ending '\0'*/
    char	acReserved[23];
    void *wdata;		/* Data written to device starts here ( coped to scsi_cmnd->request_buffer )*/
}
__attribute__ ((packed))
char_megaide_ioctl, *pchar_megaide_ioctl;

#define MEGAIDE_IOC_MAGIC	'm'
#define MIDE_IOCTL		_IOR('m', 0x95, struct _char_megaide_ioctl )
#define MEGAIDE_DEV_IOCTL _IOW('m', 0x96 , struct _char_megaide_ioctl )

/*#define MEGAIDE_DEV_IOCTL	_IOWR('m', 0x3 , char_megaide_ioctl ) */

/*
#define MEGAIDE_DEV_IOCTL_TEST _IOC(IOC_INOUT, 'm', 0x201 , sizeof(char_megaide_ioctl) )
#define MEGAIDE_DEV_IOCTL_PASS _IOC(IOC_INOUT, 'm', 0x202 , sizeof(char_megaide_ioctl) )
*/
#define NIT_MAX_DATA_TRANSFER	(128*1024)
#define CHAR_HEADER_SIZE 	(sizeof(char_megaide_ioctl) - sizeof(ioctl_data))


/* Drive ID */
typedef enum
{
    PrimaryMaster = 0,
    PrimarySlave = 1,
    SecondaryMaster = 2,
    SecondarySlave = 3,
    InvalidDrive = INVALID_DRIVE_ID
}
MP_DRIVE_ID;


/*RAID levels.*/
typedef enum
{
    None = 0x00,
    Raid0 = 0x01,      /* stripe*/
    Raid1 = 0x02,      /* mirror*/
    Raid10 = 0x03,      /* stripe+mirror*/
    SpareDrivePool = 0x08
}
MP_RAID_LEVEL;


/*RAID Array status*/

typedef enum
{
    LDS_Online = 0,
    LDS_Critical = 1,
    LDS_Degraded = 2,
    LDS_OffLine = 3,
    LDS_Deleted = 4,
    LDS_NotPresent = 0xff
}
MP_LOGICAL_DRIVE_STATUS;



/* RAID Physical Drive status.*/

typedef enum
{
    PDS_Online = 0,
    PDS_Critical = 1,
    PDS_Failed = 2,
    PDS_HotSpare = 0xd,
    PDS_Rebuilding = 0xe,
    PDS_OffLine = 0xf,
    PDS_Undefined = 0xff
}
MP_PHYSICAL_DRIVE_STATUS;


/*
Define transfer modes.
*/
typedef enum
{
    MP_PioMode0 = 0,
    MP_PioMode1 = 1,
    MP_PioMode2 = 2,
    MP_PioMode3 = 3,
    MP_PioMode4 = 4,
    MP_DmaMode0 = 5,
    MP_DmaMode1 = 6,
    MP_DmaMode2 = 7,
    MP_UdmaMode0 = 8,
    MP_UdmaMode1 = 9,
    MP_UdmaMode2 = 10,
    MP_UdmaMode3 = 11,
    MP_UdmaMode4 = 12,
    MP_UdmaMode5 = 13,
    MP_UdmaMode6 = 14
}
MP_TRANSFER_MODE, *PMP_TRANSFER_MODE;


/*
Physical Drive Information Structure
*/
typedef struct _MP_RAID_INFO_PD
{
    MP_DRIVE_ID	DriveID;
    u_char	caModelInfo [ PHYSICAL_DRIVE_MODEL_LENGTH + 1 ];
    u_char	caSerialNumber [ PHYSICAL_DRIVE_SERIAL_NO_LENGTH + 1 ];
    u_char	caFirmwareRevision[ PHYSICAL_DRIVE_FIRMWARE_REV_LENGTH + 1];
    MP_TRANSFER_MODE TransferMode;
    MP_PHYSICAL_DRIVE_STATUS PhyStatus;
    u_long	ulPhySize;
    u_char	ucIsPhyDrvPresent;
    u_char	ucIsPowerConnected;
    u_char	ucIs80PinCable;
    u_char	ucSMART_Compliant;
    u_char	ucSMART_Enabled;
}
__attribute__ ((packed))
MP_RAID_INFO_PD, *PMP_RAID_INFO_PD;


/*
Yet Another Physical Drive Info IOCTL
*/

typedef struct _IDE_PHYSICAL_DRIVE_INFO
{
    u_char	ucNumPhysicalDrives;
    MP_RAID_INFO_PD PhysicalDrives[MAX_DRIVES_PER_CONTROLLER];
}
__attribute__ ((packed))
IDE_PHYSICAL_DRIVE_INFO, * PIDE_PHYSICAL_DRIVE_INFO;


/*
Logical Drive Information
*/
typedef struct  _MP_RAID_INFO_LD
{
    MP_RAID_LEVEL Mode;
    MP_LOGICAL_DRIVE_STATUS Status;
    u_long	ulStripeSize;       /* Stripe Size in Sector Count*/
    u_long	ulStripesPerRow;
    u_long	ulTotDriveCnt;
    u_long	ulCapacity;		/* Drive Size*/
    u_long	ulArrayId;
    MP_RAID_INFO_PD phyDrives [ MAX_DRIVES_PER_CONTROLLER ];
}
__attribute__ ((packed))
MP_RAID_INFO_LD, *PMP_RAID_INFO_LD;

/*
Raid Information
*/

typedef struct _MP_IDE_RAID_INFO
{
    u_char	ucNumLogicalDrives;
    u_char	ucMajorVersion;
    u_char	ucMinorVersion;
    MP_RAID_INFO_LD LogDrives [ MAX_DRIVES_PER_CONTROLLER ];
}
__attribute__ ((packed))
MP_IDE_RAID_INFO, *PMP_IDE_RAID_INFO;


typedef struct _MP_SPAREPOOL_DATA {
    u_long	ulNumberOfDrives;
    MP_RAID_INFO_PD Info[ MAX_DRIVES_PER_CONTROLLER ];
}
__attribute__ ((packed))
MP_SPAREPOOL_DATA, *PMP_SPAREPOOL_DATA;


typedef struct _MP_CHIPSET_INFO{
    u_short	usVendorID;
    u_short	usDeviceID;
    u_char	ucPciBus;
    u_char	ucPciDevice;
    u_char	ucPciFunction;
}
__attribute__ ((packed))
MP_CHIPSET_INFO;


typedef struct _MP_IDE_VERSION {
    u_short	usMajorVer;
    u_short	usMinorVer;
    u_long	ulBuild;
}
__attribute__ ((packed))
MP_IDE_VERSION, *PMP_IDE_VERSION;

typedef enum
{
    VERSION_IRCD = 0,
    VERSION_BIOS = 1,
    VERSION_DRIVER = 2
}VERSION_INFO;

typedef struct _MP_IDE_VERSION_EX {
    VERSION_INFO version;
    u_short	usMajorVer;
    u_short	usMinorVer;
    u_long	ulBuild;
}
__attribute__ ((packed))
MP_IDE_VERSION_EX, *PMP_IDE_VERSION_EX;

typedef struct _MP_PROBE_AMIRAID {
    u_char	acData[8];
    u_short	usControllerId;	/* ZERO based - based on pci detection*/
}
__attribute__ ((packed))
MP_PROBE_AMIRAID, *PMP_PROBE_AMIRAID;


typedef struct _MP_IDE_CONTROLLER_INFO{
    u_short	ulControllerId;	/* ZERO based - based on pci detection*/
    u_long	ulPrimaryBaseAddress;	/* 1st range primary*/
    u_long	ulPrimaryControlAddress; /* 2nd range*/
    u_long	ulSecondaryBaseAddress; /* 1st range primary*/
    u_long	ulSecondaryControllAddress; /* 2nd range*/
    u_long	ulBusMasterBaseAddress;
    u_short	usIRQ; /* IRQ*/
    MP_IDE_VERSION FwVersion; /* version of hyperdsk card*/
    MP_CHIPSET_INFO ChipsetInfo;	/* controller chipset*/
}
__attribute__ ((packed))
MP_IDE_CONTROLLER_INFO, *PMP_IDE_CONTROLLER_INFO;

typedef struct{
    u_long	ulControllerCount;
    MP_IDE_CONTROLLER_INFO Info[MAX_CONTROLLERS];
}
__attribute__ ((packed))
MP_CONTROLLER_DATA,*PMP_CONTROLLER_DATA;



typedef struct _MP_RAID_OPTIONS {
    u_long	ulErrorPolling;
    u_long	ulErrorThreshold;
    u_long	ulFailDrive;
    u_long	ulAutoRebuild;
    u_long	ulCheckArray;
    u_long	ulStopRebuild;
    u_long	ulIdentical;
    u_long	ulRebuildRate;
}
__attribute__ ((packed))
MP_RAID_OPTIONS, * PMP_RAID_OPTIONS;

/* +++ edm 1 March 2002 -*/
typedef struct _MP_RAID_OPTIONS_EX {
    u_long	ulErrorPolling;
    u_long	ulErrorThreshold;
    u_long	ulFailDrive;
    u_long	ulAutoRebuild;
    u_long	ulCheckArray;
    u_long	ulStopRebuild;
    u_long	ulIdentical;
    u_long	ulRebuildRate;
    u_long	ulDiskCoercion;
    u_long	ulAutoResume;
    u_long	ulLastActiveOperation[MAX_LOGICAL_DRIVES];
    u_long	ulLastSector[MAX_LOGICAL_DRIVES];
    u_long	ulReserved[20];
}
__attribute__ ((packed))
MP_RAID_OPTIONS_EX, * PMP_RAID_OPTIONS_EX;
/* --- edm 1 March 2002 -*/


typedef enum
{
    MP_OP_NOP	= 0,
    MP_OP_REBUILD	= 1,
    MP_OP_CHKCON	= 2,
    MP_OP_MKCON	= 3
}
MP_BACKGROUND_ACTIVITY;

typedef struct _MP_RAID_STATUS_INFO {
    MP_LOGICAL_DRIVE_STATUS ArrayStatus [ MAX_DRIVES_PER_CONTROLLER ];
    MP_PHYSICAL_DRIVE_STATUS PhyStatus [ MAX_DRIVES_PER_CONTROLLER ];
    BOOLEAN	CheckErrorLog [ MAX_DRIVES_PER_CONTROLLER ];
    MP_BACKGROUND_ACTIVITY CurrentArrayActivity [ MAX_DRIVES_PER_CONTROLLER ];
}
__attribute__ ((packed))
MP_RAID_STATUS_INFO, * PMP_RAID_STATUS_INFO;


typedef struct _MP_RAID_STATUS_INFO_EX {
    u_short	usSize;
    MP_LOGICAL_DRIVE_STATUS ArrayStatus [ MAX_LOGICAL_DRIVES ];
    MP_PHYSICAL_DRIVE_STATUS PhyStatus [ MAX_DRIVES_PER_CONTROLLER ];
    BOOLEAN	CheckErrorLog [ MAX_DRIVES_PER_CONTROLLER ];
    MP_BACKGROUND_ACTIVITY CurrentArrayActivity [ MAX_LOGICAL_DRIVES ];
    u_char	PowerConnected;
    u_char	ulReserved[19];
}
__attribute__ ((packed))
MP_RAID_STATUS_INFO_EX, * PMP_RAID_STATUS_INFO_EX;

typedef enum
{
    START_REBUILDING = 0,
    CONSISTENCY_CHECK = 1,
    MAKE_CONSISTENT = 2,
    ABORT_COMMAND = 3,
    DO_NEXT_BLOCK = 4
}
MP_ACTION_TYPES ;

typedef enum
{
    ACTIVITY_NOP = 0,
    ACTIVITY_SUCCEED = 1,
    ACTIVITY_ABORTED = 2
}
MP_ACTIVITY_STATUS;

typedef struct _MP_SEND_ACTIVITY {
    MP_ACTION_TYPES Action;
    u_long	ulBlockCount;
    u_long	ulErrorCount;
    MP_DRIVE_ID	SourceDriveId;
    MP_DRIVE_ID	TargetDriveId;
}
__attribute__ ((packed))
MP_SEND_ACTIVITY , * PMP_SEND_ACTIVITY ;

typedef struct _MP_SEND_ACTIVITY_EX {
    MP_ACTION_TYPES Action;
    u_long	ulBlockCount;
    u_long	ulErrorCount;
    MP_DRIVE_ID	SourceDriveId;
    MP_DRIVE_ID	TargetDriveId;
    u_char	ucLogicalDriveId;
    u_char	ucReserved[18];
}
__attribute__ ((packed))
MP_SEND_ACTIVITY_EX , * PMP_SEND_ACTIVITY_EX ;


typedef struct _MP_GET_ACTIVITY {
    u_long	ulLogicalDriveId;
    MP_BACKGROUND_ACTIVITY Action;
    u_long	ulTotBlockCount;
    u_long	ulDoneBlockCount;
    u_long	ulInconsistencyCount;
    MP_DRIVE_ID	SourceDriveId;
    MP_DRIVE_ID	TargetDriveId;
    u_long	StartTime;
}
__attribute__ ((packed))
MP_GET_ACTIVITY , * PMP_GET_ACTIVITY ;

typedef struct _MP_GET_ACTIVITY_EX {
    u_long	ulLogicalDriveId;
    MP_BACKGROUND_ACTIVITY Action;
    u_long	ulTotBlockCount;
    u_long	ulDoneBlockCount;
    u_long	ulInconsistencyCount;
    MP_DRIVE_ID	SourceDriveId;
    MP_DRIVE_ID	TargetDriveId;
    u_long	StartTime;
    u_long	ulErrorCount;
    MP_ACTIVITY_STATUS LastStatus;
    MP_BACKGROUND_ACTIVITY LastAction;
    u_char	ucReserved[4];
}
__attribute__ ((packed))
MP_GET_ACTIVITY_EX , * PMP_GET_ACTIVITY_EX ;

typedef struct _MP_FAIL_DRIVE {
    MP_DRIVE_ID	DriveId;
}
__attribute__ ((packed))
MP_FAIL_DRIVE , * PMP_FAIL_DRIVE;

typedef struct _MP_DETECT_DRIVE {
	MP_DRIVE_ID	DriveId;
	UCHAR		ucVersionNumber; /*not used now*/
	UCHAR		ucReserved[32];
}
__attribute__ ((packed))
MP_DETECT_DRIVE, *PMP_DETECT_DRIVE;


typedef struct _MP_SET_PDRV_STATUS{
    MP_DRIVE_ID			DriveId;
    MP_PHYSICAL_DRIVE_STATUS	DriveStatus;
}MP_SET_PDRV_STATUS,*PMP_SET_PDRV_STATUS;


typedef struct _PASS_THRU_DATA {
    UCHAR	uchTargetID; /* PM=0, PS=1, SM=2, SS=3*/
    ULONG	ulSize; /* size of data buffer*/
    BOOLEAN	bIsPIO; /* 0 = DMA Transfer , 1 = PIO Transfer*/
    BOOLEAN	bIsIn; 	/* 1 = reading from device, 0 = writing to device*/
    UCHAR	aucVersionNumber; /* Version of this IOCTL*/
    UCHAR	aucReserved[7];
    UCHAR	aucTaskFile[8]; /* IDE Task File Registers*/
    UCHAR	ucTaskFileBitMap; /* bitmask of registers needing to be written to*/
    UCHAR	ucStatusByte;
    UCHAR	ucErrorByte;  /* error register value*/
    UCHAR	aucReserved2[32];
    UCHAR	aucBuffer[512]; 	/* data buffer*/
}
__attribute__ ((packed))
PASS_THRU_DATA, * PPASS_THRU_DATA;

typedef struct _PASS_THRU_DATA_EX {
    UCHAR	uchTargetID; /* PM=0, PS=1, SM=2, SS=3*/
    ULONG	ulSize; /* size of data buffer*/
    BOOLEAN	bIsPIO; /* 0 = DMA Transfer , 1 = PIO Transfer*/
    BOOLEAN	bIsIn;	/* 1 = reading from device, 0 = writing to device*/
    UCHAR	aucVersionNumber; /* Version of this IOCTL 2 means using this structure*/
    UCHAR	aucReserved[7];
    UCHAR	aucTaskFile[8];	/* IDE Task File Registers*/
    UCHAR	ucTaskFileBitMap; /* bitmask of registers needing to be written to*/
    UCHAR	ucStatusByte;
    UCHAR	ucErrorByte; /* error register value*/
    UCHAR	aucTaskFile48bit[8];/*this is to hold 48 bit address*/
    UCHAR	ucTaskFileBitMap48bit; /* bitmask of registers needing to be written in 48 bit*/
    UCHAR	aucReserved2[23];
    UCHAR	aucBuffer[1]; 	/* data buffer User need to allocate memory*/
}
__attribute__ ((packed))
PASS_THRU_DATA_EX, * PPASS_THRU_DATA_EX;

/*Diagnostic SubCodes*/
#define DIAG_WRITE_DMA			0x00
#define DIAG_WRITE_MULTIPLE_PIO		0x01
#define DIAG_WRITE_SECTOR_PIO		0x02
#define DIAG_READ_DMA			0x03
#define DIAG_READ_MULTIPLE_PIO		0x04
#define DIAG_READ_VERIFY_SECTOR_PIO	0x05
#define DIAG_READ_SECTOR_PIO		0x06

/*Diagnostic Ioctl status*/
#define DIAG_SUCCESS			0x00
#define DIAG_MEDIUM_ERROR		0x01
#define DIAG_ICRC_ERROR			0x02
#define DIAG_INVALID_SECTOR		0x03
#define DIAG_INVALID_PARAMETERS		0x04

#define DIAG_TOTAL_SECTORS		0xFF /*256*/
#define DIAG_OFFSET_FROM_END		(3000*2) /*3 MB*/

typedef struct _DIAG_PASS_THRU_DATA {
    MP_DRIVE_ID	DriveID;		/*drive Id*/
    UCHAR	diag_sub_code; /*See Sub opcode definition*/
    ULONG	sector;	/*Offset into diagnostic region*/
    ULONG	data_transfer_length; /*in bytes*/
    UCHAR	status;	/*See return status*/
    UCHAR	reserved[32];
    PASS_THRU_DATA_EX passthru_data;
}
__attribute__ ((packed))
DIAG_PASS_THRU_DATA, * PDIAG_PASS_THRU_DATA;



/* ERROR LOG*/

#define	AR_ERRLOG_ERR_TYPE_UNKNOWN	0
#define	AR_ERRLOG_ERR_TYPE_READ		1
#define	AR_ERRLOG_ERR_TYPE_WRITE	2
#define	AR_ERRLOG_ERR_TYPE_VERIFY	3
#define	AR_ERRLOG_ERR_TYPE_DRIVE	4
#define	AR_ERRLOG_ERR_TYPE_TIMESTAMP	0x80 /* this is the MSB, decides time format*/



/*When the error type MSB is enabled, use the following time format*/
typedef struct _WIN9X_METHOD
{
    UCHAR	Hour;
    UCHAR	Minute;
    UCHAR	Second;
    UCHAR	Century;
    UCHAR	Year;
    UCHAR	Month;
    UCHAR	Day;
}
__attribute__ ((packed))
WIN9X_METHOD;


/*When the error type MSB is disabled, use the following time format*/
typedef struct _WIN_METHOD
{
    long long  time; /*System time count of 100 nano seconds*/
		/*since Jan 1, 1601*/
}
__attribute__ ((packed))
WIN_METHOD;



typedef struct _MP_ERROR_LOG_HEADER
{
    u_char	Signature[6]; /* $XERR$*/
    u_char	Size; /* Size of this structure in bytes*/
    u_char	SizeErrorLogStruct; /*Size of struct ERROR_LOG*/
    u_char	ErrorLogSectors; /* Number of sectors to use for error log*/
    u_short	MaxErrorCount; /*Maximum error log records*/
    u_short	TotalErrorCount; /* Total error log records*/
    u_short	Head; /* location of the newest entry in error log*/
    u_short	Tail; /* location of the oldest entry in error log*/
}
__attribute__ ((packed))
MP_ERROR_LOG_HEADER, *PMP_ERROR_LOG_HEADER;


typedef struct _MP_ERROR_LOG_RECORD
{
    u_char	err_type; /* error types ~ MSB desinates the timestamp*/
    u_char	err_sector[4]; /* sector where error occured*/
    u_char	err_timestamp[8]; /*  time stamp, based on MSB of "err_type"*/
}
__attribute__ ((packed))
MP_ERROR_LOG_RECORD,*PMP_ERROR_LOG_RECORD;


typedef struct _MP_ERROR_LOG
{
    MP_DRIVE_ID DriveID;
    MP_ERROR_LOG_HEADER errorlog_hdr;
    MP_ERROR_LOG_RECORD errorlog_rec[38];
}
__attribute__ ((packed))
MP_ERROR_LOG, *PMP_ERROR_LOG;

/* ioctls for backplane*/

typedef enum
{
    MP_SILENCE_BEEPER = 0,
    MP_SOUND_BEEPER = 1 ,
    MP_ENABLE_BEEPER = 2,
    MP_DISABLE_BEEPER =  3,
    MP_GET_STATUS = 4,
    MP_IDENTIFY_DRIVE = 5
}
MP_BACKPLANE_REQUEST;

typedef enum
{
    MP_DRIVE_PRESENT = 0,
    MP_DRIVE_NOT_PRESENT = 1
}
MP_DRIVE_BACKPLANE_STATUS;


typedef struct _MP_BACKPLANE_ACTIVITY
{
    MP_BACKPLANE_REQUEST Action; /* IN - the request , OUT - not used*/
    ULONG	ChannelId; /* IN - MP_IDENTIFY_DRIVE only, OUT - not used*/
    MP_DRIVE_BACKPLANE_STATUS ChannelStatus [ MAX_CHANNELS_PER_CONTROLLER ]; /* OUT - MP_GET_STATUS only*/
}
MP_BACKPLANE_ACTIVITY, *PMP_BACKPLANE_ACTIVITY ;

/* ioctls for backplane - end*/
typedef enum
{
    MP_SMART_READ_DATA	 = 0xD0,
    MP_SMART_ENABLE	 = 0xD8,
    MP_SMART_DISABLE	 = 0xD9,
    MP_SMART_READ_STATUS  = 0xDA,
}MP_SMART_ACTION_TYPES;

typedef struct _MP_SMART_ACTIVITY
{
    MP_DRIVE_ID	TargetID;
    MP_SMART_ACTION_TYPES Action;
    UCHAR	ucData[512];
}
__attribute__((packed))
MP_SMART_ACTIVITY, *PMP_SMART_ACTIVITY;


typedef enum
{
    MP_IRCD_READ = 0,
    MP_IRCD_WRITE = 1,
}MP_IRCD_ACTION;

typedef struct _MP_IRCD_ACTIVITY
{
    MP_DRIVE_ID	TargetID;
    MP_IRCD_ACTION Action;
    UCHAR	ucData[512];
}
__attribute__((packed))
MP_IRCD_ACTIVITY, *PMP_IRCD_ACTIVITY;

#endif /* __HD_IOCTLS_H__ */

/*vi set ts=8:*/
